/*  mdhd.c                                                      (CRC:9683441c)
 *      relative path: ./I386
 *      creation date: 1992
 *      last modified: Wed Dec 18 07:15:53 1996
 *      header update: Wed Dec 18 07:15:53 1996
 *  
 *       Intel 386 PC  machine dependent code
 *  
 *  Copyright (c) 1992 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is a part of the Shag/OS operating system distribution, and is
 *  free software for non-profit uses; it may be freely used and distributed
 *  for non-profit purposes under the terms of the Shag/OS License, version 1.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND, 
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 *  AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT WILL THE AUTHOR, OR
 *  ANYONE RESPONSIBLE FOR THE LEGAL MODIFICATION OR REDISTRIBUTION OF THIS
 *  SOFTWARE, BE LIABLE FOR ANY DAMAGES, INCLUDING BUT NOT LIMITED TO ANY LOST
 *  PROFITS, LOST SAVINGS OR ANY INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER
 *  RESULTING FROM IMPAIRED OR LOST DATA, SOFTWARE OR COMPUTER FAILURE OR ANY
 *  OTHER CAUSE, EVEN IF THE AUTHOR IS ADVISED OF THE POSSIBILITY OF SUCH
 *  DAMAGES, OR FOR ANY OTHER CLAIM BY THE USER OR FOR ANY THIRD PARTY CLAIM.
 *  
 *  You should have received a copy of the Shag/OS License along with this
 *  program, in a file called 'license1.txt'.  If this license is missing,
 *  please contact the author and describe how this software was obtained.
 *  
 *  The author can be reached for questions, comments, and suggestions at:
 *  
 *      Frank Barrus, 4435 River Rd. Apt #3, Scottsville, NY 14546, USA
 *      email:   shaggy@csh.rit.edu
 *  
 *      For sending bug reports, please read the file 'bugs', and always
 *      include the product name and version, and the word 'BUG' in the subject.
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

/* mdhd.c
 *  hard drive support for PC's
 *   - this file is included from drv/hd.c
 */

#include <limits.h>


#define NHD	2	/* number of hard drives hardware supports */
#define HDBLOCKSIZE 512	/* number of bytes per block */

typedef struct {
	ushort	nheads;
	ushort	nsecs;
	ushort	ncyls;
	ushort	wcomp;
	ushort	lzone;
	uchar	ctrl;
} HDtype;


struct partition {
	uchar	active;
	uchar	s_head;
	uchar	s_sec;
	uchar	s_cyl;
	uchar	type;
	uchar	e_head;
	uchar	e_sec;
	uchar	e_cyl;
	ulong	sec0;
	ulong	nsecs;
};


#define HD_CTRL	  0x3f6
#define HD_DATA   0x1f0
#define HD_ERROR  0x1f1
#define HD_WCOMP  0x1f1
#define HD_NSEC	  0x1f2
#define HD_SEC	  0x1f3
#define HD_LCYL	  0x1f4
#define HD_HCYL	  0x1f5
#define HD_HEAD	  0x1f6
#define HD_STATUS 0x1f7
#define HD_CMD	  0x1f7


/* commands: */
#define HD_CMD_RESTORE	0x10
#define HD_CMD_READ	0x20
#define HD_CMD_WRITE	0x30
#define HD_CMD_VERIFY	0x40
#define HD_CMD_FORMAT	0x50
#define HD_CMD_INIT	0x60
#define HD_CMD_SEEK	0x70
#define HD_CMD_DIAGNOSE	0x90
#define HD_CMD_SPECIFY	0x91

/* status: */
#define HD_STAT_BUSY	 0x80
#define HD_STAT_READY	 0x40
#define HD_STAT_WFAULT	 0x20
#define HD_STAT_SEEKDONE 0x10
#define HD_STAT_SVCBUF	 0x08
#define HD_STAT_ECC	 0x04
#define HD_STAT_INDEX	 0x02
#define HD_STAT_ERROR	 0x01

#define port_read(port,buf,nr) \
__asm__("cld;rep;insw"::"d"(port),"D"(buf),"c"(nr):"cx","di")

#define port_write(port,buf,nr) \
__asm__("cld;rep;outsw"::"d"(port),"S"(buf),"c"(nr):"cx","di")
 

extern label_t _param, _msleep;


static HDtype hd_type[2];
static HDinfo *cur_hd;
static ulong  cur_hd_block = ULONG_MAX;
static uchar  cur_hd_buf[HDBLOCKSIZE];
static uchar  cur_hd_cmd = 0;

static void hd_svc(int port, int irq)
{
/*	cprintf("HD intr %x,%x cmd=%x stat=%x\n",port,irq, cur_hd_cmd, in_portb(HD_STATUS));
*/
	switch(cur_hd_cmd) {
	case HD_CMD_READ:
		port_read(HD_DATA,cur_hd_buf,256);
		break;
	}
	cur_hd_cmd = 0;
}

static void hd_cmd(HDinfo *hd, int cmd, ulong block)
{
int cyl, head, sec, nsec;
int k1,k2;
	if(block >= hd->nblocks) {
		cprintf("HD seek out of range (%d/%d)\n",block,hd->nblocks);
		return;
	}
	block += hd->start;
	k1 = (hd_type[hd->num].nheads);
	k2 = (hd_type[hd->num].nsecs)*k1;
	cyl = block/k2;
	sec = (block%k2)/k1+1;
	head = block%k1;
	nsec = 1;
	while(in_portb(HD_STATUS)&HD_STAT_BUSY)
		_(OBJ_THREAD, _msleep, 0, ___, ___);
	out_portb(0xA0|(hd->num<<4)|head, HD_HEAD);
	while(!(in_portb(HD_STATUS)&HD_STAT_READY))
		_(OBJ_THREAD, _msleep, 0, ___, ___);
	out_portb(hd_type[hd->num].ctrl, HD_CTRL);
	out_portb(hd_type[hd->num].wcomp, HD_WCOMP);
	out_portb(nsec, HD_NSEC);
	out_portb(sec, HD_SEC);
	out_portb(cyl&0xff, HD_LCYL);
	out_portb(cyl>>8, HD_HCYL);
	cur_hd = hd;
	cur_hd_cmd = cmd;
	cur_hd_block = block;
	out_portb(cmd, HD_CMD);
}


static void hd_readblock(HDinfo *hd, uchar *buf, ulong block)
{
	while(cur_hd_block != block || cur_hd != hd) {
		hd_cmd(hd,HD_CMD_READ,block);
		while(cur_hd_cmd)
			_(OBJ_THREAD, _msleep, 0, ___, ___);
		if(in_portb(HD_STATUS)&(HD_STAT_ECC|HD_STAT_ERROR)) 
			cprintf("HD err: status=%x error=%x\n",
				in_portb(HD_STATUS), in_portb(HD_ERROR));
	}
	memcpy(buf,cur_hd_buf,hd->blocksize);
}


#ifdef XXX
static void hd_writeblock(HDinfo *hd, uchar *buf, ulong block)
{
	hd_cmd(hd,HD_CMD_READ,block);
	port_write(HD_DATA,buf,256);
	while(cur_hd_cmd)
		_(OBJ_THREAD, _msleep, 0, ___, ___);
	if(in_portb(HD_STATUS)&(HD_STAT_ECC|HD_STAT_ERROR)) {
		cprintf("HD err: status=%x error=%x\n",
			in_portb(HD_STATUS), in_portb(HD_ERROR));
	}
}
#endif

static boolean read_hd_info(HDinfo *hd, int n)
{
	if(n != 0)
		return(FALSE);
	hd_type[n].nheads = 4;
	hd_type[n].nsecs = 17;
	hd_type[n].ncyls = 306;
	hd_type[n].lzone = 306;
	hd_type[n].wcomp = 128;
	hd_type[n].ctrl = 0;
	hd->num = n;
	hd->start = 0;
	hd->nblocks = hd_type[n].nheads*hd_type[n].nsecs*hd_type[n].ncyls;
	hd->blocksize = HDBLOCKSIZE;
	hd->type = 0;
	return(TRUE);
}



static boolean read_partition(HDinfo *partition, HDinfo *hd, int n)
{
uchar buf[HDBLOCKSIZE];
struct partition *p;

	if(n<0 || n>3)
		return(FALSE);
	memset(buf,0,hd->blocksize);
	hd_readblock(hd, buf, 0);
	if(*(ushort*)(buf+510) == 0xaa55) {
		p = &((struct partition*)(buf+446))[n];
		if(!p->nsecs)
			return(FALSE);
		partition->start = p->sec0;
		partition->nblocks = p->nsecs;
		partition->type = p->type;
		return(TRUE);
	}
	return(FALSE);
}


static __inline__ void mdhd_init(void)
{
	set_interrupt((vaddr)hd_svc, 0, INT_HD);
}

