/*  mdresource.h                                                (CRC:e060791e)
 *      relative path: ./I386
 *      creation date: 1992
 *      last modified: Wed Dec 18 07:15:58 1996
 *      header update: Wed Dec 18 07:15:58 1996
 *  
 *       Intel 386 PC  machine dependent code
 *  
 *  Copyright (c) 1992 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is a part of the Shag/OS operating system distribution, and is
 *  free software for non-profit uses; it may be freely used and distributed
 *  for non-profit purposes under the terms of the Shag/OS License, version 1.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND, 
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 *  AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT WILL THE AUTHOR, OR
 *  ANYONE RESPONSIBLE FOR THE LEGAL MODIFICATION OR REDISTRIBUTION OF THIS
 *  SOFTWARE, BE LIABLE FOR ANY DAMAGES, INCLUDING BUT NOT LIMITED TO ANY LOST
 *  PROFITS, LOST SAVINGS OR ANY INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER
 *  RESULTING FROM IMPAIRED OR LOST DATA, SOFTWARE OR COMPUTER FAILURE OR ANY
 *  OTHER CAUSE, EVEN IF THE AUTHOR IS ADVISED OF THE POSSIBILITY OF SUCH
 *  DAMAGES, OR FOR ANY OTHER CLAIM BY THE USER OR FOR ANY THIRD PARTY CLAIM.
 *  
 *  You should have received a copy of the Shag/OS License along with this
 *  program, in a file called 'license1.txt'.  If this license is missing,
 *  please contact the author and describe how this software was obtained.
 *  
 *  The author can be reached for questions, comments, and suggestions at:
 *  
 *      Frank Barrus, 4435 River Rd. Apt #3, Scottsville, NY 14546, USA
 *      email:   shaggy@csh.rit.edu
 *  
 *      For sending bug reports, please read the file 'bugs', and always
 *      include the product name and version, and the word 'BUG' in the subject.
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

#ifndef __MDRESOURCE_H__
#define __MDRESOURCE_H__

#include <mdmacro.h>
#include <segment.h>
#include <vmem.h>
#include <context.h>
#ifndef ASM
typedef void intr(int param, int irq);


#ifdef KERNEL


/* video.c: */
extern void console_getxy(short *xpos, short *ypos);
extern void console_setxyvars(short *xpos, short *ypos);
extern void video_init(void);
extern int console_write(char *buf, int len);
extern void console_beep(int freq, int delay);
extern void heartbeat(int i);
extern void heartbeat_toggle(void);


/* console.c: */
extern void console_init(void);


/* system.c: */
struct date;
extern void system_init(void);
extern void bootparam_print(void);
extern void reboot(void);
extern void halt(void);
extern uchar read_cmos_clock(int param);
extern void date_get(struct date *date);
extern void trap_svc(ulong type, ulong errcode, vaddr pc, vaddr va);
extern void context_print(void (*eprintf)(char *fmt,...), Context *context);
extern vaddr pc_traceback(int level);
extern void stack_traceback(Context *context, boolean symtrace);
extern void clock_set_hz(ushort hz);


/* interrupt.c: */
extern void interrupt_init(void);
extern void set_interrupt(vaddr func, int param, int irq);

/* vmem.c: */
extern void vm_pagefault(ulong param, int ivec, vaddr pc, vaddr va, ulong reason);
extern void vm_accv(ulong param, int ivec, vaddr pc, vaddr va, ulong reason);
extern vaddr vm_pagedir_create(void);
extern void vm_pagedir_free(vaddr pagedir);
extern void vm_restore(void);

/* stubs.s: */
extern void interrupt_enable(void);
extern void clock_enable(void);
extern void clock_disable(void);
extern void clock_set_interval(int hertz);
extern boolean is_kernel(void);
extern obj_t call_user(vaddr func, ObjFrame *frame, void *data,
		obj_t arg1, obj_t arg2, obj_t arg3);


/* object calls: */

struct objcall_frame {
	obj_t	oid;
	obj_t	op;
	obj_t	arg1;
	obj_t	arg2;
	obj_t	arg3;
};

extern obj_t ObjCall_svc(register struct objcall_frame *callframe,
			 obj_t errchk, obj_t self, vaddr pc);

#endif /* KERNEL */

#ifdef XXX
/* asm.s: */

extern void out_portb(uchar value, int port);
extern void out_portw(ushort value, int port);
extern void out_regb(uchar reg, uchar val, int port);
extern void out_regw(uchar reg, ushort val, int port);
extern uchar in_portb(int port);
extern ushort in_portw(int port);
extern uchar in_regb(uchar reg, int port);
extern ushort in_regw(uchar reg, int port);
extern uchar cmos_read(uchar adr);
extern void cmos_write(uchar val, uchar adr);
#endif
#endif /* ASM */
#endif /* __MDRESOURCE_H__ */

