/*  mdserial.c                                                  (CRC:fff3baa6)
 *      relative path: ./I386
 *      creation date: 1992
 *      last modified: Wed Dec 18 07:15:59 1996
 *      header update: Wed Dec 18 07:16:00 1996
 *  
 *       Intel 386 PC  machine dependent code
 *  
 *  Copyright (c) 1992 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is a part of the Shag/OS operating system distribution, and is
 *  free software for non-profit uses; it may be freely used and distributed
 *  for non-profit purposes under the terms of the Shag/OS License, version 1.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND, 
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 *  AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT WILL THE AUTHOR, OR
 *  ANYONE RESPONSIBLE FOR THE LEGAL MODIFICATION OR REDISTRIBUTION OF THIS
 *  SOFTWARE, BE LIABLE FOR ANY DAMAGES, INCLUDING BUT NOT LIMITED TO ANY LOST
 *  PROFITS, LOST SAVINGS OR ANY INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER
 *  RESULTING FROM IMPAIRED OR LOST DATA, SOFTWARE OR COMPUTER FAILURE OR ANY
 *  OTHER CAUSE, EVEN IF THE AUTHOR IS ADVISED OF THE POSSIBILITY OF SUCH
 *  DAMAGES, OR FOR ANY OTHER CLAIM BY THE USER OR FOR ANY THIRD PARTY CLAIM.
 *  
 *  You should have received a copy of the Shag/OS License along with this
 *  program, in a file called 'license1.txt'.  If this license is missing,
 *  please contact the author and describe how this software was obtained.
 *  
 *  The author can be reached for questions, comments, and suggestions at:
 *  
 *      Frank Barrus, 4435 River Rd. Apt #3, Scottsville, NY 14546, USA
 *      email:   shaggy@csh.rit.edu
 *  
 *      For sending bug reports, please read the file 'bugs', and always
 *      include the product name and version, and the word 'BUG' in the subject.
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

/* mdserial.c
 *  Serial I/O for PC's - to be included from drv/serial.c
 */


#define NSERIAL 4


char *devname[NSERIAL] = {"ttys0","ttys1","ttys2","ttys3"};
int devspeed[NSERIAL] = {9600,38400,9600,9600};
char *devparam[NSERIAL] = {"n81","n81","n81","n81"};

#define IS_TTY(n)	TRUE

obj_t ser_tty[NSERIAL];
SerialBuf *ser_idata[NSERIAL];
SerialBuf *ser_odata[NSERIAL];


static ushort ser_base[4] = {0x3f8, 0x2f8, 0x3e8, 0x2e8};
static uchar ser_irq[4] = {INT_SER0,INT_SER1,INT_SER0,INT_SER1};

#define SER_THR		0	/* LCR:7 = 0 */
#define SER_RDR		0	/* LCR:7 = 0 */
#define SER_BRDL	0	/* LCR:7 = 1 */
#define	SER_BRDH	1	/* LCR:7 = 1 */
#define SER_IER		1	/* LCR:7 = 0 */
#define SER_IIR		2
#define SER_LCR		3
#define SER_MCR		4
#define SER_LSR		5
#define SER_MSR		6

#define UART_HERTZ	1843200		/* 1.8432 Mhz */


extern label_t _param, _data, _speed, _break, _msleep;

static void serial_reset(int port);


static status_t ser_write(int chan)
{
int data;
	if(in_portb(ser_base[chan]+SER_LSR) & 0x20) 
		if((data=serbuf_readchar(ser_odata[chan])) >= 0) 
			out_portb(data,ser_base[chan]+SER_THR);
	return(OK);
}

static void serial_svc(int chan, int irq)
{
uchar iir;
int data;
register int base;
	base = ser_base[chan];
	while(!((iir = in_portb(base+SER_IIR)) & 1)) {
		switch(iir) {
		case 0:
			in_portb(base+SER_LSR);
			break;
		case 2:	/* transmitter ready */
			while( (in_portb(base+SER_LSR) & 0x20) 
			 && (data=serbuf_readchar(ser_odata[chan])) >= 0 ) 
				out_portb(data,base+SER_THR);
			break;
		case 4:	/* data received */
			while(in_portb(base+SER_LSR) & 1) {
				data = in_portb(ser_base[chan]+SER_RDR);
				serbuf_writechar(ser_idata[chan],data);
			}
			break;
		case 6:
			in_portb(base+SER_MSR);
			break;
		}
	}
}


static void serial_setline(int port, int speed, char *param)
{
int psel;
	switch(param[0]) {
		case 'o': case 'O': psel = 0x20; break;
		case 'e': case 'E': psel = 0x30; break;
		case 'm': case 'M': psel = 0x28; break;
		case 's': case 'S': psel = 0x38; break;
		case 'i': case 'I': 
		case 'n': case 'N': 	
		default:  psel = 0x00; 
	}
	out_portb(0x80, ser_base[port]+SER_LCR);
	out_portw(UART_HERTZ/(16*speed), ser_base[port]+SER_BRDL);
	out_portb((param[1]-'0'-5) | psel, ser_base[port]+SER_LCR);
	serial_reset(port);
}


static void serial_break(int port)
{
uchar b;
	b = in_portb(ser_base[port]+SER_LCR);
	out_portb(0x40|b, ser_base[port]+SER_LCR);
	_(OBJ_THREAD, _msleep, 1000, ___, ___);
	out_portb(b, ser_base[port]+SER_LCR);
}

static void serial_hangup(int port)
{
uchar b;
	b = in_portb(ser_base[port]+SER_MCR);
	out_portb(b&0xfe, ser_base[port]+SER_MCR);
	_(OBJ_THREAD, _msleep, 1000, ___, ___);
	out_portb(b, ser_base[port]+SER_MCR);
}


static void serial_reset(int port)
{
int i;
/* XXX - this works -- but isn't there an easier way? */
	in_portb(ser_base[port]+SER_IIR);
	in_portb(ser_base[port]+SER_MSR);
	in_portb(ser_base[port]+SER_IIR);
	in_portb(ser_base[port]+SER_LSR);
	in_portb(ser_base[port]+SER_IIR);
	for(i=0; i<16; i++) {
		in_portb(ser_base[port]+SER_RDR);
		if (!(in_portb(ser_base[port]+SER_LSR) & 1))
			break;
	}
	in_portb(ser_base[port]+SER_IIR);
	in_portb(ser_base[port]+SER_LSR);
}


static __inline__ void mdserial_init(void)
{
}

static __inline__ boolean serial_init_port(int i, char *msg)
{
	out_portb(0x0, ser_base[i]+SER_LCR);
	out_portb(0xb, ser_base[i]+SER_MCR);
	out_portb(0x3, ser_base[i]+SER_IER);
	if((in_portb(ser_base[i]+SER_IER) & 0xf0) == 0x00) {
		serial_reset(i);
		set_interrupt((vaddr)serial_svc, i, ser_irq[i]);
		sprintf(msg,"port 0x%x  irq %d", ser_base[i], ser_irq[i]);
		return(TRUE);
	} 
	return(FALSE);
}

