/*  mdthread.c                                                  (CRC:905b88ce)
 *      relative path: ./I386
 *      creation date: 1992
 *      last modified: Wed Dec 18 07:16:00 1996
 *      header update: Wed Dec 18 07:16:01 1996
 *  
 *       Intel 386 PC  machine dependent code
 *  
 *  Copyright (c) 1992 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is a part of the Shag/OS operating system distribution, and is
 *  free software for non-profit uses; it may be freely used and distributed
 *  for non-profit purposes under the terms of the Shag/OS License, version 1.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND, 
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 *  AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT WILL THE AUTHOR, OR
 *  ANYONE RESPONSIBLE FOR THE LEGAL MODIFICATION OR REDISTRIBUTION OF THIS
 *  SOFTWARE, BE LIABLE FOR ANY DAMAGES, INCLUDING BUT NOT LIMITED TO ANY LOST
 *  PROFITS, LOST SAVINGS OR ANY INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER
 *  RESULTING FROM IMPAIRED OR LOST DATA, SOFTWARE OR COMPUTER FAILURE OR ANY
 *  OTHER CAUSE, EVEN IF THE AUTHOR IS ADVISED OF THE POSSIBILITY OF SUCH
 *  DAMAGES, OR FOR ANY OTHER CLAIM BY THE USER OR FOR ANY THIRD PARTY CLAIM.
 *  
 *  You should have received a copy of the Shag/OS License along with this
 *  program, in a file called 'license1.txt'.  If this license is missing,
 *  please contact the author and describe how this software was obtained.
 *  
 *  The author can be reached for questions, comments, and suggestions at:
 *  
 *      Frank Barrus, 4435 River Rd. Apt #3, Scottsville, NY 14546, USA
 *      email:   shaggy@csh.rit.edu
 *  
 *      For sending bug reports, please read the file 'bugs', and always
 *      include the product name and version, and the word 'BUG' in the subject.
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

/* mdthread.c
 *  machine dependent thread control
 *  for I386 architectures
 *  to be included by sys/thread.c
 */



extern label_t _data;



static __inline__ void context_switch(void)
{
	gdt[GDT_THREAD] = GDT_TSS((vaddr)thread_context,sizeof(Context));
	cswitch_enable = TRUE;
	if(int_level == 0) 
		__asm__("ljmp %0, $0"::"i"(SEG_THREAD));
	else 
		cswitch_force = TRUE;
		/* when int_level = 0, the switch will occur */
}

static __inline__ void context_set(Context *context)
{
	gdt[GDT_THREAD] = GDT_TSS((vaddr)context,sizeof(Context)); 
	/* XXX - should adjust for SEG_THREAD values: */
	__asm__("movw %0,%%ax"::"i"(SEG_THREAD));
	__asm__("ltr %ax"); 
}

#ifdef XXX
Context *context_get(void)
{
vaddr n;
	n = ((gdt[4].lo >> 16)&0xffff) | ((gdt[4].hi & 0xff)<<16)
	    |(gdt[4].hi & 0xff000000);
	return((Context*)n);
}
#endif

static void force_exit(void)
{
obj_t rv;
	__asm__("movl %%eax,%0":"g="(rv));	/* get return value */
	err_(OBJ_THREAD,_free,rv,___,___);
	while(1)
		continue;
}

static __inline__ void mdthread_spawn(Thread *new, vaddr func, ulong param)
{
ulong *sp;
vaddr tmp;
extern void force_exit(void);
	new->context.pagedir = vm_pagedir_create();
	new->context.pdbr = vm_getphys(new->context.pagedir);
	new->context.pc = func;
	new->context.eflags = 0x200;	/* enable interrupts */
	new->context.usp = new->ustack;
	if(func < KERNBASE) {
		new->context.seg[SEG_CS] = SEG_USER_CODE+RPL_USER;
		new->context.seg[SEG_DS] = SEG_USER_DATA+RPL_USER;
		new->context.seg[SEG_ES] = SEG_USER_DATA+RPL_USER;
		new->context.seg[SEG_SS] = SEG_USER_DATA+RPL_USER;
		sp = (ulong*)new->ustack;
	} else {
		new->context.seg[SEG_CS] = SEG_KERN_CODE;
		new->context.seg[SEG_DS] = SEG_KERN_DATA;
		new->context.seg[SEG_ES] = SEG_KERN_DATA;
		new->context.seg[SEG_SS] = SEG_KERN_DATA;
		sp = (ulong*)new->kstack;
	}
	tmp = vm_temp_share(&new->stack, (vaddr)sp - PAGE_SIZE, 
				PAGE_SIZE, VM_KRW);
	new->context.reg[REG_ESP] = (vaddr)sp - 12;
	new->context.reg[REG_EBP] = 0;
	new->context.sp[0].esp = new->kstack;
	sp = (ulong*)(tmp+PAGE_SIZE);
	*(--sp) = 0;		/* dummy- just in case */
	*(--sp) = param; 	/* put parameter on stack */
	*(--sp) = (vaddr)force_exit;	/* put a forced exit on stack */
	vm_temp_free(tmp);
}


static __inline__ void mdthread_destroy(Thread *node)
{
	vm_pagedir_free(node->context.pagedir);
}



/* set up the initial thread's context */
static __inline__ void mdthread_init(void)
{
extern vaddr kpagedir;
	/* set ldt, sp[0-2], debug, iomap -- assumed to be 0 */
	thread_context->pagedir = kpagedir;
	thread_context->pdbr = vm_getphys(kpagedir);
	thread_context->sp[0].ss = SEG_KERN_DATA;
	kprintf("pdbr=%x ", thread_context->pagedir,
			thread_context->pdbr);
	cswitch_force = FALSE;
}

