/*  $Id: mdvct.c,v 1.1 1996/11/12 03:24:53 shaggy Exp shaggy $
 *   driver for basic vt100 console I/O
 *   to be included from drv/vct.c
 *
 *      relative path: I386
 *  
 *       Intel 386 PC  machine dependent code
 *  
 *  Copyright (C) 1992  Frank E. Barrus
 *  
 *  This file is a part of the Shag/OS operating system distribution.
 *  
 *  Shag/OS is free software, for non-profit uses; it may be freely distributed
 *  for non-profit purposes under the terms of the Shag/OS License, version 0.
 *  This software comes with absolutely no warranty, expressed or implied.
 *  
 *  You should have received a copy of version 0 of the Shag/OS License along
 *  with this program, in a file called 'COPYING'.  If you did not receive this
 *  license, please contact the author and describe how this file was received.
 *  
 *  The author can be reached for questions, comments,
 *  suggestions, and bug-reports at:
 *  
 *      Frank Barrus, 4435 River Rd. Apt #3, Scottsville, NY 14546, USA
 *      email:   shaggy@csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 *  
 */

/*  $Log: mdvct.c,v $
 *  Revision 1.1  1996/11/12  03:24:53  shaggy
 *  Initial revision
 *
 *      creation date: 1992
 */


#include <stdio.h>
#include "videoparam.h"

static void console_in(void);
static void console_out(int vcnum);
static void key_init(void);
static void kb_setrate(uchar rate, uchar delay);
static void kbd_svc(int param, int irq);


extern label_t _rnoblock, _wnoblock;
extern label_t _stdin, _stdout, _stderr;
extern label_t _wid, _hgt;
extern label_t _reboot, _halt;
extern label_t _msleep;


#define VT_SET_CURSOR(vt,x,y) \
	if(vct == (vt)->id)     \
	out_regw(0x0e,(x)+((y)+ystart[(vt)->id])*(vt)->wid,crtc_adr); 



static ushort *video_base;
static short crtc_adr;

#define NVCT	8
#define MAX_INPUT	256
#define MAX_OUTPUT	2048

static struct vt_state vtstate[NVCT];
static obj_t tty[NVCT], tty_in[NVCT], tty_out[NVCT];
static ushort *framebuf[NVCT];
static boolean suspend[NVCT];
static int ystart[NVCT];

static int vct;
static int whgt;

#define KEYBUFSIZE 64
static uchar keybuf[KEYBUFSIZE];
static int key_in, key_out;




obj_t Main(ObjFrame *frame, vaddr base, obj_t arg, obj_t env)
{
paddr vbase;
struct videoparam *videoparam;
obj_t rv;
char buf[40];
int i;
int wid,hgt;
obj_t class_buffer, class_tty;
	cprintf("Virtual Console Terminal (vct) v0.10  Dec 1992 (irq1=%x)\n",
			kbd_svc);
	key_init();

	videoparam = (struct videoparam*)dir_parse(OBJ_ROOT,"/lib/mkern/videoparam");

	wid = videoparam->wid;
	hgt = videoparam->hgt;
	whgt = hgt/4;

	video_base = videoparam->video_base;
	// (ushort*)_(OBJ_TASK, _vm_map, vbase, wid*hgt*2, ___);
	crtc_adr = videoparam->crtc_adr;
	crtc_adr = 0x3b4;

	class_buffer = dir_parse(OBJ_ROOT,"/class/Buffer");
	class_tty = dir_parse(OBJ_ROOT,"/class/TTY");

	for(i=0; i<NVCT; i++) {
		vtstate[i].wid = wid;
		if(i<4) {
			vtstate[i].hgt = hgt;
			ystart[i] = 0;
		} else {
			vtstate[i].hgt = whgt;
			ystart[i] = (i-4)*whgt;
		}
		framebuf[i] = (ushort*)_(OBJ_TASK, _mem_alloc,
				 vtstate[i].wid*vtstate[i].hgt*2, ___, ___);
		memset_w(framebuf[i],0x700,vtstate[i].wid*vtstate[i].hgt);
		if(i==0)
			vtstate[i].base = video_base;
		else
			vtstate[i].base = framebuf[i];
		vtstate[i].id = i;
		vt_init(&vtstate[i]);	
		suspend[i] = FALSE;
		tty_in[i] = _(class_buffer, _new, ___, ___, ___);
		tty_out[i] = _(class_buffer, _new, ___, ___, ___);
		tty[i] = _(class_tty, _new, ___, ___, ___);
		_(tty[i], _set, _stdin, tty_in[i], ___);
		_(tty[i], _set, _stdout, tty_out[i], ___);
		_(tty[i], _set, _wid, vtstate[i].wid, ___);
		_(tty[i], _set, _hgt, vtstate[i].hgt, ___);
		sprintf(buf,"/dev/tty%d", i);
		rv = dir_create(OBJ_ROOT,buf,tty[i]); 
		if(err(rv))
			cprintf("\t%s: %s\n", buf, strerror(rv));
		else
			cprintf("\t%s: %dx%d\n", buf, vtstate[i].wid, vtstate[i].hgt);
		_(OBJ_THREAD, _set, _stdin, tty_out[i], ___);
		_(OBJ_THREAD, _set, _stdout, tty_in[i], ___);
		_(OBJ_THREAD, _set, _stderr, tty[i], ___);
		sprintf(buf,"vct_tty%d", i);
		_(OBJ_THREAD, _new, Label(buf), Ptr(console_out), i);
	}
	vct = 0;
	key_in = 0;
	key_out = 0;

	set_interrupt((vaddr)kbd_svc,0,INT_KB);

	console_getxy(&vtstate[vct].xpos, &vtstate[vct].ypos);
	console_setxyvars(&vtstate[vct].xpos, &vtstate[vct].ypos);

	_(OBJ_THREAD, _new, Label("vct_kb"), Ptr(console_in), 0);

	kb_setrate(0,1);

	return(OK);
}


static void console_out(int vcnum)
{
uchar cb[MAX_OUTPUT];
int len;
boolean active;
	while(1) {
		active = FALSE;
		if(!suspend[vcnum] &&
		 ((len = _(tty[vcnum], _devread, Ptr(cb), 2048, ___)) > 0)) {
			active = TRUE;
			vt_write(&vtstate[vcnum], cb, len);
		}
		if(!active)
			_(OBJ_THREAD, _msleep, 0, ___, ___);
	}
}

static void console_in(void)
{
uchar buf[64];
int n;	
	while(1) {
		n = 0;
		while(key_in != key_out) {
			buf[n++] = keybuf[key_out];
			key_out = (key_out+1)%KEYBUFSIZE;
		}
		if(n) 
			_(tty[vct], _devwrite, Ptr(buf), n, ___);
		else {
			_(OBJ_THREAD, _msleep, 0, ___, ___);
		}
	}
}


static void vct_switch(int n)
{
int i;
	if(vct < 4) {
		memcpy(framebuf[vct],vtstate[vct].base,
			vtstate[vct].wid*vtstate[vct].hgt*2);
		vtstate[vct].base = framebuf[vct];
	} else {
		for(i=4; i<NVCT; i++) {
			memcpy(framebuf[i],vtstate[i].base,
				vtstate[i].wid*vtstate[i].hgt*2);
			vtstate[i].base = framebuf[i];
		}	
	}
	vct = n;
	if(vct < 4) {
		memcpy(video_base,framebuf[vct],
			vtstate[vct].wid*vtstate[vct].hgt*2);
		vtstate[vct].base = video_base;
	} else {
		for(i=4; i<NVCT; i++) {
			vtstate[i].base = video_base
				+(i-4)*(vtstate[i].wid)*whgt;
			memcpy(vtstate[i].base, framebuf[i],
				vtstate[i].wid*whgt*2);
		}
	}
	VT_SET_CURSOR(&vtstate[vct], vtstate[vct].xpos, 
		vtstate[vct].ypos+ystart[vct]);
	console_setxyvars(&vtstate[vct].xpos, &vtstate[vct].ypos);
}


static void video_bell(int freq, int delay)
{
int n;
int oldkb;
	n = TIMER_HERTZ/freq;
	/* set PIT2 (programmable interrupt timer) to frequency: */ 
	out_portb(0xb6,0x43);
	out_portb(n&0xff,0x42);
	out_portb(n>>8,0x42);
	/* activate speaker: */
	oldkb = in_portb(0x61);
	out_portb(oldkb | 3, 0x61);	
	_(OBJ_THREAD, _msleep, delay, ___, ___);
	/* turn it off: */
	out_portb(oldkb, 0x61);	
}



static void video_scroll_up(struct vt_state *vt,
			int x1, int y1, int x2, int y2, int n)
/* XXX - currently x1 and x2 ignored */
{
	if(x1 > x2 || y1 > y2)
		return;
	memcpy(&vt->base[y1*vt->wid],
	      &vt->base[(y1+n)*vt->wid], (vt->wid)*(y2-y1-n+1)*sizeof(short));
	memset_w(&vt->base[(vt->wid)*(y2-n+1)], vt->atval, vt->wid*n);
}


static void video_scroll_down(struct vt_state *vt,
			int x1, int y1, int x2, int y2, int n)
{
int i;
	if(x1 > x2 || y1 > y2)
		return;
	i = y2;
	while(i >= y1+n) {
		memcpy(&vt->base[i*vt->wid+x1],
		      &vt->base[(i-n)*vt->wid+x1],
		      (x2-x1+1)*sizeof(short));
		i--;
	}	
	while(i >= y1) { 
		memset_w(&vt->base[i*(vt->wid)+x1], vt->atval, (x2-x1+1));
		i--;
	}
}

static void video_scroll_right(struct vt_state *vt,
			 int x1, int y1, int x2, int y2, int n)
{
int y,yofs,i;
	if(x1 > x2 || y1 > y2)
		return;
	for(y=y1; y<=y2; y++) {
		yofs = y*vt->wid;
		i = x2;
		while(i >= x1+n) {		
			vt->base[yofs+i] = vt->base[yofs+i-n];
			i--;
		}
	}	
}

static void video_scroll_left(struct vt_state *vt,
			int x1, int y1, int x2, int y2, int n)
{
int y,yofs,i;
	if(x1 > x2 || y1 > y2)
		return;
	for(y=y1; y<=y2; y++) {
		yofs = y*vt->wid;
		i = x1;
		while(i <= x2-n) {		
			vt->base[yofs+i] = vt->base[yofs+i+n];
			i++;
		}
	}	
}


static void video_clear(struct vt_state *vt,
			int x1, int y1, int x2, int y2)
{
int a1,a2;
	if(x1 > x2 || y1 > y2)
		return;
	a1 = y1*vt->wid	+ x1;
	a2 = y2*vt->wid + x2;
	memset_w(&vt->base[a1], vt->atval, a2-a1+1);		
}

static void video_setcolor(struct vt_state *vt)
{
uchar fg,bg,t;
	fg = vt->fg | ((vt->attr&2) << 2);
	bg = vt->bg;
	if(vt->attr&0x80) {
		t = fg; fg = bg; bg = t;
	}
	vt->atval = ((bg&15) << 12) | ((fg&15) << 8);
}



#define VT_PUTCHAR(vt,x,y,ch)		(vt)->base[(y)*((vt)->wid)+(x)] \
					= (ch) | (vt)->atval;

#define VT_SCROLL_UP(vt,x1,y1,x2,y2,n)    video_scroll_up(vt,x1,y1,x2,y2,n);
#define VT_SCROLL_DOWN(vt,x1,y1,x2,y2,n)  video_scroll_down(vt,x1,y1,x2,y2,n);
#define VT_SCROLL_RIGHT(vt,x1,y1,x2,y2,n) video_scroll_right(vt,x1,y1,x2,y2,n);
#define VT_SCROLL_LEFT(vt,x1,y1,x2,y2,n)  video_scroll_left(vt,x1,y1,x2,y2,n);
#define VT_CLEAR(vt,x1,y1,x2,y2)	  video_clear(vt,x1,y1,x2,y2);

#define VT_SET_COLOR(vt)		video_setcolor(vt);
#define VT_BEEP(vt)			video_bell(800,20);



#define NONE	0xff
 
static uchar key_unshift[] = {	
0,     033  , '1'  , '2'  , '3'  , '4'  , '5'  , '6'  ,	 /* scan  0- 7 */ 
'7'  , '8'  , '9'  , '0'  , '-'  , '='  , 8 ,'\t'  ,	 /* scan  8-15 */ 
 
'q'  , 'w'  , 'e'  , 'r'  , 't'  , 'y'  , 'u'  , 'i'  ,	 /* scan 16-23 */ 
'o'  , 'p'  , '['  , ']'  , '\r' , NONE  , 'a'  , 's'  , /* scan 24-31 */ 
 
'd'  , 'f'  , 'g'  , 'h'  , 'j'  , 'k'  , 'l'  , ';'  ,  /* scan 32-39 */ 
'\'' , '`'  , NONE  , '\\' , 'z'  , 'x'  , 'c'  , 'v'  , /* scan 40-47 */ 
 
'b'  , 'n'  , 'm'  , ','  , '.'  , '/'  , NONE  ,   '*', /* scan 48-55 */ 
NONE  , ' '  , NONE,     'f',     0, ' '  ,     0,     0, /* scan 56-63 */ 
 
    0,     0,     0,     0,     0, NONE, NONE,   '7',	 /* scan 64-71 */ 
  '8',   '9',   '-',   '4',   '5',   '6',   '+',   '1',	 /* scan 72-79 */ 
 
  '2',   '3',   '0',   '.',     0,     0,     0,     0,	 /* scan 80-87 */ 
0,0,0,0,0,0,0,0, 
0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 
0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,	

0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,	
0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,	
0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,	
0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,	
0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,	
0,0,0,127,0,0,0,0, 0,0,0,0,0,0,0,0,	
0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,	
0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,	
} ; 
 
static uchar key_shift[] = {
0,     033  , '!'  , '@'  , '#'  , '$'  , '%'  , '^'  ,	 /* scan  0- 7 */ 
'&'  , '*'  , '('  , ')'  , '_'  , '+'  , 8 ,'\t'  ,	 /* scan  8-15 */ 
'Q'  , 'W'  , 'E'  , 'R'  , 'T'  , 'Y'  , 'U'  , 'I'  ,	 /* scan 16-23 */ 
'O'  , 'P'  , '{'  , '}'  , '\r' , NONE  , 'A'  , 'S'  , /* scan 24-31 */ 
'D'  , 'F'  , 'G'  , 'H'  , 'J'  , 'K'  , 'L'  , ':'  ,	 /* scan 32-39 */ 
'"'  , '~'  , NONE  , '|'  , 'Z'  , 'X'  , 'C'  , 'V'  , /* scan 40-47 */ 
'B'  , 'N'  , 'M'  , '<'  , '>'  , '?'  , NONE  ,   '*', /* scan 48-55 */ 
NONE  , ' '  , NONE,     0,     0, ' '  ,     0,     0,	 /* scan 56-63 */ 
    0,     0,     0,     0,     0, NONE, NONE,   '7',	 /* scan 64-71 */ 
  '8',   '9',   '-',   '4',   '5',   '6',   '+',   '1',	 /* scan 72-79 */ 
  '2',   '3',   '0',   '.',     0,     0,     0,     0,	 /* scan 80-87 */ 
0,0,0,0,0,0,0,0, 
0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0, 
0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,	} ; 
 
static uchar key_ctrl[] = {	
0,     033  , '!'  , 000  , '#'  , '$'  , '%'  , 036  ,	 /* scan  0- 7 */ 
'&'  , '*'  , '('  , ')'  , 037  , '+'  , 0177  , 034,	 /* scan  8-15 */ 
021  , 027  , 005  , 022  , 024  , 031  , 025  , 011  ,	 /* scan 16-23 */ 
017  , 020  , 033  , 035  , '\r' , NONE  , 001  , 023 ,  /* scan 24-31 */ 
004  , 006  , 007  , 010  , 012  , 013  , 014  , ';'  ,	 /* scan 32-39 */ 
'\'' , '`'  , NONE  , 034  , 032  , 030  , 003  , 026  , /* scan 40-47 */ 
002  , 016  , 015  , '<'  , '>'  , '?'  , NONE  ,   '*', /* scan 48-55 */ 
NONE  , ' '  , NONE,     0,     0, ' '  ,     0,     0,	 /* scan 56-63 */ 
NONE,     0,     0,     0,     0,     0,     0,     0,	 /* scan 64-71 */ 
    0,     0,     0,     0,     0,     0,     0,     0,	 /* scan 72-79 */ 
    0,     0,     0,     0,     0,     0,     0,     0,	 /* scan 80-87 */ 
    0,     0,   033, '7'  , '4'  , '1'  ,     0, NONE,	 /* scan 88-95 */ 
'8'  , '5'  , '2'  ,     0, NONE, '9'  , '6'  , '3'  ,	 /*scan  96-103*/ 
'.'  ,     0, '*'  , '-'  , '+'  ,     0,     0,     0,	 /*scan 104-111*/ 
0,0,0,0,0,0,0,0, 0,0,0,0,0,0,0,0,	} ; 


static ulong keymap[16];
static uchar kb_leds = 0;
static int kb_data = -1;

static void key_press(int key)
{
	keymap[key>>4] |= (1 << (key&0xf));
}

static void key_release(int key)
{
	keymap[key>>4] &= ~(1 << (key&0xf));
}

static boolean key_pressed(int key1, int key2)
{
ulong v1,v2;
	v1 = keymap[key1>>4] & (1 << (key1&0xf));
	v2 = keymap[key2>>4] & (1 << (key2&0xf));
	return((v1|v2) > 0);
}


static __inline__ void kb_wait(void)
{
	while(in_portb(KB_STATUS) & 1)
		continue;
}


static void kb_setrate(uchar rate, uchar delay)
{
	kb_wait();
	kb_data = (delay<<5)|rate;
	out_portb(0xf3,KB_DATA);
}

static void kb_setleds(uchar leds)
{
	kb_wait();
	kb_data = leds;
	out_portb(0xed,KB_DATA);
}

static void kb_setled(uchar led)
{
	kb_leds |= (1 << led);
	kb_setleds(kb_leds);
}

static void kb_clearled(uchar led)
{
	kb_leds &= ~(1 << led);
	kb_setleds(kb_leds);
}


static void key_init(void)
{
int i;
	for(i=0; i<16; i++)
		keymap[i] = 0;
}


/* note: right shift,ctrl,alt, and all extended (duplicate) keys
 * are preceded by 0xe0 - we don't use this here though
 */

#define	KBD_BRK	0x80 /* key is breaking contact, not making contact */ 
#define KBD_EXT 0xe0 /* extended key sequence */


static void kbd_svc(int param, int irq)
{
static boolean ext = 0;
uchar ch;
uchar data;
boolean alt,ctrl,shift;
uchar scancode;
	if (in_portb(KB_STATUS)&KB_STAT_RDY) 
		return;
	scancode = in_portb(KB_DATA);
	if(scancode == KBD_EXT) {
		ext = TRUE;
		return;
	}
	if(scancode == 0xfa) {
		if(kb_data >= 0)
			out_portb(kb_data,KB_DATA);
		kb_data = -1;
		return;
	}
	if(scancode == 0xfe) {
		kb_wait();
		out_portb(0xf4,KB_DATA);
		kb_data = -1;
		return;
	}	
	data = scancode & 0x7f;
	if(ext) {
		data |= 0x80;
		ext = FALSE;
	}
	if(scancode & KBD_BRK) {
		key_release(data);
		return;
	}
	if(data == KEY_CAPS_LOCK) {
		if(key_pressed(KEY_CAPS_LOCK,KEY_NONE))
			return;
		key_press(KEY_CAPS_LOCK);
		if(key_pressed(KEY_CAPS_LOCKED,KEY_NONE)) {
			key_release(KEY_CAPS_LOCKED);
			kb_clearled(2);
		} else {
			key_press(KEY_CAPS_LOCKED);
			kb_setled(2);
		}
		return;
	}
	if(data == KEY_NUM_LOCK) {
		if(key_pressed(KEY_NUM_LOCK,KEY_NONE))
			return;
		key_press(KEY_NUM_LOCK);
		if(key_pressed(KEY_NUM_LOCKED,KEY_NONE)) {
			key_release(KEY_NUM_LOCKED);
			kb_clearled(1);
		} else {
			key_press(KEY_NUM_LOCKED);
			kb_setled(1);
		}
		heartbeat_toggle();
		return;
	}
	key_press(data);	
	ch = NONE;
	alt = key_pressed(KEY_LALT, KEY_RALT);
	ctrl = key_pressed(KEY_LCTRL, KEY_RCTRL);
	shift = key_pressed(KEY_LSHIFT, KEY_RSHIFT);
	if(ctrl && alt) {
		switch(data) {
		case KEY_DELETE:
			_(OBJ_SYS, _reboot, ___, ___, ___);
			break;
		case KEY_INSERT:
			_(OBJ_SYS, _halt, ___, ___, ___);
			break;
		}
		return;
	}
	if(alt) {
		if(data >= KEY_F1 && data < KEY_F1+NVCT)
			vct_switch(data-KEY_F1);
		return;
	}
	if(ctrl)
		ch = key_ctrl[scancode];
	else if (shift)
		ch = key_shift[scancode];
	else {
		ch = key_unshift[scancode];
		if (key_pressed(KEY_CAPS_LOCKED, KEY_NONE)
		 && (ch>='a' && ch<='z'))
			ch -= 'a'-'A';
	}
	if(ch == NONE)
		return; 
	keybuf[key_in] = ch;
	key_in = (key_in+1)%KEYBUFSIZE;
} 


