/*  segment.h                                                   (CRC:3a99295a)
 *      relative path: ./I386
 *      creation date: 1992
 *      last modified: Wed Dec 18 07:16:07 1996
 *      header update: Wed Dec 18 07:16:08 1996
 *  
 *       Intel 386 PC  machine dependent code
 *  
 *  Copyright (c) 1992 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is a part of the Shag/OS operating system distribution, and is
 *  free software for non-profit uses; it may be freely used and distributed
 *  for non-profit purposes under the terms of the Shag/OS License, version 1.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND, 
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 *  AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT WILL THE AUTHOR, OR
 *  ANYONE RESPONSIBLE FOR THE LEGAL MODIFICATION OR REDISTRIBUTION OF THIS
 *  SOFTWARE, BE LIABLE FOR ANY DAMAGES, INCLUDING BUT NOT LIMITED TO ANY LOST
 *  PROFITS, LOST SAVINGS OR ANY INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER
 *  RESULTING FROM IMPAIRED OR LOST DATA, SOFTWARE OR COMPUTER FAILURE OR ANY
 *  OTHER CAUSE, EVEN IF THE AUTHOR IS ADVISED OF THE POSSIBILITY OF SUCH
 *  DAMAGES, OR FOR ANY OTHER CLAIM BY THE USER OR FOR ANY THIRD PARTY CLAIM.
 *  
 *  You should have received a copy of the Shag/OS License along with this
 *  program, in a file called 'license1.txt'.  If this license is missing,
 *  please contact the author and describe how this software was obtained.
 *  
 *  The author can be reached for questions, comments, and suggestions at:
 *  
 *      Frank Barrus, 4435 River Rd. Apt #3, Scottsville, NY 14546, USA
 *      email:   shaggy@csh.rit.edu
 *  
 *      For sending bug reports, please read the file 'bugs', and always
 *      include the product name and version, and the word 'BUG' in the subject.
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

#ifndef __SEGMENT_H__
#define __SEGMENT_H__


#ifndef ASM

typedef struct {
	ulong	lo;
	ulong	hi;
} IDTentry;

typedef struct {
	ulong	lo;
	ulong	hi;
} GDTentry;

#endif



#define GDT_NULL __extension__ (GDTentry){0,0} 
#define GDT_SEG(base,limit,type) __extension__(GDTentry){\
		((limit)&0xffff)|(((base)&0xffff) << 16),\
		(type)|(((base)>>16)&0xff)|((limit)&0xf0000)| \
		((base)&0xff000000)|0xc00000}
#define GDT_GATE(seg,ofs,type,stk) __extension__(GDTentry){\
		((seg)<<16)|((ofs)&0xffff),\
		((ofs)&0xffff0000)|(type)|(stk)|0x0c00}	
#define GDT_TSS(base,limit) __extension__(GDTentry){\
		((limit)&0xffff)|(((base)&0xffff) << 16),\
		(((base)>>16)&0xff)|((limit)&0xf0000)| \
		 ((base)&0xff000000)|0x8900}
#define GDT_SET_TASK(tss) __extension__(GDTentry){\
		((tss)<<16),\
		0x8500}
#define IDT(seg,ofs,type) __extension__(IDTentry){\
		((seg)<<16)|((ofs)&0xffff),\
		((ofs)&0xffff0000)|(type)|0x0e00}	

#define GDT_ACC_ISSET(gdt)	(((gdt).hi&0x0100)>0)
#define GDT_ACC_CLEAR(gdt)	(gdt).hi = ((gdt).hi & 0xfeff);


#define IDT_SIZE	48
#define GDT_SIZE	12


/* the following segment descriptors are set up in vmem.c,
 * except for GDT_THREAD which is in mdthread.c:
 */
#define GDT_INIT_CODE	1
#define GDT_INIT_DATA	2
#define GDT_KERN_CODE	3
#define GDT_KERN_DATA	4
#define GDT_USER_CODE	5
#define GDT_USER_DATA	6
#define GDT_THREAD	7
#define GDT_OBJCALL	8
#define GDT_LABEL	9
#define GDT_OBJSTRING	10
#define SEG_INIT_CODE	0x08
#define SEG_INIT_DATA	0x10
#define SEG_KERN_CODE	0x18
#define SEG_KERN_DATA	0x20
#define SEG_USER_CODE	0x28
#define SEG_USER_DATA	0x30
#define SEG_THREAD	0x38
#define SEG_OBJCALL	0x40
#define SEG_LABEL	0x48
#define SEG_OBJSTRING	0x50

#define RPL_KERNEL	0	/* request kernel mode */
#define RPL_USER	3	/* request user mode */


#define SEGTYPE_DR	0x0000	/* data read-only */
#define SEGTYPE_DRW	0x0200	/* data read/write */
#define SEGTYPE_DR_R	0x0400	/* data read-only reversed direction */
#define SEGTYPE_DRW_R	0x0600	/* data read/write reversed direction */
#define SEGTYPE_CX	0x0800	/* code execute only */
#define SEGTYPE_CXR	0x0a00	/* code execute/read */
#define SEGTYPE_CXC	0x0c00	/* code execute only - conforming */
#define SEGTYPE_CXRC	0x0e00	/* code execute/read - conforming */

#define SEGTYPE_USER	0xe000	
#define SEGTYPE_KERNEL	0x8000

#define SEGTYPE_APP	0x1000
#define SEGTYPE_SYS	0x0000

#define SEGLIMIT_MAX	0xfffff /* maximum segment limit */


#endif  /* __SEGMENT_H__ */

