/*  $Id: vmem.c,v 1.1 1996/11/07 11:59:32 shaggy Exp shaggy $
 *      virtual memory management 
 *      
 *      relative path: ./I386
 *  
 *       Intel 386 PC  machine dependent code
 *  
 *  Copyright (c) 1992,1996 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is a part of the Shag/OS operating system distribution, and is
 *  free software for non-profit uses; it may be freely used and distributed
 *  for non-profit purposes under the terms of the Shag/OS License, version 1.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND, 
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 *  AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT WILL THE AUTHOR, OR
 *  ANYONE RESPONSIBLE FOR THE LEGAL MODIFICATION OR REDISTRIBUTION OF THIS
 *  SOFTWARE, BE LIABLE FOR ANY DAMAGES, INCLUDING BUT NOT LIMITED TO ANY LOST
 *  PROFITS, LOST SAVINGS OR ANY INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER
 *  RESULTING FROM IMPAIRED OR LOST DATA, SOFTWARE OR COMPUTER FAILURE OR ANY
 *  OTHER CAUSE, EVEN IF THE AUTHOR IS ADVISED OF THE POSSIBILITY OF SUCH
 *  DAMAGES, OR FOR ANY OTHER CLAIM BY THE USER OR FOR ANY THIRD PARTY CLAIM.
 *  
 *  You should have received a copy of the Shag/OS License along with this
 *  program, in a file called 'license1.txt'.  If this license is missing,
 *  please contact the author and describe how this software was obtained.
 *  
 *  The author can be reached for questions, comments, and suggestions at:
 *  
 *      Frank Barrus, 4435 River Rd. Apt #3, Scottsville, NY 14546, USA
 *      email:   shaggy@csh.rit.edu
 *  
 *      For sending bug reports, please read the file 'bugs', and always
 *      include the product name and version, and the word 'BUG' in the subject.
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

/*  $Log: vmem.c,v $
 *  Revision 1.1  1996/11/07  11:59:32  shaggy
 *  Initial revision
 *
 *
 *  Created in 1992
 */


#define KERNEL
#include <sys/types.h>
#include <sys/resource.h>
#include <sys/global.h>
#include <error.h>
#include <stdio.h>
#include <boot/bootparam.h>
#include <boot/libsolo.h>


/* NOTE:  this memory management system 
 *        is limited to a physical size of 
 *        64K pages (256 meg, with 4K pagesize)
 *        in order to conserve space on the tables
 *        that map available memory.
 *
 *        if it is rewritten for larger systems, 
 *        vm_refcount must first be changed to a ulong *
 *        and many other references must be changed appropriately
 */

static ulong vm_npages;
static ulong vm_nfreepages;
static ushort *vm_refcount;
static ushort vm_firstfree;
static lock_t pagelock;	  /* lock for available page freelist */


ulong *kpagedir;	/* master page directory */

extern BootParam *bootparam;	/* boot parameters from SOLO */

#ifdef OLD_BOOT_LOADER
extern void etext(void),edata(void),end(void);	/* address of last readable page */
#endif

/* routines common to 386 and VAX memory mapping: */
#include "../common/vmem_386_VAX.c"

#define EXTMEMBASE	256	/* page # where extended memory starts --
				 * due to "backward compatability", there
				 * is a 384K "hole" between the end of
				 * base memory and the start of extended
				 * memory -- (or a 512K hole for machines
				 * with only 512K base memory)
				 */
/* #define NEATCHIP -- XXX - not for now  - should be configurable */

#define ROMBASE 	(0xc0000/PAGE_SIZE)
#define ROMSIZE		(256*K/PAGE_SIZE)


extern void ObjCall_gate(void);
extern void Label_gate(void);
extern void ObjString_gate(void);

extern void load_gdt(GDTentry *adr, size_t size);

static void vm_ptmap(ulong *pdir, Vmap *vm, vaddr base, ulong size);


GDTentry gdt[GDT_SIZE] = {
	GDT_NULL,
#ifdef OLD_BOOT_LOADER	/* strange remappings used before */
	GDT_SEG(-KERNBASE,SEGLIMIT_MAX,SEGTYPE_CXR|SEGTYPE_APP|SEGTYPE_KERNEL),
	GDT_SEG(-KERNBASE,SEGLIMIT_MAX,SEGTYPE_DRW|SEGTYPE_APP|SEGTYPE_KERNEL),
#else			/* new simple mappings */
	GDT_SEG(0,SEGLIMIT_MAX,SEGTYPE_CXR|SEGTYPE_APP|SEGTYPE_KERNEL),
	GDT_SEG(0,SEGLIMIT_MAX,SEGTYPE_DRW|SEGTYPE_APP|SEGTYPE_KERNEL),
#endif
	GDT_SEG(0,SEGLIMIT_MAX,SEGTYPE_CXR|SEGTYPE_APP|SEGTYPE_KERNEL),
	GDT_SEG(0,SEGLIMIT_MAX,SEGTYPE_DRW|SEGTYPE_APP|SEGTYPE_KERNEL),
	GDT_SEG(0,SEGLIMIT_MAX,SEGTYPE_CXR|SEGTYPE_APP|SEGTYPE_USER),
	GDT_SEG(0,SEGLIMIT_MAX,SEGTYPE_DRW|SEGTYPE_APP|SEGTYPE_USER),
	GDT_NULL,	/* used for task switching */
	GDT_NULL,
	GDT_NULL,
	GDT_NULL,
	GDT_NULL
};


#ifdef OLD_BOOT_LOADER
static vaddr lomem = (vaddr)end;
#else
static vaddr lomem;	/* can't make assumptions about available memory */
#endif

#ifdef NEATCHIP
static boolean neatchip;
static uchar save65,save68,save69;	
#endif


/****************************************************************************/
#ifdef OLD_BOOT_LOADER


void vm_init(void)
{	
int i, n, len;
ulong readadr;
extern vaddr lomem;	/* imported from sys/kmem.c */
ulong nbasepages; /* number of pages of base memory (n<640K) */
ulong nextpages;  /* number of pages of extended memory */
ulong holesize;	  /* size of "hole" -- see note below */
	mem_init();
	LOCK_CLEAR(pagelock);
	gdt[GDT_OBJCALL] = GDT_GATE(SEG_KERN_CODE,
		(vaddr)ObjCall_gate,SEGTYPE_USER,0);
	gdt[GDT_LABEL] = GDT_GATE(SEG_KERN_CODE,
		(vaddr)Label_gate,SEGTYPE_USER,0);
	gdt[GDT_OBJSTRING] = GDT_GATE(SEG_KERN_CODE,
		(vaddr)ObjString_gate,SEGTYPE_USER,0);
	nbasepages = bootparam.memsize/(PAGE_SIZE/K);
	nextpages = bootparam.extmem/(PAGE_SIZE/K);
	holesize = EXTMEMBASE - nbasepages;
	vm_npages = nbasepages + nextpages;
#ifdef NEATCHIP
	/* get an extra 256K from the ununsed "shadow ROM" space */
	out_portb(0x68,0x22); save68=in_portb(0x23); out_portb(0xff,0x23);
	out_portb(0x69,0x22); save69=in_portb(0x23); out_portb(0xff,0x23);
	out_portb(0x65,0x22); save65=in_portb(0x23); out_portb(0x0f,0x23);
	*(ulong*)0xf00c0000 = 0x12345678;
	if(*(ulong*)0xf00c0000 == 0x12345678) {
		vm_npages += ROMSIZE;
		neatchip = TRUE;
	} else
		neatchip = FALSE;
#endif

	/* make a page directory and clear it: */
	lomem = align(lomem,PAGE_SIZE);
	kpagedir = (ulong*)lomem;
	lomem += PAGE_SIZE;
	memset(kpagedir, 0, PAGE_SIZE);

	/* set up page tables for kernel: */
	kernel.data.ptlen = KERNSIZE/PAGE_SIZE;
	kernel.data.pte = (ulong*)lomem;
	lomem += kernel.data.ptlen*sizeof(long);
	kernel.data.vbase = KERNBASE;

	vm_refcount = (ushort*)lomem;
	lomem += (vm_npages+holesize)*sizeof(ushort);
	memset(vm_refcount, 0, (vm_npages+holesize)*sizeof(ushort));

	vm_vmap_init(&kernel.data);
	/* map the kernel at KERNBASE: */
	len = (kernel.data.ptlen*sizeof(long)-1)/PAGE_SIZE+1;
	for(i=0; i<len; i++) {
		n = (((vaddr)kernel.data.pte + PAGE_SIZE*i) & 0xfff000) 
			| VM_URW | VM_USED | VM_VALID;
		kpagedir[i+KERNBASE/(PAGE_SIZE*PAGE_SIZE/sizeof(long))] = n;	
	}

	/* set up free-list for physical pages: */
	vm_nfreepages = vm_npages;
	vm_firstfree = align(lomem-KERNBASE,PAGE_SIZE)/PAGE_SIZE;
	n = vm_firstfree;
	for(i=1; i<nbasepages-vm_firstfree; i++) { 
		vm_refcount[n] = i;
		n = i;
	}
	if(nextpages) {
		for(i=EXTMEMBASE; i<EXTMEMBASE+vm_npages; i++) {
			vm_refcount[n] = i;
			n = i;
		}
	} 
#ifdef NEATCHIP
	if(neatchip)
		for(i=ROMBASE; i<ROMBASE+ROMSIZE; i++) {
			vm_refcount[n] = i;
			n = i;
		}
#endif	
	vm_refcount[n] = 0;

	/* actually map the kernel, with first segment read only */
	readadr = ((ulong)edata & ~(PAGE_SIZE-1));
	vm_map(&kernel.data, 0, lomem-KERNBASE, VM_KRW);
	/* make the code portion of it readable to user programs: */
	vm_setaccess(&kernel.data, KERNBASE, readadr-KERNBASE, VM_URKW);
	SETPAGEDIR(((vaddr)kpagedir) - KERNBASE);
}


void vm_restore()
{
#ifdef NEATCHIP
	if(neatchip) {
		out_portb(0x68,0x22); out_portb(save68,0x23);
		out_portb(0x69,0x22); out_portb(save69,0x23);
		out_portb(0x65,0x22); out_portb(save65,0x23);
	}
#endif
	kpagedir[0] = kpagedir[KERNBASE/(PAGE_SIZE*NPTEPAGE)];
	SETPAGEDIR((vaddr)kpagedir-KERNBASE);
}



/****************************************************************************/
#else	/* new boot loader (SOLO) */


/* This only works when the lower 1 Meg of memory is identity mapped
 * to virtual memory, which is the state of memory when SOLO 
 * initially loads the kernel, until remapping occurs
 * Previously, we could assume that a virtual address could be 
 * converted to a physical address by simply masking out the high 8 bits,
 * because of the simple linear mapping, but now we can make no
 * such assumptions, so each virtual address must be explicitly looked up.
 * (actually, within each segment (code/data/bss/etc...) there is
 * still a linear mapping, but it's simple enough just to translate 
 * every address to be safe, and besides, this is only done once
 * at initialization time, so efficiency issues are moot)
 */


static paddr boot_get_phys(vaddr va)
{
	paddr pt;
	vaddr base;
	
	base = va & ~(PAGE_SIZE*NPTEPAGE-1);
	pt = kpagedir[base/(PAGE_SIZE*NPTEPAGE)] & ~(PAGE_SIZE-1);
	return (*(ulong*)(pt+(va-base)/PAGE_SIZE*sizeof(pte_t))
			& ~(PAGE_SIZE-1)) +(va%PAGE_SIZE);
}


static void* vm_MapLow4MB(void *new_map)
{
	ulong *pagedir;
	ulong old_map;

	if(thread)
		pagedir = (ulong*)thread_context->pagedir;
	else
		pagedir = kpagedir;
	old_map = pagedir[0];
	pagedir[0] = (ulong)new_map;
	TBIR(0,4*MEG);
	return (void*)old_map;
}

static void verify_mapping(const char *name, vaddr addr, vaddr expected)
{
	if(addr != expected) {
		panic(0, "%s: %x != %x", name, addr, expected);
		solo_exit(1);
	}
}


void vm_init(void)
{
	ulong basepagehi; /* highest base page number + 1 */
	ulong extpagelo;  /* lowest extended page number */
	ulong extpagehi;  /* highest extended page number + 1 */
	int i, n;
	ulong pdbr;
	vaddr lomem_start;

	/* just to be safe... */
	thread = NULL;

	/* since there's already a kernel page directory,
	 * we might as well use it... 
	 */
	__asm__("mov %%cr3,%0":"r="(pdbr));
	kpagedir = (ulong*)(pdbr & ~(PAGE_SIZE-1));

	lomem = lomem_start = align(bootparam->lomem, PAGE_SIZE);
	basepagehi = bootparam->base_himem/PAGE_SIZE;
	if(bootparam->ext_size) {
		extpagelo = bootparam->ext_addr/PAGE_SIZE;
		extpagehi = extpagelo + bootparam->ext_size/PAGE_SIZE;
	} else
		extpagelo = extpagehi = 0;

	mem_init();
	LOCK_CLEAR(pagelock);
	gdt[GDT_OBJCALL] = GDT_GATE(SEG_KERN_CODE,
		(vaddr)ObjCall_gate,SEGTYPE_USER,0);
	gdt[GDT_LABEL] = GDT_GATE(SEG_KERN_CODE,
		(vaddr)Label_gate,SEGTYPE_USER,0);
	gdt[GDT_OBJSTRING] = GDT_GATE(SEG_KERN_CODE,
		(vaddr)ObjString_gate,SEGTYPE_USER,0);

	load_gdt(gdt, sizeof(gdt));

	/* set up page tables for kernel: */
	kernel.data.ptlen = KERNSIZE/PAGE_SIZE;
	kernel.data.pte = (ulong*)lomem;
	lomem += kernel.data.ptlen*sizeof(long);
	kernel.data.vbase = KERNBASE;

	vm_refcount = (ushort*)lomem;
	if(extpagelo)
		vm_npages = extpagehi;
	else
		vm_npages = basepagehi;

	lomem += vm_npages*sizeof(ushort);
	memset(vm_refcount, 0, vm_npages*sizeof(ushort));

	vm_vmap_init(&kernel.data);

	/* set up free-list for physical pages: */
	lomem = align(lomem, PAGE_SIZE);
	vm_firstfree = boot_get_phys(lomem)/PAGE_SIZE;
	n = vm_firstfree;
	vm_nfreepages = 1;
	for(i=vm_firstfree+1; i<basepagehi; i++) {
		vm_refcount[n] = i;
		vm_nfreepages++;
		n = i;
	}
	if(extpagelo) {
		for(i=extpagelo; i<extpagehi; i++) {
			vm_refcount[n] = i;
			vm_nfreepages++;
			n = i;
		}
	} 
	vm_refcount[n] = 0;

	/* actually map the kernel, with first segment readable 
         * by user programs so they can share code 
         */
	verify_mapping("text", 
		vm_map(&kernel.data, bootparam->text.paddr,
		bootparam->text.vsize, VM_URKW), bootparam->text.vaddr);
	verify_mapping("data",
		vm_map_join(&kernel.data, bootparam->data.paddr,
		bootparam->data.vsize, VM_URKW), bootparam->data.vaddr);
	verify_mapping("bss",
		vm_map_join(&kernel.data,
			align(bootparam->bss.paddr, PAGE_SIZE),
			bootparam->bss.vsize & ~(PAGE_SIZE-1), VM_URKW),
				align(bootparam->bss.vaddr, PAGE_SIZE));
	verify_mapping("kmem",
		vm_map_join(&kernel.data, boot_get_phys(lomem_start),
			lomem-lomem_start, VM_KRW), lomem_start);

	/* map the kernel at KERNBASE: */
	vm_ptmap(kpagedir, &kernel.data, KERNBASE, KERNSIZE);

	/* map in the page directory we're using */
	kpagedir = (ulong*)vm_map(&kernel.data, (paddr)kpagedir, PAGE_SIZE, VM_KRW);

	/* Set up a handler for BootParam mapping */
	solo_set_MapBootParam(vm_MapLow4MB);
	solo_set_UnmapBootParam(vm_MapLow4MB);

	/* unmap the low 1MB of memory (BIOS, BootParam, SOLO, etc) */
	solo_UnmapBootParam(0);
}


void vm_restore()
{
	solo_MapBootParam();
}

#endif	/*  new boot loader */

/****************************************************************************/



vaddr vm_pagedir_create(void)
{
vaddr pdir;
	pdir = vm_alloc(&kernel.data,PAGE_SIZE,VM_KRW);
	memset((void*)pdir, 0, PAGE_SIZE);
	vm_ptmap((ulong*)pdir, &kernel.data, KERNBASE, KERNSIZE);
	return(pdir);
}


void vm_pagedir_free(vaddr pdir)
{
	vm_free(&kernel.data, pdir);
}



static void vm_ptmap(ulong *pdir, Vmap *vm, vaddr base, ulong size)
{
int i, n;
int npages;
vaddr pte;
	n = base/(PAGE_SIZE*NPTEPAGE);
	if(vm) {
		npages = (vm->ptlen*sizeof(long)-1)/PAGE_SIZE+1;
		pte = (vaddr)vm->pte;
		if(vm->vbase == STACKBASE) {
			pte += ((MAXPTE-npages*NPTEPAGE)*4);
			i = n+((MAXPTE-npages)/(PAGE_SIZE/sizeof(long)));
			while(n < i) 
				pdir[n++] = VM_NA;
		}
		for(i=0; i<npages; i++) {
			pdir[n++] = vm_getphys(pte) | VM_URW | VM_USED | VM_VALID;
			pte += PAGE_SIZE;
		}
	} else 
		npages = 0;
	i = (base+size)/(PAGE_SIZE*NPTEPAGE);
	while(n < i)
		pdir[n++] = VM_NA;
	TBIA();
	/* XXX - should use segment sizes or something to enforce 'ptlen'	
	 * limits correctly  - otherwise tasks can access each other
	 */
}

void vm_task_ptmap(Thread *node, Vmap *vmap)
{
	vm_ptmap((ulong*)node->context.pagedir, vmap, TASKBASE, TASKSIZE);
}

void vm_stack_temp_ptmap(Thread *node, Vmap *vm, Vmap *vmtemp)
{
	vm_ptmap((ulong*)node->context.pagedir, vm, STACKBASE, STACKSIZE);
	vm_ptmap((ulong*)node->context.pagedir, vmtemp, TEMPBASE, TEMPSIZE);
}

void vm_kern_ptmap(Thread *node, Vmap *vmap)
{
	vm_ptmap((ulong*)node->context.pagedir, vmap, KERNBASE, KERNSIZE);
}

