/*  vmem.h                                                      (CRC:1653f05b)
 *      relative path: ./I386
 *      creation date: 1992
 *      last modified: Wed Dec 18 07:16:11 1996
 *      header update: Wed Dec 18 07:16:12 1996
 *  
 *       Intel 386 PC  machine dependent code
 *  
 *  Copyright (c) 1992 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is a part of the Shag/OS operating system distribution, and is
 *  free software for non-profit uses; it may be freely used and distributed
 *  for non-profit purposes under the terms of the Shag/OS License, version 1.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND, 
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 *  AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT WILL THE AUTHOR, OR
 *  ANYONE RESPONSIBLE FOR THE LEGAL MODIFICATION OR REDISTRIBUTION OF THIS
 *  SOFTWARE, BE LIABLE FOR ANY DAMAGES, INCLUDING BUT NOT LIMITED TO ANY LOST
 *  PROFITS, LOST SAVINGS OR ANY INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER
 *  RESULTING FROM IMPAIRED OR LOST DATA, SOFTWARE OR COMPUTER FAILURE OR ANY
 *  OTHER CAUSE, EVEN IF THE AUTHOR IS ADVISED OF THE POSSIBILITY OF SUCH
 *  DAMAGES, OR FOR ANY OTHER CLAIM BY THE USER OR FOR ANY THIRD PARTY CLAIM.
 *  
 *  You should have received a copy of the Shag/OS License along with this
 *  program, in a file called 'license1.txt'.  If this license is missing,
 *  please contact the author and describe how this software was obtained.
 *  
 *  The author can be reached for questions, comments, and suggestions at:
 *  
 *      Frank Barrus, 4435 River Rd. Apt #3, Scottsville, NY 14546, USA
 *      email:   shaggy@csh.rit.edu
 *  
 *      For sending bug reports, please read the file 'bugs', and always
 *      include the product name and version, and the word 'BUG' in the subject.
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

#ifndef __VMEM_H__
#define __VMEM_H__


#ifndef ASM

#include <sys/types.h>
#include <object.h>

typedef ulong	pte_t;


#define VMSEG_SHARED	0x01	/* shared memory */
#define VMSEG_OBJREAD	0x02	/* demand paging */
#define VMSEG_OBJWRITE	0x04	/* demand paged writeback */

typedef struct vmseg {
	ulong	start;	
	ushort	size;
	ushort	type;
	struct vmap *vmshare;
	union {
		struct vmap *vmap;
		obj_t	obj;
	} src;
	ulong	index;
	struct vmseg *nextseg;
	struct vmseg *nextshare;
} VmSeg;

typedef struct vmap {
	pte_t	*pte;		/* pointer to page table */
	long	ptlen;		/* length of page table */
	vaddr	vbase;		/* mapped address */
	VmSeg	*seg;		/* segment map */
	VmSeg	*share;		/* shared segment map */
	lock_t	lock;		/* vmap lock */
} Vmap;

#endif


/* Page Table Entry (PTE) access modes: */
#define VM_ACC		0x0006		/* bitmask for access mode */
#define VM_NA		0x0000		/* not available */
#define VM_URW		0x0006		/* user read/write */
#define VM_UR		0x0004		/* user read only */
#define VM_URKW		0x0004		/* user read, kernel read/write */
#define VM_KRW		0x0002		/* kernel read/write */
#define VM_KR		0x0000		/* kernel read  --- but thanks to
					   Intel, supervisor mode can
					   still write anyways, and this
					   bit has no effect... 
					   Thanks, Intel....    :-)  */
					/* XXX - is this true for the 486
					   also? */
/* software values for PTE's:  */
#define VM_SOFT		0x0e00		/* bitmask for software values */
#define VM_FREE		0x0000		/* available virtual space */
#define VM_USED		0x0200		/* generic in-use */
#define VM_LOCKED	0x0400		/* locked-in page */
#define VM_RESERVED	0x0600		/* reserved page (illegal page) */
#define VM_SHARED	0x0800		/* shared page */
#define VM_RESV4	0x0a00
#define VM_RESV5	0x0c00
#define VM_EOS		0x0e00		/* end-of-segment (illegal page) */

/* access flags (for valid pages) or software values for invalid */
#define VM_FLAGS	0x0061		/* bitmask for flags */
/* hardware: valid page */
#define VM_VALID	0x0001		/* valid entry */
#define VM_DIRTY	0x0040		/* dirty page (written to) */
#define VM_ACCESS	0x0020		/* read access to page */
/* software: invalid page */
#define VM_ZERO		0x0000		/* zero out the page */
#define VM_WFILL	0x0020		/* fill with 'page' value words */
#define VM_DEMAND	0x0040		/* demand page (page=kernel object) */
#define VM_COPY		0x0060		/* copy-on-write */


#define PTE_ISDIRTY(pte)	(((pte) & VM_DIRTY) > 0)	/* check dirty bit */
#define PTE_ISREF(pte)		(((pte) & VM_ACCESS) > 0)	/* check reference bit */
#define PTE_ISVALID(pte)	(((pte) & VM_VALID) > 0)	/* check valid bit */
#define PTE_ISFREE(pte)		(((pte) & VM_SOFT) == VM_FREE)	/* check if entry is free */
#define PTE_ISEOS(pte)		(((pte) & VM_SOFT) == VM_EOS)	/* check if entry is end of segment */
#define PTE_SETVALID(pte)	(pte) |= VM_VALID;
#define PTE_SETINVALID(pte)	(pte) &= ~VM_VALID;
#define PTE_GETMODE(pte,mode) 	(mode) = ((pte) & 0xfff);
#define PTE_GETACC(pte,mode)	(mode) = ((pte) & VM_ACC);
#define PTE_GETFLAGS(pte,mode) 	(mode) = ((pte) & VM_FLAGS);
#define PTE_GETSOFT(pte,mode)	(mode) = ((pte) & VM_SOFT);
#define PTE_GETPAGE(pte,page)	(page) = ((pte) >> 12); 
#define PTE_SETMODE(pte,m)	(pte) = ((pte) & ~0xfff) | ((m) & 0xfff);
#define PTE_SETACC(pte,m)	(pte) = ((pte) & ~VM_ACC) | ((m) & VM_ACC);
#define PTE_SETFLAGS(pte,m)	(pte) = ((pte) & ~VM_FLAGS) | ((m) & VM_FLAGS);
#define PTE_SETSOFT(pte,m)	(pte) = ((pte) & ~VM_SOFT) | ((m) & VM_SOFT);
#define PTE_SETPAGE(pte,page)	(pte) = ((pte) & 0xfff) | ((page) << 12);
#define PTE_SETEOS(pte)		(pte) = (VM_EOS|VM_NA);
#define PTE_SETFREE(pte)	(pte) = (VM_FREE|VM_NA);
#define PTE_SET(pte,page,mode)	(pte) = ((page) << 12) | (mode);


/* page-fault error codes macros */
#define VM_FAULT_USER(x)	(((x)&1)>0)
#define VM_FAULT_KERNEL(x)	(((x)&1)==0)
#define VM_FAULT_WRITE(x)	(((x)&2)>0)
#define VM_FAULT_READ(x)	(((x)&2)==0)
#define VM_FAULT_ACCV(x)	(((x)&4)>0)
#define VM_FAULT_NOPAGE(x)	(((x)&4)==0)

#endif /* __VMEM_H__ */

