/*  console_mvax.c                                              (CRC:97daf022)
 *      relative path: ./VAX
 *      creation date: 1992
 *      last modified: Wed May 22 21:28:00 1996
 *      header update: Wed Dec 18 07:16:20 1996
 *  
 *       Digital VAX machine dependent code
 *  
 *  Copyright (c) 1992 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is a part of the Shag/OS operating system distribution, and is
 *  free software for non-profit uses; it may be freely used and distributed
 *  for non-profit purposes under the terms of the Shag/OS License, version 1.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND, 
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 *  AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT WILL THE AUTHOR, OR
 *  ANYONE RESPONSIBLE FOR THE LEGAL MODIFICATION OR REDISTRIBUTION OF THIS
 *  SOFTWARE, BE LIABLE FOR ANY DAMAGES, INCLUDING BUT NOT LIMITED TO ANY LOST
 *  PROFITS, LOST SAVINGS OR ANY INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER
 *  RESULTING FROM IMPAIRED OR LOST DATA, SOFTWARE OR COMPUTER FAILURE OR ANY
 *  OTHER CAUSE, EVEN IF THE AUTHOR IS ADVISED OF THE POSSIBILITY OF SUCH
 *  DAMAGES, OR FOR ANY OTHER CLAIM BY THE USER OR FOR ANY THIRD PARTY CLAIM.
 *  
 *  You should have received a copy of the Shag/OS License along with this
 *  program, in a file called 'license1.txt'.  If this license is missing,
 *  please contact the author and describe how this software was obtained.
 *  
 *  The author can be reached for questions, comments, and suggestions at:
 *  
 *      Frank Barrus, 4435 River Rd. Apt #3, Scottsville, NY 14546, USA
 *      email:   shaggy@csh.rit.edu
 *  
 *      For sending bug reports, please read the file 'bugs', and always
 *      include the product name and version, and the word 'BUG' in the subject.
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

/* console_mvax
 *  console I/O for MicroVAX systems
 */


#include <mvax.h>
#include <cmos.h>

static volatile RSerial *rserial = (RSerial*)0x200a0000;
static volatile RIntr   *rintr = (RIntr*)0x20080000;

#define wserial	((WSerial*)rserial)
#define wintr ((WIntr*)rintr)


#define CONSOLE_AUX 2
#define CONSOLE_CON 3
#define CONSOLE_VID 0

static int console = 3;



int console_write(char *buf, int len)
{
int chan = console;
int line;
ulong tcr;
int t, i;
ulong data;
int ipl;
	i = len;
        while(i--) {
		ipl = set_iplmin(5); 
		if(*buf == '\n')
			console_write("\r",1);
                wintr->int_msk &= ~IMASK_ST;
                tcr = 0;
                t = 30000;
                wserial->ser_tcr |= (1 << chan);
                while(1) {
                        while(((data = rserial->ser_csr) & bit(15)) == 0 && t--)
                                continue;
                        wserial->ser_csr &= ~bit(15);
                        line = (data >> 8) & 3;
                        if (line == chan)
                                break;
                        tcr |= (1 << line);
                        wserial->ser_tcr &= ~(1 << line);
                }
                wserial->ser_tdr = (*buf++) & 0xff;
                while(1) {
                        while(((data=rserial->ser_csr) & bit(15)) == 0 && t--)
                                continue;
                        line = (data >> 8) & 3;
                        if (line == chan)
                                break;
                        tcr |= (1 << line);
                        wserial->ser_tcr &= ~(1 << line);
                }
                wserial->ser_tcr &= ~(1 << line);
                if (!tcr) {
			wintr->int_clr = IMASK_ST;
			wintr->int_msk |= IMASK_ST;
                } else {
			wintr->int_msk |= IMASK_ST;
                        wserial->ser_tcr |= tcr;
                }
        	set_ipl(ipl); 
        }
	return(len);
}


static __inline__ void serial_setline(int port, int rate, int bits, int parity)
{
	wserial->ser_lpr = 0x1000 | port | (rate << 8) | parity
		| ((bits-5) << 3);
}


static void console_hardware_init(void)
{
	if(((nvram[CONSOLE_ID] >> 2) & 0xff) == 1)
		console = CONSOLE_CON;
	else
		console = CONSOLE_VID;
	rserial = (RSerial*)vm_map(&kernel.data, 0x200a0000, 512, VM_KRW);
	rintr = (RIntr*)vm_map(&kernel.data, 0x20080000, 512, VM_KRW);
	wserial->ser_csr = 0x0030;
	while(rserial->ser_csr & 0x0010)
		continue;
	wserial->ser_csr = 0x0020;
	wserial->ser_tcr = DTR_2 | RTS_2;
	serial_setline(0, BR4800, 8, PAR_NONE);
	serial_setline(1, BR4800, 8, PAR_ODD);
	serial_setline(2, BR9600, 7, PAR_EVEN);
	serial_setline(3, BR9600, 8, PAR_NONE);
	wintr->int_clr = IMASK_SR | IMASK_ST;
	wintr->int_msk |= (IMASK_SR | IMASK_ST);
}

char console_readkey(void)
{
int data;
register int chan;
	while((data = rserial->ser_rbuf) & bit(15)) {
		if(data & 0x6000)
			continue;
		chan = (data >> 8) & 0x03;
		data &= 0xff;
		if(chan == console)
			return(data);
	}
	return(0);
}


