/*  interrupt.c                                                 (CRC:7d3d0ce1)
 *      relative path: ./VAX
 *      creation date: 1992
 *      last modified: Wed May 22 21:28:04 1996
 *      header update: Wed Dec 18 07:16:22 1996
 *  
 *       Digital VAX machine dependent code
 *  
 *  Copyright (c) 1992 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is a part of the Shag/OS operating system distribution, and is
 *  free software for non-profit uses; it may be freely used and distributed
 *  for non-profit purposes under the terms of the Shag/OS License, version 1.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND, 
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 *  AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT WILL THE AUTHOR, OR
 *  ANYONE RESPONSIBLE FOR THE LEGAL MODIFICATION OR REDISTRIBUTION OF THIS
 *  SOFTWARE, BE LIABLE FOR ANY DAMAGES, INCLUDING BUT NOT LIMITED TO ANY LOST
 *  PROFITS, LOST SAVINGS OR ANY INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER
 *  RESULTING FROM IMPAIRED OR LOST DATA, SOFTWARE OR COMPUTER FAILURE OR ANY
 *  OTHER CAUSE, EVEN IF THE AUTHOR IS ADVISED OF THE POSSIBILITY OF SUCH
 *  DAMAGES, OR FOR ANY OTHER CLAIM BY THE USER OR FOR ANY THIRD PARTY CLAIM.
 *  
 *  You should have received a copy of the Shag/OS License along with this
 *  program, in a file called 'license1.txt'.  If this license is missing,
 *  please contact the author and describe how this software was obtained.
 *  
 *  The author can be reached for questions, comments, and suggestions at:
 *  
 *      Frank Barrus, 4435 River Rd. Apt #3, Scottsville, NY 14546, USA
 *      email:   shaggy@csh.rit.edu
 *  
 *      For sending bug reports, please read the file 'bugs', and always
 *      include the product name and version, and the word 'BUG' in the subject.
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

/* interrupt.c
 */

#define KERNEL
#include <sys/types.h>
#include <sys/resource.h>
#include <sys/global.h>
#include <stdio.h>
#include <error.h>
#include <privreg.h>

extern void isr_jmp(void);

extern void isr_emulate(void);
extern void isr_em_cont(void);

#define NIVEC 256

ulong int_param[NIVEC];
vaddr int_vec[NIVEC];

ulong *scb;



static __inline__ void set_ivec(vaddr func, int param, int ivec)
{
	int_param[ivec] = param;
	int_vec[ivec] = func;
}


static void trap_svc(ulong type, int ivec, vaddr pc, vaddr va, ulong errcode)
{
	panic(pc,"%s; errcode=%x ivec=%x", trap_msg(type|(errcode&0xff)),
		errcode, ivec);
}


void interrupt_init(void)
{
int i;
paddr scb_phys;
	scb_phys = vm_getphys((vaddr)scb);
	mtpr(scb_phys, R_SCBB);
	for(i=0; i<NIVEC; i++)
		scb[i] = (vaddr)isr_jmp + i*4;	
	scb[0x02]++;		/* use interrupt stack */
	for(i=32; i<50; i++)
		scb[i]++;
	scb[0x32] = (vaddr)isr_emulate;
	scb[0x33] = (vaddr)isr_em_cont;
	for(i=0; i<NIVEC; i++)
		set_interrupt((vaddr)trap_svc, TRAP_INVALID, i);
	set_ivec((vaddr)trap_svc, TRAP_KSTACK , 0x02);
	set_ivec((vaddr)trap_svc, TRAP_PRIVINST, 0x04);
	set_ivec((vaddr)trap_svc, TRAP_RESVOP, 0x05);
	set_ivec((vaddr)trap_svc, TRAP_RESVOP, 0x06);
	set_ivec((vaddr)trap_svc, TRAP_RESVADR, 0x07);
	set_ivec((vaddr)trap_svc, TRAP_TRACE, 0x0a);
	set_ivec((vaddr)trap_svc, TRAP_BKPT, 0x0b);
	set_ivec((vaddr)trap_svc, TRAP_ARITH , 0x0d);
	set_ivec((vaddr)vm_accv, 0, 0x08);
	set_ivec((vaddr)vm_pagefault, 0, 0x09);
	int_level = 0;
}


void set_interrupt(vaddr func, int param, int ivec)
{
	int_param[ivec] = param;
	int_vec[ivec] = func;
}

