/*  mdserial.c                                                  (CRC:308ab9f8)
 *      relative path: ./VAX
 *      creation date: 1992
 *      last modified: Wed May 22 21:28:10 1996
 *      header update: Wed Dec 18 07:16:31 1996
 *  
 *       Digital VAX machine dependent code
 *  
 *  Copyright (c) 1992 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is a part of the Shag/OS operating system distribution, and is
 *  free software for non-profit uses; it may be freely used and distributed
 *  for non-profit purposes under the terms of the Shag/OS License, version 1.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND, 
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 *  AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT WILL THE AUTHOR, OR
 *  ANYONE RESPONSIBLE FOR THE LEGAL MODIFICATION OR REDISTRIBUTION OF THIS
 *  SOFTWARE, BE LIABLE FOR ANY DAMAGES, INCLUDING BUT NOT LIMITED TO ANY LOST
 *  PROFITS, LOST SAVINGS OR ANY INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER
 *  RESULTING FROM IMPAIRED OR LOST DATA, SOFTWARE OR COMPUTER FAILURE OR ANY
 *  OTHER CAUSE, EVEN IF THE AUTHOR IS ADVISED OF THE POSSIBILITY OF SUCH
 *  DAMAGES, OR FOR ANY OTHER CLAIM BY THE USER OR FOR ANY THIRD PARTY CLAIM.
 *  
 *  You should have received a copy of the Shag/OS License along with this
 *  program, in a file called 'license1.txt'.  If this license is missing,
 *  please contact the author and describe how this software was obtained.
 *  
 *  The author can be reached for questions, comments, and suggestions at:
 *  
 *      Frank Barrus, 4435 River Rd. Apt #3, Scottsville, NY 14546, USA
 *      email:   shaggy@csh.rit.edu
 *  
 *      For sending bug reports, please read the file 'bugs', and always
 *      include the product name and version, and the word 'BUG' in the subject.
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

/* mdserial.c
 *  Serial I/O for microVAXen - to be included from drv/serial.c
 */


#include <mvax.h>


#define NSERIAL 4

char *devname[NSERIAL] = {"mouse","kb","ttys0","ttys1"};
int devspeed[NSERIAL] = {4800, 4800, 9600, 9600};
char *devparam[NSERIAL] = {"n81","o81","e71","n81"};

#define IS_TTY(n) 	((n)>=2)


obj_t ser_tty[NSERIAL];
SerialBuf *ser_idata[NSERIAL];
SerialBuf *ser_odata[NSERIAL];

static RSerial *rserial;
static RIntr   *rintr;

#define wserial ((WSerial*)rserial)
#define wintr ((WIntr*)rintr)


extern label_t _param, _data, _speed, _break, _msleep;
extern label_t _halt, _reboot;


static status_t ser_write(int port)
{
	wserial->ser_tcr |= bit(port);
	return(OK);
}


static void serial_read_svc(int param, int ivec)
{
register int chan;
int data;
static tilde = FALSE;
	wintr->int_clr = IMASK_SR;
	while((data = rserial->ser_rbuf) & bit(15)) {
		if(data & 0x6000)
			continue;
		chan = (data >> 8) & 0x03;
		data &= 0xff;
		serbuf_writechar(ser_idata[chan],data);
		if(chan==3) {
			if(tilde && data=='h')
				_(OBJ_SYS, _halt, ___, ___, ___);
			if(tilde && data=='b')
				_(OBJ_SYS, _reboot, ___, ___, ___);
			tilde = (data=='~');
		}
	}	
}


static void serial_write_svc(int param, int ivec)
{
register int chan;
int data;
	wintr->int_clr = IMASK_ST;
	while((chan = rserial->ser_csr) & bit(15)) {
		chan = (chan >> 8) & 0x03;
		if((data=serbuf_readchar(ser_odata[chan])) < 0) {
			wserial->ser_tcr &= ~(1 << chan); /* no xmit */ 
			break;
		}
		wserial->ser_tdr = data;
	}
}


static void serial_setline(int port, int speed, char *param)
{
int rate, pcode;
	if(speed == 300)
		rate = BR300;
	else if(speed == 1200)
		rate = BR1200;
	else if(speed == 2400)
		rate = BR2400;
	else if(speed == 4800)
		rate = BR4800;
	else if(speed == 9600)
		rate = BR9600;
	else if(speed == 19200)
		rate = BR19200;
	else	
		rate = BR9600;
	switch(param[0]) {
	case 'e': case 'E':
		pcode = PAR_EVEN; break;
	case 'o': case 'O':
		pcode = PAR_ODD; break;
	default:
		pcode = PAR_NONE;
	}
	wserial->ser_lpr = 0x1000 | port | (rate << 8) | pcode
		| ((param[1]-'0'-5) << 3);
}


static obj_t serial_break(ObjFrame *frame, int port)
{
	wserial->ser_tdr = bit(8+port);
	_(OBJ_THREAD, _msleep, 1000, ___, ___);
	wserial->ser_tdr = 0;
	return(OK);
}


static obj_t serial_hangup(ObjFrame *frame, int port)
{
	if(port != 2)
		return(ERR_NOMETHOD);
	wserial->ser_tcr = 0;
	_(OBJ_THREAD, _msleep, 1000, ___, ___);
	wserial->ser_tcr = DTR_2 | RTS_2;
	return(OK);
}


static __inline__ void mdserial_init(void)
{
	rserial = (RSerial*)_(OBJ_TASK, _vm_map, 0x200a0000, 512, ___);
        rintr = (RIntr*)_(OBJ_TASK, _vm_map, 0x20080000, 512, ___);
        wserial->ser_csr = 0x0030;
        while(rserial->ser_csr & 0x0010)
                continue;
        wserial->ser_csr = 0x0020;
        wserial->ser_tcr = DTR_2 | RTS_2;
        wintr->int_clr = IMASK_SR | IMASK_ST;
        wintr->int_msk |= (IMASK_SR | IMASK_ST);
	set_interrupt((vaddr)serial_read_svc, 0, INT_RSERIAL);
	set_interrupt((vaddr)serial_write_svc, 0, INT_WSERIAL);
}


static __inline__ boolean serial_init_port(int i, char *msg)
{
	*msg = '\0';
	return(TRUE);
}

