/*  vid_qdss.c                                                  (CRC:5514722a)
 *      relative path: ./VAX
 *      creation date: 1992
 *      last modified: Wed May 22 21:28:20 1996
 *      header update: Wed Dec 18 07:16:48 1996
 *  
 *       Digital VAX machine dependent code
 *  
 *  Copyright (c) 1992 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is a part of the Shag/OS operating system distribution, and is
 *  free software for non-profit uses; it may be freely used and distributed
 *  for non-profit purposes under the terms of the Shag/OS License, version 1.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND, 
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 *  AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT WILL THE AUTHOR, OR
 *  ANYONE RESPONSIBLE FOR THE LEGAL MODIFICATION OR REDISTRIBUTION OF THIS
 *  SOFTWARE, BE LIABLE FOR ANY DAMAGES, INCLUDING BUT NOT LIMITED TO ANY LOST
 *  PROFITS, LOST SAVINGS OR ANY INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER
 *  RESULTING FROM IMPAIRED OR LOST DATA, SOFTWARE OR COMPUTER FAILURE OR ANY
 *  OTHER CAUSE, EVEN IF THE AUTHOR IS ADVISED OF THE POSSIBILITY OF SUCH
 *  DAMAGES, OR FOR ANY OTHER CLAIM BY THE USER OR FOR ANY THIRD PARTY CLAIM.
 *  
 *  You should have received a copy of the Shag/OS License along with this
 *  program, in a file called 'license1.txt'.  If this license is missing,
 *  please contact the author and describe how this software was obtained.
 *  
 *  The author can be reached for questions, comments, and suggestions at:
 *  
 *      Frank Barrus, 4435 River Rd. Apt #3, Scottsville, NY 14546, USA
 *      email:   shaggy@csh.rit.edu
 *  
 *      For sending bug reports, please read the file 'bugs', and always
 *      include the product name and version, and the word 'BUG' in the subject.
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

/* vid_qdss.c
 *  AVSS/QDSS video subsystems
 *  (KIMS, etc)
 */


static Adder *adder;
static Cursor *cursor;
static KimsReg *kimsreg;
static ScrollFill *sc_fill;
static ScrollStart *sc_start;
static ScrollStart *sc_start_b;
static ScrollEnd *sc_end;
static ScrollEnd *sc_end_b;

static struct vt100_state vtstate;


#define VROM_FONT_INDEX	0x00b4
#define PROM_FONT_VECTOR 0x001d

void video_fill(), video_setclip();
int video_write();


ushort block_data[16] = {
	0x003f,0x003f,0x003f,0x003f,0x003f,0x003f,0x003f,0x003f,
	0x003f,0x003f,0x003f,0x003f,0x003f,0x003f,0x003f,0x003f};
ushort block_mask[16] = {
	0,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0};



Font font[2];
Font *console_font;
int console_hgt, console_wid;


#define VT_SET_CURSOR(vt,x,y) \
	set_cursor_xy((x)*console_font->wid,(y)*console_font->hgt+6);
#define VT_PUTCHAR(vt,x,y,ch) \
	video_putchar(console_font, \
		(x)*console_font->wid,(y)*console_font->hgt, (ch));
#define VT_SCROLL_UP(vt,x1,y1,x2,y2,n) \
	video_scroll_up((x1)*console_font->wid,(y1)*console_font->hgt, \
		((x2)-(x1)+1)*console_font->wid,\
		((y2)-(y1)+1)*console_font->hgt,\
		console_font->hgt*(n));
#define VT_SCROLL_DOWN(vt,x1,y1,x2,y2,n) \
	video_scroll_up((x1)*console_font->wid,(y1)*console_font->hgt, \
		((x2)-(x1)+1)*console_font->wid,\
		((y2)-(y1)+1)*console_font->hgt,\
		console_font->hgt*(n));
#define VT_SCROLL_RIGHT(vt,x1,y1,x2,y2,n) \
	video_scroll_up((x1)*console_font->wid,(y1)*console_font->hgt, \
		((x2)-(x1)+1)*console_font->wid,\
		((y2)-(y1)+1)*console_font->hgt,\
		console_font->hgt*(n));
#define VT_SCROLL_LEFT(vt,x1,y1,x2,y2,n) \
	video_scroll_up((x1)*console_font->wid,(y1)*console_font->hgt, \
		((x2)-(x1)+1)*console_font->wid,\
		((y2)-(y1)+1)*console_font->hgt,\
		console_font->hgt*(n));
#define VT_CLEAR(vt,x1,y1,x2,y2) \
	video_fill((x1)*console_font->wid,(y1)*console_font->hgt, \
		((x2)-(x1)+1)*console_font->wid,\
		((y2)-(y1)+1)*console_font->hgt,\
		0x0000);
#define VT_SET_COLOR(vt)	;
#define VT_BEEP(vt)		video_beep();


init_video()
{
	adder = (Adder*)map_device("adder",0x3c000000,sizeof(Adder));
	cursor = (Cursor*)map_device("cursor",0x3c000400,sizeof(Cursor));
	kimsreg = (KimsReg*)map_device("kimsreg",0x38000000,sizeof(KimsReg));
	sc_fill = (ScrollFill*)map_device("sc_fill",
		0x38008000,sizeof(ScrollFill));
	sc_start = (ScrollStart*)map_device("sc_start",
		0x38012000,sizeof(ScrollStart));
	sc_start_b = (ScrollStart*)map_device("sc_start_b",
		0x38013000,sizeof(ScrollStart));
	sc_end = (ScrollEnd*)map_device("sc_end",
		0x38014000,sizeof(ScrollEnd));
	sc_end_b = (ScrollEnd*)map_device("sc_end_b",
		0x38015000,sizeof(ScrollEnd));
#ifdef KIMS
	screen.nplanes = 2;
	screen.vis_wid = 1536;
	screen.max_wid = 2048;
	screen.vis_hgt = 1152;
	screen.max_hgt = 2048; 
	screen.cur_minx = 0;
	screen.cur_miny = 0;
	screen.cur_maxx = 1535;
	screen.cur_maxy = 1151;
#else
	XXX
#endif
	video_fill(0,0,screen.max_wid,screen.max_hgt,0x0000);
	avss_init();
	font_init();
	video_setclip(0,0,screen.vis_wid,screen.vis_hgt); 
	console_font = &font[1];
	console_wid = (screen.vis_wid) / (console_font->wid);
	console_hgt = (screen.vis_hgt) / (console_font->hgt);
	set_cursor_mode(3,0);
	set_cursor_sprite(block_data,block_mask);
	vtstate.wid = console_wid;
	vtstate.hgt = console_hgt;
	vtstate.xpos = 0;
	vtstate.ypos = 0;
	vtstate.y1 = 0;
	vtstate.y2 = console_hgt-1;
	vtstate.attr = 0;
	vtstate.fg = 7;
	vtstate.bg = 0;
	vtstate.atval = 7;
	vtstate.esc = 0;
	vtstate.pcount = 0;
}

ushort cursor_mode;

set_cursor_mode(mode,crosshair)
int mode, crosshair;
{
	if(crosshair)
		crosshair = 1;
	switch(mode) {
		case 0:
			cursor_mode = 0xe000;
			break;
		case 1:
			cursor_mode = 0xe001+(crosshair<<4);
			break;
		case 2:
			cursor_mode = 0xe004+(crosshair<<4);
			break;
		case 3:
			cursor_mode = 0xe005+(crosshair<<4);
			break;
	}	
	cursor->command = cursor_mode;
}


set_cursor_xy(x,y)
int x,y;
{
#ifdef KIMS
        cursor->xpos = x/2 + 8;
        cursor->ypos = y + 40;
        kimsreg->vid_csr = (kimsreg->vid_csr & ~bit(5)) | ((x&1) << 5);
#else
        cursor->xpos = x + 8;
        cursor->ypos = y + 40;
#endif
}


set_cursor_sprite(data,mask)
ushort *data;
ushort *mask;
{
int i;
	cursor->command = cursor_mode | bit(12);
	for(i=0; i<16; i++)
		cursor->load = data[i];
	for(i=0; i<16; i++)
		cursor->load = mask[i];
	cursor->command = cursor_mode;

}


#ifdef KIMS
ushort avss_init_tab[] = {
        0x60, 0x01,     /* select plane 0 */
        0x83, 0x01,     /* address 1 */
        0x60, 0x02,     /* select plane 1 */
        0x83, 0x00,     /* address 0 */
        0x60, 0x03,     /* planes 0 and 1 */
        0x84, 0xca,     /* f(0) = s */
        0x85, 0xce,     /* f(1) = not(s) */
        0x86, 0xce,     /* f(2) = d or s */
        0x87, 0xc6,     /* f(3) = d xor s */
        0x88, 0xffff,   /* mask 1 & 2 */
        0x80, 0x00,     /* one plane res */
        0x81, 0x0c,     /* 16-bit bus */
        0x8a, 0xffff,   /* source register */
        0x8e, 0x0000,   /* background */
        0x8f, 0xffff,   /* foreground */
        0x90, 0x29,     /* CSR0 */
        0x60, 0x01,     /* select plane 0 */
        0x94, 0x35,     /* CSR4 - SRC1 */
        0x60, 0x02,     /* select plane 1 */
        0x94, 0x21,     /* CSR4 - SRC1 */
        0x60, 0x03,     /* select both planes */
        0x91, 0x05,     /* CSR1 - SRC2 */
        0x92, 0x00,     /* CSR2 - DEST1 */
        0x95, 0x09,     /* CSR5 */
        0x96, 0x00      /* CSR6 */
};      
#endif


void avss_id_write(adr,value)
        int adr;
        int value;
{
int t;
	t = 60000;
        while(!(adder->stat & bit(4)) && t--)
                continue;
	t = 60000;
        while(!(adder->stat & bit(6)) && t--)
                continue;
        adder->idd = value;
        adder->cmd = 0x100 + adr;
}




avss_init()
{
int i, t;
        t = 30000;
        while((!(adder->stat & bit(3)) || !(adder->stat & bit(4))) && t--)
                continue;
        for(i=0; i<sizeof(avss_init_tab)/sizeof(ushort); i += 2)
                avss_id_write(avss_init_tab[i], avss_init_tab[i+1]);
}



void video_setclip(x1,y1,x2,y2)
int x1,y1,x2,y2;
{
	adder->cxmn = x1;
	adder->cxmx = x2;
	adder->cymn = y1;
	adder->cymx = y2;
}


void video_setcolor(fg,bg)
int fg,bg;
{
	avss_id_write(0x60,0xff);	/* all planes */
	avss_id_write(0xa4,fg);		/* z-mode foreground load */
	avss_id_write(0xac,bg);		/* z-mode background load */
#ifdef KIMS
	if(bg & 1)
		sc_fill->fill_0 = 0xffff;
	else
		sc_fill->fill_0 = 0x0000;
	if(bg & 2)
		sc_fill->fill_1 = 0xffff;
	else
		sc_fill->fill_1 = 0x0000;
#endif
}


void video_fill(dx,dy,wid,hgt,fill)
	int dx,dy,wid,hgt,fill;
{
int t, i;
uchar *vp;
#ifdef QDSS
	t = 30000;
	while(!(adder->stat & bit(3)) && t--)
		continue;
	avss_id_write(0x60,0xff);	/* all planes */
	avss_id_write(0x88, 0xffff);    /* mask 1 & mask 2 */
        avss_id_write(0x8a, fill);      /* source register */
        while(!(adder->stat & bit(4)))
                continue;
        while(!(adder->stat & bit(6)))
                continue;
        adder->dxo = dx;
        adder->dyo = dy;
        adder->fdx = wid;
        adder->fdy = 0;
        adder->sdx = 0;
        adder->sdy = hgt;
        adder->fsc = 0x1fff;
        adder->ssc = 0x1fff;
        adder->mde = 0xc0;
        adder->cmd = 0x6c4;
#endif
}


void video_line(x1, y1, x2, y2, sdx, sdy, func, source)
	int x1, y1, x2, y2, sdx, sdy, func, source;
{
int t;
int dx,dy;
        t = 30000;
        while(!(adder->stat & bit(3)) && t--)
                continue;
        avss_id_write(0x88, 0xffff);    /* mask 1 & mask 2 */
        avss_id_write(0x8a, source);    /* source register */
        while(!(adder->stat & bit(4)))
                continue;
        while(!(adder->stat & bit(6)))
                continue;
        dx = (x2-x1);
        dy = (y2-y1);
	if (abs(dx) > abs(dy))
        	if (dx < 0)
                	dx--;
        	else
                	dx++;
	else
        	if (dy < 0)
                	dy--;
        	else
                	dy++;
        adder->dxo = x1 & 0x0fff;
        adder->dyo = y1 & 0x0fff;
        adder->fdx = dx & 0x0fff;
        adder->fdy = dy & 0x0fff;
        adder->sdx = sdx;
        adder->sdy = sdy;
        adder->fsc = 0x1fff;
        adder->ssc = 0x1fff;
        adder->mde = 0xc0;
        adder->cmd = 0x6c4 + (func << 4);
}


void video_box(x1,y1,wid,hgt,size,func,source)
	int x1,y1,wid,hgt,func,source;
{
	video_line(x1,y1,x1,y1+hgt,size,1,func,source);
	video_line(x1,y1,x1+wid,y1,1,size,func,source);
	video_line(x1,y1+hgt,x1+wid,y1+hgt,1,size,func,source);
	video_line(x1+wid,y1,x1+wid,y1+hgt,size,1,func,source);
}
	

void video_clear(dx,dy,dfvx,dsvy)
	int dx,dy,dfvx,dsvy;
{
	video_fill(dx,dy,dfvx,dsvy,0);
}


void video_copy(sx, sy, wid, hgt, dx, dy)
        int sx, sy, wid, hgt, dx, dy;
{
int t;
        t = 30000;
        while(!(adder->stat & bit(3)) && t--)
                continue;
	avss_id_write(0x60, 0xff);	/* all planes */
        avss_id_write(0x88, 0xffff);    /* mask 1 & mask 2 */
        avss_id_write(0x8a, 0xffff);    /* source register */
        while(!(adder->stat & bit(4)))
		continue;
        while(!(adder->stat & bit(6)))
                continue;
        adder->sxo = sx;
        adder->syo = sy;
        adder->fsdx = wid;
        adder->ssdy = hgt;
        adder->dxo = dx;
        adder->dyo = dy;
        adder->fdx = wid;
        adder->fdy = 0;
        adder->sdx = 0;
        adder->sdy = hgt;
        adder->fsc = 0x1fff;
        adder->ssc = 0x1fff;
        adder->mde = 0xc0;
        adder->cmd = 0xec4;
}

void video_put(buf, x,y,wid,hgt)
        short *buf;
        int x,y,wid,hgt;
{
int t;
register int r11,r10,r9,r8;
        t = 30000;
        while(!(adder->stat & bit(3)) && t--)
                continue;
        avss_id_write(0x88, 0xffff);    /* mask 1 & mask 2 */
        avss_id_write(0x8a, 0xffff);    /* source register */
        while(!(adder->stat & bit(4)))
                continue;
        while(!(adder->stat & bit(6)))
                continue;
        adder->sxo = 0;
        adder->syo = 0;
        adder->fsdx = wid;
        adder->ssdy = hgt;
        adder->dxo = x;
        adder->dyo = y;
        adder->fdx = wid;
        adder->fdy = 0;
        adder->sdx = 0;
        adder->sdy = hgt;
        adder->fsc = 0x1fff;
        adder->ssc = 0x1fff;
        adder->mde = 0xc0;
        adder->cmd = 0x7c1;
        r11 = (int)&adder->stat;
        r10 = (int)&adder->idd;
        r8 = (int)buf;
        r9 = r8 + ((wid*hgt)/16 + 1) * 2;
/*
        asm("gpl1: bbc $6,(r11),gpl1");
        asm("gpl2: movw (r8)+,(r10)");
        asm("      bbs $6,(r11),gpl2"); 
        asm("      cmpl r8,r9");
        asm("      blss gpl1");
	asm("halt");
*/
        asm("gpl1: bbc $6,(r11),gpl1");
        asm("gpl2: movw (r8)+,(r10)");
        asm("gpl3: bbc $6,(r11),gpl3"); 
        asm("      cmpl r8,r9");
        asm("      blss gpl1");
}



void video_scroll_up(x1,y1,wid,hgt,ofs)
	int x1,y1,wid,hgt,ofs;
{
int t,i;
char *vp;
#ifdef QDSS
        t = 30000;
        while(!(adder->stat & bit(4)) && t--)
                continue;
        while(!(adder->stat & bit(3)) && t--)
                continue;
	kimsreg->vid_csr &= 0xf4;
	kimsreg->vid_csr |= 8;
	while(sc_start->csr & SCROLL_GO)
		continue;
	sc_start->src_x = x1;
	sc_start->src_y = y1+ofs;
	sc_end->src_x = x1+wid-1;
	sc_end->src_y = y1+hgt-1;
	sc_start->dest_x = x1;
	sc_start->dest_y = y1;
	sc_end->dest_x = x1+wid-1;
	sc_end->dest_y = y1+hgt-1;
	sc_end->clip_x1 = x1;
	sc_end->clip_y1 = y1;
	sc_end->clip_x2 = x1+wid-1;
	sc_end->clip_y2 = y1+hgt-1;
	kimsreg->vid_csr &= 0xf4;
	kimsreg->vid_csr |= 8;
	sc_start->csr = SCROLL_GO | SCROLL_FAST | SCROLL_SXINC | SCROLL_SYINC
		| SCROLL_SMAJ | SCROLL_DXINC | SCROLL_DYINC | SCROLL_DMAJ;
	while(sc_start->csr & SCROLL_GO)
		continue;
	kimsreg->vid_csr = 0xf6;
        if(adder->stat & bit(3))
		return;
	adder->stat = 0;
        while(!(adder->stat & bit(4)) && t--)
                continue;
#endif
}




#ifdef KIMS
get_vromchar(buf,ch)
        ushort *buf;
        int ch;
{
int base, i;
	base = vrom[VROM_FONT_INDEX+ch*2];
        for(i=0; i<22; i++)
                buf[i] = vrom[VROM_FONT_INDEX+base+i*2];
}
#endif


get_promchar(buf,ch)
        ushort *buf;
        int ch;
{
uchar *cp;
int i;
	cp = (uchar*)prom+(prom[PROM_FONT_VECTOR] - 0x20040000);
	ch = (ch+96)%128;
        for(i=0; i<15; i++)
                buf[i] = cp[ch*15+i];
}


font_init()
{
int i,n;
short buf[32];
	font[0].ypos = screen.vis_hgt;
	font[0].xpos = 0;
	font[0].rwid = 128;
	font[0].swid = 16;
	font[0].wid = 12;
	font[0].hgt = 22;
#ifdef QDSS
	font[1].ypos = screen.vis_hgt*0 + 22;
	font[1].xpos = 0;
	font[1].rwid = 128;
	font[1].swid = 16;
	font[1].wid = 10;
	font[1].hgt = 15;
#endif
        for(i = 0; i<128; i++) {
#ifdef KIMS
                get_vromchar(buf, i);
                video_put(buf, i*font[0].swid, font[0].ypos, font[0].swid,
				font[0].hgt);
		get_promchar(buf, i,n);
                video_put(buf, i*font[1].swid, font[1].ypos, font[1].swid,
				font[1].hgt);
#endif
        }
}


Font *video_getfont(num)
int num;
{
	return(&font[num]);
}


video_putchar(font,x,y,ch)
Font *font;
int ch;
{
int yp,xp,i;
char *vp;
#ifdef KIMS
	xp = font->xpos + (ch % font->rwid)*(font->swid);
        yp = font->ypos + (ch / font->rwid)*(font->hgt);
        video_copy(xp, yp, font->wid, font->hgt, x, y);
#endif
}



