/*  video.c                                                     (CRC:495c3de5)
 *      relative path: ./VAX
 *      creation date: 1992
 *      last modified: Wed May 22 21:28:25 1996
 *      header update: Wed Dec 18 07:16:51 1996
 *  
 *       Digital VAX machine dependent code
 *  
 *  Copyright (c) 1992 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is a part of the Shag/OS operating system distribution, and is
 *  free software for non-profit uses; it may be freely used and distributed
 *  for non-profit purposes under the terms of the Shag/OS License, version 1.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND, 
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 *  AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT WILL THE AUTHOR, OR
 *  ANYONE RESPONSIBLE FOR THE LEGAL MODIFICATION OR REDISTRIBUTION OF THIS
 *  SOFTWARE, BE LIABLE FOR ANY DAMAGES, INCLUDING BUT NOT LIMITED TO ANY LOST
 *  PROFITS, LOST SAVINGS OR ANY INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER
 *  RESULTING FROM IMPAIRED OR LOST DATA, SOFTWARE OR COMPUTER FAILURE OR ANY
 *  OTHER CAUSE, EVEN IF THE AUTHOR IS ADVISED OF THE POSSIBILITY OF SUCH
 *  DAMAGES, OR FOR ANY OTHER CLAIM BY THE USER OR FOR ANY THIRD PARTY CLAIM.
 *  
 *  You should have received a copy of the Shag/OS License along with this
 *  program, in a file called 'license1.txt'.  If this license is missing,
 *  please contact the author and describe how this software was obtained.
 *  
 *  The author can be reached for questions, comments, and suggestions at:
 *  
 *      Frank Barrus, 4435 River Rd. Apt #3, Scottsville, NY 14546, USA
 *      email:   shaggy@csh.rit.edu
 *  
 *      For sending bug reports, please read the file 'bugs', and always
 *      include the product name and version, and the word 'BUG' in the subject.
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

/* video.c 
 *  PC/AT basic video routines
 *  to be included from I386/console.c
 */



static int panic_lines;
static short wid, hgt, xpos, ypos;

static short *xpos_ptr, *ypos_ptr;

#ifdef QDSS
#include <vid_qdss.c>
#endif
#ifdef VS2000
#include <vid_vs2000.c>
#endif


void console_getxy(short *xptr, short *yptr) 
{
	*xptr = *xpos_ptr;
	*yptr = *ypos_ptr;
}

void console_setxyvars(short *xptr, short *yptr) 
{
	xpos_ptr = xptr;
	ypos_ptr = yptr;
}


void video_init(void)
{
paddr vbase;
int n;
	wid = bootparam.scr_wid;
	hgt = bootparam.scr_hgt;
	xpos = bootparam.cur_x;
	ypos = bootparam.cur_y;
	vbase = (paddr)(bootparam.video_seg << 4);
	video_base = (short*)vm_map(&kernel.data,vbase,wid*hgt*2,VM_KRW);
	panic_lines = 0;
	crtc_adr = bootparam.crtc_adr;
	n = in_regw(0xe,crtc_adr);
	ypos = n/wid;
	xpos = n%wid; 
	console_setxyvars(&xpos,&ypos);
}


static void fault_loop(void)
{
int cflag;
	CRITICAL_BEGIN(cflag)
	panic_event.event = 0;
	video_write("\007 Fault loop!\n",14);
	halt();
	/* theoretically, a user can
	 * choose to continue from halt().. */
	panic_lines = 0;
	CRITICAL_END(cflag)
}


int video_write(char *buf, int len)
{
char ch;
int x,n,i;
char *bp;
	xpos = *xpos_ptr;
	ypos = *ypos_ptr;
	i = len;
        while(i--) {
                ch = *buf++;
                switch(ch) {
                case '\b':
                        if (xpos > 0)
                                xpos--;
                        break;
                case '\t':
                        xpos = (xpos & ~7)+8;
                        break;
		case '\007':
			video_beep(400,100);
			break;
		case '\013':
			xpos = 0;
			ypos = 0;
			break;
                case '\r':
                        xpos = 0;
                        break;
                case '\n':
			xpos = 0;
                        ypos++;
			if (panic_event.event) {
				if (panic_lines++ >= hgt-8) 
					fault_loop();
			}
                        break;
                default:
			video_putchar(xpos,ypos,ch);
                        if ((xpos++) >= wid) {
                                xpos = 0;
                                ypos++;
                        }
                }
                if (ypos >= hgt) {
			x = len;
			n = 1;
			bp = buf;
			while(x-- > 0 && n < hgt) 
				if (*bp++ == '\n')
					n++;
                        ypos = hgt-n;
			video_scroll_up(0,0,wid,hgt,n);
                }
        }
	*xpos_ptr = xpos;
	*ypos_ptr = ypos;
	video_setcursor(xpos,ypos);
	return(len);
}


video_beep()
{
        serial_write_polled(0,"\247",1);
}



