/*  error.c                                                     (CRC:4dfae50e)
 *      relative path: ./clib
 *      creation date: 1992
 *      last modified: Wed May 22 21:31:39 1996
 *      header update: Wed Dec 18 07:16:56 1996
 *  
 *       C Library
 *  
 *  Copyright (c) 1992 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is a part of the Shag/OS operating system distribution, and is
 *  free software for non-profit uses; it may be freely used and distributed
 *  for non-profit purposes under the terms of the Shag/OS License, version 1.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND, 
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 *  AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT WILL THE AUTHOR, OR
 *  ANYONE RESPONSIBLE FOR THE LEGAL MODIFICATION OR REDISTRIBUTION OF THIS
 *  SOFTWARE, BE LIABLE FOR ANY DAMAGES, INCLUDING BUT NOT LIMITED TO ANY LOST
 *  PROFITS, LOST SAVINGS OR ANY INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER
 *  RESULTING FROM IMPAIRED OR LOST DATA, SOFTWARE OR COMPUTER FAILURE OR ANY
 *  OTHER CAUSE, EVEN IF THE AUTHOR IS ADVISED OF THE POSSIBILITY OF SUCH
 *  DAMAGES, OR FOR ANY OTHER CLAIM BY THE USER OR FOR ANY THIRD PARTY CLAIM.
 *  
 *  You should have received a copy of the Shag/OS License along with this
 *  program, in a file called 'license1.txt'.  If this license is missing,
 *  please contact the author and describe how this software was obtained.
 *  
 *  The author can be reached for questions, comments, and suggestions at:
 *  
 *      Frank Barrus, 4435 River Rd. Apt #3, Scottsville, NY 14546, USA
 *      email:   shaggy@csh.rit.edu
 *  
 *      For sending bug reports, please read the file 'bugs', and always
 *      include the product name and version, and the word 'BUG' in the subject.
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

/* error.c
 *   routines to deal with and record errors
 *   and traps
 */

#include <sys/types.h>
#include <error.h>
#include <string.h>
#include <stdio.h>

char *strerror(obj_t oid)
{
	switch(oid) {
		case ERR_FAILURE:  	return("general failure");

		case ERR_INVPARAM: 	return("invalid parameter");
		case ERR_INVTYPE:	return("invalid type"); 

		case ERR_NOFILE:   	return("file not found");
		case ERR_INVFILE:  	return("invalid file format");
		case ERR_SHORTFILE:	return("unexpected EOF");
		case ERR_EOF:		return("end of file");

		case ERR_EOB:		return("end of buffer");
		case ERR_NOMEM:		return("not enough memory");
		case ERR_NOBUF:		return("not enough buffers");

		case ERR_NOMETHOD:	return("method not found");
		case ERR_NOATTR:	return("attribute not found");
		case ERR_EXISTATTR:	return("attribute exists");
		case ERR_EOA:		return("end of attributes");
		case ERR_INVOBJ:	return("invalid object");
		case ERR_NOREAD:	return("no read access");
		case ERR_NOWRITE:	return("no write access");

		case ERR_NOTHREAD: 	return("thread not found");

		case ERR_NOTASK: 	return("task not found");

		default:		return("unknown error");
	}
}

void print_error(const char *s, obj_t errcode)
{
	oprintf(OBJ_STDERR, "%s: %s\n", s, strerror(errcode));
}


/* XXX - how is this going to get the last error?? 
void perror(const char *s)
{
	printf("%s: %s\n", s, strerror(
}
*/


