/*  stdio.c                                                     (CRC:77c3a33a)
 *      relative path: ./clib
 *      creation date: 1992
 *      last modified: Wed May 22 21:31:51 1996
 *      header update: Wed Dec 18 07:16:58 1996
 *  
 *       C Library
 *  
 *  Copyright (c) 1992 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is a part of the Shag/OS operating system distribution, and is
 *  free software for non-profit uses; it may be freely used and distributed
 *  for non-profit purposes under the terms of the Shag/OS License, version 1.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND, 
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 *  AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT WILL THE AUTHOR, OR
 *  ANYONE RESPONSIBLE FOR THE LEGAL MODIFICATION OR REDISTRIBUTION OF THIS
 *  SOFTWARE, BE LIABLE FOR ANY DAMAGES, INCLUDING BUT NOT LIMITED TO ANY LOST
 *  PROFITS, LOST SAVINGS OR ANY INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER
 *  RESULTING FROM IMPAIRED OR LOST DATA, SOFTWARE OR COMPUTER FAILURE OR ANY
 *  OTHER CAUSE, EVEN IF THE AUTHOR IS ADVISED OF THE POSSIBILITY OF SUCH
 *  DAMAGES, OR FOR ANY OTHER CLAIM BY THE USER OR FOR ANY THIRD PARTY CLAIM.
 *  
 *  You should have received a copy of the Shag/OS License along with this
 *  program, in a file called 'license1.txt'.  If this license is missing,
 *  please contact the author and describe how this software was obtained.
 *  
 *  The author can be reached for questions, comments, and suggestions at:
 *  
 *      Frank Barrus, 4435 River Rd. Apt #3, Scottsville, NY 14546, USA
 *      email:   shaggy@csh.rit.edu
 *  
 *      For sending bug reports, please read the file 'bugs', and always
 *      include the product name and version, and the word 'BUG' in the subject.
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

/*  stdio.c
 *    standard I/O routines for user programs 
 */


#include <sys/types.h>
#include <sys/resource.h>
#include <stdio.h>
#include <error.h>
#include <string.h>
#include <stdarg.h>



/* var-arg printf to any object: */

int voprintf(obj_t oid, char *fmt, va_list args)
{
char buf[512];
int len;
	len = vsprintf(buf, fmt, args);
	return(_(oid, _write, Ptr(buf), len, ___));
}


/* printf to any object: */

int oprintf(obj_t oid, char *fmt, ...)
{
va_list args;
int len;
	va_start(args, fmt);
	len = voprintf(oid, fmt, args);
	va_end(args);
	return(len);
}


/* normal printf() to standard out: */
 
int printf(char *fmt, ...)
{
va_list args;
int len;
	va_start(args, fmt);
	len = voprintf(OBJ_STDOUT, fmt, args);
	va_end(args);
	return(len);
}


/* printf() to standard error: */

int eprintf(char *fmt, ...)
{
va_list args;
int len;
	va_start(args, fmt);
	len = voprintf(OBJ_STDERR, fmt, args);
	return(len);
}


/* printf() to console: */

int cprintf(char *fmt, ...)
{
va_list args;
int len;
	va_start(args, fmt);
	len = voprintf(OBJ_SYS, fmt, args);
	return(len);
}


/* variable-argument list printf(): */

int vprintf(char *fmt, va_list args)
{
	return(voprintf(OBJ_STDOUT, fmt, args));
}





/* put a string to standard out */

int puts(char *s)
{
	_(OBJ_STDOUT, _write, Ptr(s), strlen(s), ___);
	_(OBJ_STDOUT, _write, Ptr("\n"), 1, ___);
	return(0);
}



/* put a character to standard out */

int putchar(char ch)
{
	_(OBJ_STDOUT, _write, Ptr(&ch), 1, ___);
	return(ch);
}



/* get a character from standard in */

int getchar()
{
char ch;
	while(!_(OBJ_STDIN, _read, Ptr(&ch), 1, ___)) 
		continue;
	return(ch);
}


/* note:  getstrn() has replaced gets() 
 *  	(read about Internet worms)
 */
char *getstrn(char *str, int len)
{
	return(ogets(str, len, OBJ_STDIN));
}


char *ogets(char *str, int len, obj_t obj)
{
obj_t rv;
	while(len > 1) {
		rv = err_(obj, _read, Ptr(str), 1, ___);
		if(!err(rv) && rv > 0) {
			if(*str == '\n')
				break;
			str++;
			len--;
		} else {	
			*str = '\0';
			return(NULL);
		}	
	}
	*str = '\0';
	return(str);
}


