/*  $Id$
 *      virtual memory management 
 *      that is common for both 386 and VAX
 *      
 *      relative path: ./common
 *  
 *       Shared machine dependent code
 *  
 *  Copyright (c) 1992 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is a part of the Shag/OS operating system distribution, and is
 *  free software for non-profit uses; it may be freely used and distributed
 *  for non-profit purposes under the terms of the Shag/OS License, version 1.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND, 
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 *  AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT WILL THE AUTHOR, OR
 *  ANYONE RESPONSIBLE FOR THE LEGAL MODIFICATION OR REDISTRIBUTION OF THIS
 *  SOFTWARE, BE LIABLE FOR ANY DAMAGES, INCLUDING BUT NOT LIMITED TO ANY LOST
 *  PROFITS, LOST SAVINGS OR ANY INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER
 *  RESULTING FROM IMPAIRED OR LOST DATA, SOFTWARE OR COMPUTER FAILURE OR ANY
 *  OTHER CAUSE, EVEN IF THE AUTHOR IS ADVISED OF THE POSSIBILITY OF SUCH
 *  DAMAGES, OR FOR ANY OTHER CLAIM BY THE USER OR FOR ANY THIRD PARTY CLAIM.
 *  
 *  You should have received a copy of the Shag/OS License along with this
 *  program, in a file called 'license1.txt'.  If this license is missing,
 *  please contact the author and describe how this software was obtained.
 *  
 *  The author can be reached for questions, comments, and suggestions at:
 *  
 *      Frank Barrus, 4435 River Rd. Apt #3, Scottsville, NY 14546, USA
 *      email:   shaggy@csh.rit.edu
 *  
 *      For sending bug reports, please read the file 'bugs', and always
 *      include the product name and version, and the word 'BUG' in the subject.
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

/*  $Log$
 *      creation date: 1992
 */


void vm_print_pt(Vmap *vm);
static int vm_freepage(int page);
static int vm_allocpage(void);
static ZoneP seg_zone = NULL;

static int vm_size(Vmap *vm, int vpage)
{
int size;	
register pte_t pte;
	size = 0;
	pte = vm->pte[vpage];
	while(!PTE_ISFREE(pte)) {
		size++;
		if(PTE_ISEOS(pte) || ++vpage >= vm->ptlen)
			break;
		pte = vm->pte[vpage];
	}
	return(size);
}


static int vm_freesize(Vmap *vm, int vpage)
{
int size;	
	size = 0;
	while(vpage < vm->ptlen && PTE_ISFREE(vm->pte[vpage])) {
		vpage++;
		size++;
	}
	return(size);
}


ulong vm_getpsize(Vmap *vm, ulong shared)
{
int n;
int vpage;
int len,slen;
int page;
pte_t pte;
	if(vm->vbase == STACKBASE) 
		vpage = MAXPTE-vm->ptlen;
	else 
		vpage = 0;
	len = 0;
	slen = 0;
	n = vm->ptlen;
	while(n--) {
		pte = vm->pte[vpage++];
		if(PTE_ISVALID(pte))
			slen += PAGE_SIZE;
		if(PTE_ISEOS(pte)) {
			PTE_GETPAGE(pte,page);
			if((!page && !shared) || (page>0 && shared>0))
				len += slen;
			slen = 0;
		}
	}
	if(slen > 0)
		panic(0,"Unterminated segment");
	return(len);
}


static vaddr vm_findfree(Vmap *vm, int np) 
{
register int vpage;
register int i;
	vpage =  0;
	while((i=vm_freesize(vm,vpage)) < np) {
		if(i > 0)
			vpage += i;
		else {
			i = vm_size(vm,vpage);
			if(!i)
				panic(0,"Virtually exhausted; vpage=%x",
					vpage);
			vpage += i;
		}
	}
	return(vpage);
}


Vmap *vm_getmap(vaddr adr)
{
	if(adr <= TASKBASE+TASKSIZE) {
		if(thread && thread->task)
			return(&thread->task->data);
	} else if(adr >= TEMPBASE && adr < TEMPBASE+TEMPSIZE) {
		if(thread)
			return(&thread->temp);
	} else if(adr >= STACKTOP-STACKSIZE && adr < STACKTOP) {
		if(thread)
			return(&thread->stack);
	} else if(adr >= KERNBASE && adr < KERNBASE+KERNSIZE) {
		return(&kernel.data);
	}
	return(NULLP(Vmap));
}


paddr vm_getphys(vaddr adr)
{
int ofs;
Vmap *vm;
int page;
int vpage;
int smin,smax;
	ofs = adr & (PAGE_SIZE-1);
	adr -= ofs;
	vm = vm_getmap(adr);
	vpage = (adr-vm->vbase)/PAGE_SIZE;
	if(vm->vbase == STACKBASE) {
		smin = MAXPTE-vm->ptlen;
		smax = MAXPTE-1;
	} else {
		smin = 0;
		smax = vm->ptlen-1;
	}
	if(vpage < smin || vpage > smax || !PTE_ISVALID(vm->pte[vpage]))
		panic(0,"invalid vm_getphys()");
	PTE_GETPAGE(vm->pte[vpage], page);
	return(page*PAGE_SIZE+ofs);
}


vaddr vm_alloc(Vmap *vm, int len, ulong accmode)
{
vaddr start;
register int vpage;
register int np;
	accmode &= VM_ACC;
	LOCK(vm->lock);
	np = (len-1)/PAGE_SIZE+1;
	vpage = vm_findfree(vm, np+1);
	start = vm->vbase + vpage*PAGE_SIZE;	
	while(np-- > 0) {
		PTE_SET(vm->pte[vpage], 0, VM_ZERO|VM_USED|accmode);
		vpage++;
	}
	/* mark end of segment with illegal page for boundary checks: */
	PTE_SETEOS(vm->pte[vpage]);
	UNLOCK(vm->lock) 
	TBIR(start,len+PAGE_SIZE);
	return(start);
}


vaddr vm_reserve(Vmap *vm, int len)
{
vaddr start;
register int vpage;
register int np;
	LOCK(vm->lock);
	np = (len-1)/PAGE_SIZE+1;
	vpage = vm_findfree(vm, np+1);
	start = vm->vbase + vpage*PAGE_SIZE;	
	while(np-- > 1) {
		PTE_SET(vm->pte[vpage], 0, VM_RESERVED|VM_NA);
		vpage++;
	}
	/* mark end of segment with illegal page for boundary checks: */
	PTE_SETEOS(vm->pte[vpage]);
	UNLOCK(vm->lock) 
	TBIR(start,len+PAGE_SIZE);
	return(start);
}

static vaddr vm_map_ex(Vmap *vm, paddr phys, int len, ulong accmode, int join)
{
vaddr start;
register int vpage;
register int np;
int ofs;
int page;
	accmode &= VM_ACC;
	ofs = phys & (PAGE_SIZE-1);
	phys -= ofs;
	len += ofs;
	LOCK(vm->lock);
	np = (len-1)/PAGE_SIZE+1;
	vpage = vm_findfree(vm, np+1);

	if(join && vpage && PTE_ISEOS(vm->pte[vpage-1]))
		vpage--;

	start = vm->vbase + vpage*PAGE_SIZE;	
	page = phys/PAGE_SIZE;
	while(np-- > 0) {
		PTE_SET(vm->pte[vpage], page, VM_LOCKED|VM_VALID|accmode);
		page++;
		vpage++;
	}
	/* mark end of segment with illegal page for boundary checks: */
	PTE_SETEOS(vm->pte[vpage]);
	UNLOCK(vm->lock) 
	TBIR(start,len+PAGE_SIZE);
	return(start+ofs);
}

vaddr vm_map(Vmap *vm, paddr phys, int len, ulong accmode)
{
	return vm_map_ex(vm, phys, len, accmode, FALSE);
}

vaddr vm_map_join(Vmap *vm, paddr phys, int len, ulong accmode)
{
	return vm_map_ex(vm, phys, len, accmode, TRUE);
}

vaddr vm_share(Vmap *vm, Vmap *vmsrc, vaddr adr, int len, ulong accmode)
{
vaddr start;
register int vpage;
register int spage;
register int np;
int ofs;
int smin,smax;
VmSeg *seg;
	accmode &= VM_ACC;
	ofs = adr & (PAGE_SIZE-1);
	adr -= ofs;
	len += ofs;
	if(!seg_zone)
		seg_zone = zinit(sizeof(VmSeg), "vmseg");
	seg = (VmSeg*)zalloc(seg_zone);
	LOCK(vm->lock);
	np = (len-1)/PAGE_SIZE+1;
	vpage = vm_findfree(vm, np+1);
	spage = (adr - vmsrc->vbase)/PAGE_SIZE;
	if(vmsrc->vbase == STACKBASE) {
		smin = MAXPTE-vmsrc->ptlen;
		smax = MAXPTE-1;
	} else {
		smin = 0;
		smax = vmsrc->ptlen-1;
	}
	start = vm->vbase + vpage*PAGE_SIZE;	

	seg->type = VMSEG_SHARED;
	seg->start = vpage;
	seg->size = np;
	seg->vmshare = vm;
	seg->src.vmap = vmsrc;
	seg->index = spage;
	seg->nextseg = vm->seg;
	seg->nextshare = vmsrc->share;
	vm->seg = seg;
	vmsrc->share = seg;

	while(np-- > 0) {
		if(spage < smin || spage > smax)
			panic(0,"vm_share: invalid source");
		vm->pte[vpage] = vmsrc->pte[spage];
		PTE_SETSOFT(vm->pte[vpage], VM_SHARED);
		PTE_SETACC(vm->pte[vpage], accmode);
		vpage++;
		spage++;
	}
	/* mark end of segment with illegal page for boundary checks: */
	PTE_SETEOS(vm->pte[vpage]);
	UNLOCK(vm->lock) 
	TBIR(start,len+PAGE_SIZE);
	return(start+ofs);
}


vaddr vm_stack_alloc(Vmap *vm, int len, ulong accmode) 
{
register int vpage;
register int np;
vaddr start;
int page;
	accmode &= VM_ACC;
	vpage = MAXPTE-1;
	LOCK(vm->lock);
	np = (len-1)/PAGE_SIZE+1;
	while(!PTE_ISFREE(vm->pte[vpage])) {
		vpage--;
		if(vpage < MAXPTE-vm->ptlen)
			break;
	}
	start = vm->vbase + vpage*PAGE_SIZE;
	vpage -= np;
	if(vpage < MAXPTE-vm->ptlen) 
		panic(0,"Out of stack space");
	while(np-- > 0) {
		page = vm_allocpage();
		PTE_SET(vm->pte[vpage], page, VM_USED|VM_VALID|accmode);
		vpage++;
	}
	/* mark end of segment with illegal page for boundary checks: */
	PTE_SETEOS(vm->pte[vpage]);
	UNLOCK(vm->lock) 
	TBIR(start,len+PAGE_SIZE);
	return(start);
}


vaddr vm_realloc(Vmap *vm, vaddr adr, int size)
{
int np;
int i, n;
int vpage;
int cursize;
vaddr start;
ulong accmode;
int page;
/* XXX - should probably check for shared segments, and change their
 * references as well
 */
	LOCK(vm->lock);
	np = (size-1)/PAGE_SIZE + 2;	
	vpage = (adr-vm->vbase)/PAGE_SIZE;
	PTE_GETACC(vm->pte[vpage],accmode);
	cursize = vm_size(vm,vpage);
	if(np == cursize) {
		UNLOCK(vm->lock);
		return(adr);
	}
	if(np < cursize) {
		PTE_SETEOS(vm->pte[vpage+np-1]);
		UNLOCK(vm->lock);
		vm_free(vm,adr+np*PAGE_SIZE);
		TBIR(adr,(np+1)*PAGE_SIZE);
		return(adr);
	}
	i = np-cursize;
	n = vpage+cursize-1;
	if(vm_freesize(vm,n+1) >= i) {
		while(i--) {
			page = vm_allocpage();
			PTE_SET(vm->pte[n], page,
				 accmode|VM_USED|VM_VALID);
			n++;
		}
		PTE_SETEOS(vm->pte[n]);
		UNLOCK(vm->lock);
		TBIR(adr,size+PAGE_SIZE);
		return(adr);
	}
	n = vm_findfree(vm,np);
	start = vm->vbase + n*PAGE_SIZE;
	for(i=0; i<cursize-1; i++) {
		vm->pte[n++] = vm->pte[vpage];
		PTE_SETFREE(vm->pte[vpage]);
		vpage++;
	}
	for(i=cursize-1; i<np-1; i++) {
		page = vm_allocpage();
		PTE_SET(vm->pte[n], page, accmode|VM_USED|VM_VALID);
		n++;
	}
	PTE_SETEOS(vm->pte[n]);
	UNLOCK(vm->lock);
	TBIR(start,size+PAGE_SIZE);
	return(start);
}


void vm_free(Vmap *vm, vaddr adr)
{
/* XXX - should check if this range is part of a segment
 * and if so, free it */
int vpage;
int page;
pte_t pte;
int len;
ulong soft;
	LOCK(vm->lock);
	vpage = (adr-vm->vbase)/PAGE_SIZE;
	pte = vm->pte[vpage];
	len = PAGE_SIZE;
	while(!PTE_ISEOS(pte)) {
		if(PTE_ISFREE(pte))
			panic(0,"Invalid vm_free(%x,%x)", vm, adr);
		PTE_GETSOFT(pte, soft);
		PTE_GETPAGE(pte, page);
		PTE_SETFREE(vm->pte[vpage]);
		if(soft == VM_USED) 
			vm_freepage(page);
		len += PAGE_SIZE;
		if(++vpage >= vm->ptlen)
			break;
		pte = vm->pte[vpage];
	}
	PTE_SETFREE(vm->pte[vpage]);
	TBIR(adr,len);
	UNLOCK(vm->lock);
}


vaddr vm_temp_alloc(int len, ulong acc)
{
	if(thread != null_thread)
		return(vm_alloc(&thread->temp, len, acc));
	else
		return(vm_alloc(&kernel.data, len, acc));
}

vaddr vm_temp_map(paddr phys, int len, ulong acc)
{
	if(thread != null_thread)
		return(vm_map(&thread->temp, phys, len, acc));
	else
		return(vm_map(&kernel.data, phys, len, acc));
}

vaddr vm_temp_share(Vmap *vmsrc, vaddr adr, int len, ulong acc)
{
	if(thread != null_thread)
		return(vm_share(&thread->temp, vmsrc, adr, len, acc));
	else
		return(vm_share(&kernel.data, vmsrc, adr, len, acc));
}


void vm_temp_free(vaddr adr)
{
	if(thread != null_thread)
		vm_free(&thread->temp, adr);
	else
		vm_free(&kernel.data, adr);
}


void vm_copyto(Vmap *vm, vaddr dest, vaddr src, int len)
/* src is current, dest is in vm struct */
{
vaddr tmp;
	if(vm == &kernel.data) {
		memcpy((void*)dest,(void*)src,len);
		return;
	}
	tmp = vm_temp_share(vm, dest, len, VM_KRW);
	memcpy((void*)tmp,(void*)src,len);
	vm_temp_free(tmp);
}


void vm_copyfrom(Vmap *vm, vaddr src, vaddr dest, int len)
/* dest is current, src is in vm struct */
{
vaddr tmp;
	if(vm == &kernel.data) {
		memcpy((void*)dest,(void*)src,len);
		return;
	}
	tmp = vm_temp_share(vm, src, len, VM_KRW);
	memcpy((void*)dest,(void*)tmp,len);
	vm_temp_free(tmp);
}

void vm_setaccess(Vmap *vm, vaddr adr, int len, ulong access)
{
int vpage;
int np;
	access &= VM_ACC;
	len += (adr & (PAGE_SIZE-1));
	adr &= ~(PAGE_SIZE-1);
	LOCK(vm->lock);
	vpage = (adr-vm->vbase)/PAGE_SIZE;
	np = (adr+len-vm->vbase)/PAGE_SIZE-vpage+1;
	while(np-- > 0) {
		PTE_SETACC(vm->pte[vpage], access);
		vpage++;
	}
	TBIR(adr,len);
	UNLOCK(vm->lock)
}


static int vm_freepage(int page)
{
int n;
	LOCK(pagelock) 
	if (page < vm_npages && page >= 0) {
		n = --vm_refcount[page];
		if(!n) {
			vm_refcount[page] = vm_firstfree;
			vm_firstfree = page;
			vm_nfreepages++;
		}
	} else
		n = 0;
	UNLOCK(pagelock) 
	return(n);
}


static int vm_allocpage(void)
{
int n;
	n = -1;
	LOCK(pagelock) 
	if (vm_nfreepages > 0) {
		n = vm_firstfree;
		vm_firstfree = vm_refcount[n];
		vm_refcount[n] = 1;
		vm_nfreepages--;
	}
	UNLOCK(pagelock) 
	if(vm_nfreepages<(32*K/PAGE_SIZE))
		kprintf("*** Warning: pages left: %d\n", vm_nfreepages);
	if(n<0)
		panic(0,"out of physical memory");
	return(n);
}



void vm_print_pt(Vmap *vm)
{ 
int i;
ulong acc,nextacc,first;
ulong page;
ulong soft;
char accstr[4];
        kprintf("pte(%8x): pte=%x/%x ptlen=%x base=%x\n", (vaddr)vm,
                        (vaddr)vm->pte, vm_getphys((vaddr)vm->pte), 
                        vm->ptlen, vm->vbase);
	first = 0;
	nextacc = 0;
	for(i=0; i<vm->ptlen; i++) {
		PTE_GETACC(vm->pte[i], acc);
		if(i<vm->ptlen-1) 
			PTE_GETACC(vm->pte[i+1], nextacc);
		PTE_GETSOFT(vm->pte[i+1],soft)
		if(soft != VM_EOS && i != vm->ptlen-1 && nextacc == acc) 
			continue;
		PTE_GETSOFT(vm->pte[first],soft);
		kprintf("%08x-%08x: s:%08x ",
			vm->vbase+first*PAGE_SIZE,
			vm->vbase+(i+1)*PAGE_SIZE-1,
			(i+1-first)*PAGE_SIZE);
		switch(acc) {
		case VM_KR: strcpy(accstr,"KR "); break;
		case VM_UR: strcpy(accstr,"UR "); break;
		case VM_KRW: strcpy(accstr,"KRW"); break;
		case VM_URW: strcpy(accstr,"URW"); break;
#ifdef VAX
		case VM_URKW: strcpy(accstr,"URKW"); break;
#endif
		}
		switch(soft) {
		case VM_EOS:
			kprintf("stray EOS\n");
			break;
		case VM_FREE:
			kprintf("free\n");
			break;
		case VM_RESERVED:
			kprintf("reserved\n");
			break;
		case VM_SHARED:
			kprintf("shared a:%s\n", accstr);
			break;
		case VM_LOCKED:
		case VM_USED:
			PTE_GETPAGE(vm->pte[first],page);
			kprintf("p:%08x a:%s\n",
				page*PAGE_SIZE, accstr);
			break;
		}
		PTE_GETSOFT(vm->pte[i+1],soft)
		if(soft == VM_EOS)
			i++;
		first = i+1;
        }
}

void vm_status(void)
{
        kprintf("vmstat: npages=%d nfreepages=%d (%dK, %dK free)\r\n",
                vm_npages, vm_nfreepages,
                vm_npages*PAGE_SIZE/K, vm_nfreepages*PAGE_SIZE/K );
        kprintf("kernel usage: %d virtual pages (%dK)\r\n",
                kernel.data.ptlen, kernel.data.ptlen*PAGE_SIZE/K);
}

void vm_list(void)
{
	vm_print_pt(&kernel.data);
}

static void vm_vmap_init(Vmap *vm)
{
int i;
        for(i=0; i<vm->ptlen; i++)
                PTE_SETFREE(vm->pte[i]);
	vm->seg = NULLP(VmSeg);
	LOCK_CLEAR(vm->lock); 
}


void vm_task_open(Vmap *vm)
{
        vm->ptlen = align(((64*K)/PAGE_SIZE),NPTEPAGE);
        vm->vbase = TASKBASE;
        vm->pte = (pte_t*)vm_alloc(&kernel.data,
			 vm->ptlen*sizeof(pte_t), VM_KRW);
        vm_vmap_init(vm);
}


void vm_stack_temp_open(Vmap *vm, Vmap *vmtemp)
{
int i, n;
	vm->ptlen = align((STACKSIZE/PAGE_SIZE),NPTEPAGE);
#ifdef I386
	n = vm->ptlen;
#else	/* VAX */
	n = vm->ptlen + align((TEMPSIZE/PAGE_SIZE),NPTEPAGE);
#endif
        vm->pte = (pte_t*)vm_alloc(&kernel.data,
			n*sizeof(pte_t),VM_KRW)
			 + (n - MAXPTE);
	for(i=MAXPTE-n; i<MAXPTE; i++)
                PTE_SETFREE(vm->pte[i]);
        vm->vbase = STACKBASE;
	vm->seg = NULLP(VmSeg);
        vmtemp->ptlen = align((TEMPSIZE/PAGE_SIZE),NPTEPAGE);
        vmtemp->vbase = TEMPBASE;
#ifdef I386
        vmtemp->pte = (pte_t*)vm_alloc(&kernel.data,
			 vmtemp->ptlen*sizeof(pte_t), VM_KRW);
#else	/* VAX */
	vmtemp->pte = vm->pte + (TEMPBASE-STACKBASE)/PAGE_SIZE;
#endif
        vm_vmap_init(vmtemp);
	LOCK_CLEAR(vm->lock);
}


void vm_task_close(Vmap *vm)
{
int n;
pte_t *tp;
int page;
ulong soft;
	tp = vm->pte;
	n = vm->ptlen;
        while(n--) {
		PTE_GETSOFT(*tp,soft);
                if(soft == VM_USED && PTE_ISVALID(*tp)) {
                        PTE_GETPAGE(*tp, page);
                        vm_freepage(page);
                }
		tp++;
        }
        vm_free(&kernel.data,(vaddr)vm->pte);
	vm->pte = NULLP(pte_t);
}

void vm_stack_temp_close(Vmap *vm, Vmap *vmtemp)
{
#ifdef I386	
	vm->pte = &vm->pte[MAXPTE-vm->ptlen];
	vm_task_close(vm);
	vm_task_close(vmtemp);
#endif
#ifdef VAX
	vm->pte = &vm->pte[MAXPTE-vm->ptlen-vmtemp->ptlen];
	vm_task_close(vm);
#endif
}


static void vm_fault(ulong reason, vaddr va, vaddr pc, char *msg)
{
char *acctype;
	if(VM_FAULT_KERNEL(reason)) {
		if(VM_FAULT_READ(reason))
			acctype = "kread";
		else
			acctype = "kwrite";
	} else {
		if(VM_FAULT_READ(reason))
			acctype = "read";
		else
			acctype = "write";
	}
	panic(pc,"%s; %s ref=%x", msg, acctype, va);
}


static void vm_pagemap(ulong reason, vaddr va, vaddr pc,
			Vmap *vmdest, int ndest,
			Vmap *vmsrc, int nsrc)
{
ulong acc, soft, flags;
ulong page;
register pte_t pte;
	pte = vmsrc->pte[nsrc];
	PTE_GETACC(pte,acc);
	PTE_GETSOFT(pte,soft);
	PTE_GETFLAGS(pte,flags);
	PTE_GETPAGE(pte,page);
	switch(soft) {
	case VM_FREE:
		vm_fault(reason, va, pc, "Unallocated memory access");
		break;
	case VM_USED:
	case VM_LOCKED:
		if(PTE_ISVALID(pte)) { 
			vm_fault(reason, va, pc, "Bus error"); 
		} else switch(flags) {
		case VM_ZERO:
			page = vm_allocpage();
			PTE_SET(vmsrc->pte[nsrc],page,VM_USED|VM_VALID|acc);
			PTE_SETPAGE(vmdest->pte[ndest], page);
			PTE_SETVALID(vmdest->pte[ndest]);
			TBIS(va);
			memset((void*)(va & ~(PAGE_SIZE-1)), 0, PAGE_SIZE);
			break;
		default:
			vm_fault(reason, va, pc, "Invalid PTE");
		}
		break;
	case VM_SHARED: {
		VmSeg *seg;
		int index;

		seg = vmsrc->seg;
		while(seg && 
		 (nsrc < seg->start || nsrc >= seg->start+seg->size))
			seg = seg->nextseg;
		if(!seg)
			vm_fault(reason, va, pc, "Illegal shared page");
		if(seg->type == VMSEG_SHARED) {
			index = seg->index + (nsrc - seg->start);
			vm_pagemap(reason, va, pc, vmdest, ndest, 
						seg->src.vmap, index);
		} else
			vm_fault(reason, va, pc, "Illegal shared segment");
		break; }
	case VM_RESERVED:
		vm_fault(reason, va, pc, "Reserved page");
		break;
	case VM_EOS:
		vm_fault(reason, va, pc, "Segmentation boundary fault");
		break;
	default:
		vm_fault(reason, va, pc, "Invalid PTE type");
	
	}
}

static int vm_getpage(ulong reason, vaddr va, Vmap *vm, vaddr pc)
{
int n;
	if(va == 0)
		vm_fault(reason, va, pc, "Null pointer");
	if(!vm)
		vm_fault(reason, va, pc, "Stray pointer");
	n = (va-vm->vbase)/PAGE_SIZE;
	if(vm->vbase == STACKBASE) {
		if(va > thread->kstack)
			vm_fault(reason, va, pc, "Kernel stack underflow"); 
		if(va > thread->ustack) 
			vm_fault(reason, va, pc, "Stack underflow");
		if(n < MAXPTE-vm->ptlen) 		
			vm_fault(reason, va, pc, "Stack overflow");
		else
			vm_fault(reason, va, pc, "Stack fault");
	}
	if(n >= vm->ptlen)
		vm_fault(reason, va, pc, "Segmentation fault");
	return(n);
}


void vm_pagefault(ulong param, int ivec, vaddr pc, vaddr va, ulong reason)
{
Vmap *vm;
int n;
	vm = vm_getmap(va);
	n = vm_getpage(reason, va, vm, pc);
	vm_pagemap(reason, va, pc, vm, n, vm, n);
}


void vm_accv(ulong param, int ivec, vaddr pc, vaddr va, ulong reason)
{
Vmap *vm;
int n;
	vm = vm_getmap(va);
	n = vm_getpage(reason, va, vm, pc);
	vm_fault(reason, va, pc, "Access violation");
}

