/*  hd.c                                                        (CRC:4d00ef89)
 *      hard drive support for PC's
 *
 *      relative path: ./drv
 *      creation date: 1992
 *      last modified: Thu Dec 19 18:16:44 1996
 *      header update: Fri Dec 20 09:09:07 1996
 *  
 *       Device drivers
 *  
 *  Copyright (c) 1992 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is a part of the Shag/OS operating system distribution, and is
 *  free software for non-profit uses; it may be freely used and distributed
 *  for non-profit purposes under the terms of the Shag/OS License, version 1.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND, 
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 *  AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT WILL THE AUTHOR, OR
 *  ANYONE RESPONSIBLE FOR THE LEGAL MODIFICATION OR REDISTRIBUTION OF THIS
 *  SOFTWARE, BE LIABLE FOR ANY DAMAGES, INCLUDING BUT NOT LIMITED TO ANY LOST
 *  PROFITS, LOST SAVINGS OR ANY INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER
 *  RESULTING FROM IMPAIRED OR LOST DATA, SOFTWARE OR COMPUTER FAILURE OR ANY
 *  OTHER CAUSE, EVEN IF THE AUTHOR IS ADVISED OF THE POSSIBILITY OF SUCH
 *  DAMAGES, OR FOR ANY OTHER CLAIM BY THE USER OR FOR ANY THIRD PARTY CLAIM.
 *  
 *  You should have received a copy of the Shag/OS License along with this
 *  program, in a file called 'license1.txt'.  If this license is missing,
 *  please contact the author and describe how this software was obtained.
 *  
 *  The author can be reached for questions, comments, and suggestions at:
 *  
 *      Frank Barrus, 4435 River Rd. Apt #3, Scottsville, NY 14546, USA
 *      email:   shaggy@csh.rit.edu
 *  
 *      For sending bug reports, please read the file 'bugs', and always
 *      include the product name and version, and the word 'BUG' in the subject.
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */


#define KERNEL
#include <sys/types.h>
#include <sys/resource.h>
#include <error.h>
#include <stdio.h>
#include <string.h>
#include <dir.h>


typedef struct {
	ulong	start;
	ulong	nblocks;
	ushort	blocksize;
	uchar	num;
	uchar	type;
} HDinfo;


#include <mdhd.c>



static obj_t hd_init(ObjFrame *frame, HDinfo *hd, int num, ulong start, ulong nblocks)
{
	frame->data = hd = KMALLOC(HDinfo);
	frame->size = sizeof(HDinfo);
	hd->num = num;
	hd->start = start;
	hd->nblocks = nblocks;
	return(OFLAG_HASVEC|OFLAG_KERNEL);
}


static obj_t hd_readat(ObjFrame *frame, HDinfo *hd, uchar *buf, int len, int index)
{
uchar *tmp;
int n;
	tmp = (uchar*)alloca(hd->blocksize);
	n = 0;
	while(len > 0 && (index < hd->blocksize*hd->nblocks)) {
		hd_readblock(frame->data,tmp,index/hd->blocksize);
		memcpy(buf,tmp,min(hd->blocksize,len));
		buf = (uchar*)align((ulong)buf+1,hd->blocksize);
		len -= hd->blocksize;
		index += hd->blocksize;
		n += hd->blocksize;
	}
	return(n);
}

static MethodTable hd_method[] = {
	{&_init,(obj_method)hd_init},
	{&_readat,(obj_method)hd_readat}};

static ObjClass Class_HardDrive = {
	hd_method,
	NULLP(AttrTable),
	arraysize(hd_method),
	0};


static void hd_loop(void)
{
	while(1) {
		_(OBJ_THREAD, _msleep, 1000, ___, ___);
	}
}


obj_t Main(ObjFrame *frame, vaddr base, obj_t arg, obj_t env)
{
int i,n;
char str[40];
obj_t hd, pd;
obj_t class_hd;
obj_t class_object;
HDinfo partition;
HDinfo hdinfo;
	cprintf("Hard Disk Driver v0.01  Dec 1992\n");

	class_object = dir_parse(OBJ_ROOT, "/class/Object");
	class_hd = _(class_object, _class, Ptr(&Class_HardDrive), ___, ___);

	mdhd_init();
		
	for(i=0; i<NHD; i++)
	    if(read_hd_info(&hdinfo,i)) {
		hd = _(class_hd, _new, hdinfo.num, 
				hdinfo.start, hdinfo.nblocks);
		sprintf(str,"/dev/hd%d",i);
		dir_create(OBJ_ROOT,str,hd);
		cprintf("\t%s: ",str);
		n = 0;
		while(read_partition(&partition,&hdinfo,n)) {
			if(!n)
				cprintf("\n");
			sprintf(str,"/dev/hd%d%c",i,'a'+n);
			pd = _(class_hd, _new, partition.num, 
					partition.start, partition.nblocks);
			dir_create(OBJ_ROOT,str,pd);
			cprintf("\t  %s: type:%2x start:%-7d size:%-7d\n",
				str, partition.type,partition.start,
					partition.nblocks);
			n++;
		} 
		if(!n)
			cprintf("cannot read partition table\n");
	}
	_(OBJ_THREAD, _new, Label("hd"), Ptr(hd_loop), ___);
	return(OK);
}

