/*  init.c                                                      (CRC: 7655690)
 *      kernel initialization
 *
 *      relative path: ./drv
 *      creation date: 1992
 *      last modified: Thu Dec 19 18:17:15 1996
 *      header update: Fri Dec 20 09:09:08 1996
 *  
 *       Device drivers
 *  
 *  Copyright (c) 1992 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is a part of the Shag/OS operating system distribution, and is
 *  free software for non-profit uses; it may be freely used and distributed
 *  for non-profit purposes under the terms of the Shag/OS License, version 1.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND, 
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 *  AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT WILL THE AUTHOR, OR
 *  ANYONE RESPONSIBLE FOR THE LEGAL MODIFICATION OR REDISTRIBUTION OF THIS
 *  SOFTWARE, BE LIABLE FOR ANY DAMAGES, INCLUDING BUT NOT LIMITED TO ANY LOST
 *  PROFITS, LOST SAVINGS OR ANY INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER
 *  RESULTING FROM IMPAIRED OR LOST DATA, SOFTWARE OR COMPUTER FAILURE OR ANY
 *  OTHER CAUSE, EVEN IF THE AUTHOR IS ADVISED OF THE POSSIBILITY OF SUCH
 *  DAMAGES, OR FOR ANY OTHER CLAIM BY THE USER OR FOR ANY THIRD PARTY CLAIM.
 *  
 *  You should have received a copy of the Shag/OS License along with this
 *  program, in a file called 'license1.txt'.  If this license is missing,
 *  please contact the author and describe how this software was obtained.
 *  
 *  The author can be reached for questions, comments, and suggestions at:
 *  
 *      Frank Barrus, 4435 River Rd. Apt #3, Scottsville, NY 14546, USA
 *      email:   shaggy@csh.rit.edu
 *  
 *      For sending bug reports, please read the file 'bugs', and always
 *      include the product name and version, and the word 'BUG' in the subject.
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */


#include <sys/types.h>
#include <sys/resource.h>
#include <stdio.h>
#include <error.h>
#include <event.h>
#include <dir.h>


static obj_t devload(char *path);


extern label_t _stdin, _stdout, _stderr;
extern label_t _icrnl, _onlcr, _name, _echo, _echonl;
extern label_t _msleep, _exec, _wait;

#define MAXDEV 10

char devname[MAXDEV][16];
char drvname[MAXDEV][32];
obj_t thread_id[MAXDEV];

static obj_t spawn_login(int n)
{
obj_t fd;
obj_t tty;
char ttyname[24];
	sprintf(ttyname,"/dev/%s",devname[n]);
	tty = dir_parse(OBJ_ROOT, ttyname);
	if(err(tty)) {
		cprintf("%s: %s\n", ttyname, strerror(tty));
		return(ERR_NOATTR);
	}
	_(OBJ_THREAD, _set, _name, Label(drvname[n]),___);
	_(OBJ_THREAD, _set, _stdin, tty,___);
	_(OBJ_THREAD, _set, _stdout, tty,___);
	_(OBJ_THREAD, _set, _stderr, tty,___);
	err_(OBJ_STDOUT, _set, _onlcr, OBJ_TRUE, ___);
	err_(OBJ_STDIN, _set, _icrnl, OBJ_TRUE, ___);
	fd = dir_parse(OBJ_ROOT,drvname[n]);
	_(OBJ_TASK, _exec, fd, ___, ___);
	return(0);
}


static char *ogets_at(char *buf, int len, obj_t obj, int *index)
{
obj_t rv;
int n;
	rv = err_(obj, _readat, Ptr(buf), len-1, *index);
	if(err(rv) || !rv)
		return(NULL);
	n = 0;
	while(n < len-1 && buf[n] != '\n')
		n++;
	buf[n] = '\0';
	*index += (n+1);
	return(buf);
}


obj_t Main(ObjFrame *frame, vaddr base, obj_t arg, obj_t env)
{
obj_t fd;
char buf[80];
int ndev, i;
char *cp;
char *dp;
Event event;
	cprintf("Init v0.10  Oct 1992\n");

	fd = dir_parse(OBJ_ROOT,"/etc/dev");
	i = 0;
	while(ogets_at(buf,sizeof(buf),fd,&i)) {
		if(buf[0] == '#' || buf[0] <= 32)
			continue;
		devload(buf);
	}
	fd = dir_parse(OBJ_ROOT,"/etc/ttys");
	ndev = 0;
	cprintf("\n");

	i = 0;
	while(ogets_at(buf,sizeof(buf),fd,&i)) {
		if(buf[0] == '#' || buf[0] <= 32)
			continue;
		cp = buf;
		dp = devname[ndev];
		while(*cp > 32) 
			*dp++ = *cp++;
		*dp = '\0';
		while(*cp <= 32)
			cp++;
		dp = drvname[ndev];
		while(*cp > 32) 
			*dp++ = *cp++;
		*dp = '\0';
		thread_id[ndev] = _(OBJ_THREAD, _new, Label("spawn"),
				 Ptr(spawn_login), ndev);
		ndev++;
	}
	while(1) {
		_(OBJ_THREAD, _wait, EVENT_ANY, PARAM_ANY, Ptr(&event));
		for(i=0; i<ndev; i++)
			if(thread_id[i] == event.sender)
				thread_id[i] = _(OBJ_THREAD, _new,
				 Label("spawn"), Ptr(spawn_login), i);
	}
}




static obj_t devload(char *path)
{
obj_t  fd;
	cprintf("-- ");
	fd = dir_parse(OBJ_ROOT,path);
	if(err(fd)) {
		cprintf("%s: %s\n", path, strerror(fd));
		return(OK);
	}
	fd = err_(OBJ_KERNEL, _exec, fd, ___, ___);
	if(err(fd))
		cprintf("%s failed: %s\n", path, strerror(fd));
	return(OK);
}

