/*  event.h                                                     (CRC:1767da1c)
 *      relative path: ./include
 *      creation date: 1992
 *      last modified: Wed May 22 21:33:32 1996
 *      header update: Wed Dec 18 07:17:20 1996
 *  
 *       Include files
 *  
 *  Copyright (c) 1992 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is a part of the Shag/OS operating system distribution, and is
 *  free software for non-profit uses; it may be freely used and distributed
 *  for non-profit purposes under the terms of the Shag/OS License, version 1.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND, 
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 *  AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT WILL THE AUTHOR, OR
 *  ANYONE RESPONSIBLE FOR THE LEGAL MODIFICATION OR REDISTRIBUTION OF THIS
 *  SOFTWARE, BE LIABLE FOR ANY DAMAGES, INCLUDING BUT NOT LIMITED TO ANY LOST
 *  PROFITS, LOST SAVINGS OR ANY INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER
 *  RESULTING FROM IMPAIRED OR LOST DATA, SOFTWARE OR COMPUTER FAILURE OR ANY
 *  OTHER CAUSE, EVEN IF THE AUTHOR IS ADVISED OF THE POSSIBILITY OF SUCH
 *  DAMAGES, OR FOR ANY OTHER CLAIM BY THE USER OR FOR ANY THIRD PARTY CLAIM.
 *  
 *  You should have received a copy of the Shag/OS License along with this
 *  program, in a file called 'license1.txt'.  If this license is missing,
 *  please contact the author and describe how this software was obtained.
 *  
 *  The author can be reached for questions, comments, and suggestions at:
 *  
 *      Frank Barrus, 4435 River Rd. Apt #3, Scottsville, NY 14546, USA
 *      email:   shaggy@csh.rit.edu
 *  
 *      For sending bug reports, please read the file 'bugs', and always
 *      include the product name and version, and the word 'BUG' in the subject.
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

#ifndef __EVENT_H__
#define __EVENT_H__

/* high bit set means external event, with low 31 bits for value */
#define EVENT_ANY		0xffffffff
#define EVENT_CHILDTERM		0x00010000
#define EVENT_CHILDSTOP		0x00020000
#define EVENT_BUFIN		0x00100000
#define EVENT_BUFOUT		0x00200000
#define EVENT_LOCK		0x00400000
#define EVENT_SUSPEND		0x01000000
#define EVENT_WAKEUP   		0x02000000
#define EVENT_QUIT		0x04000000
#define EVENT_TERMINATE		0x08000000
#define EVENT_TRAP		0x80000000


#define PARAM_ANY		0xffffffff

/* trap events: */
#define TRAP_FAULT		0x0100	/* general fault */
#define TRAP_BKPT		0x0200	/* breakpoint */
#define TRAP_RESVOP		0x0300	/* reserved operation */
#define TRAP_RESVADR		0x0400	/* reserved addressing mode */
#define TRAP_PRIVINST		0x0500	/* privileged instruction */
#define TRAP_KSTACK		0x0600	/* invalid kernel stack */
#define TRAP_TRACE		0x0700	/* trace pending */
#define TRAP_ARITH		0x0800	/* arithmetic fault */
#define TRAP_DOUBLE		0x0900	/* double fault */
#define TRAP_INVTASK		0x0a00	/* invalid task */
#define TRAP_ACCV		0x1000	/* access violation */
#define TRAP_SEGV		0x1100	/* segmentation fault */
#define TRAP_PGFAULT		0x1200	/* page fault */
#define	TRAP_INVALID		0x1300	/* invalid interrupt */

/* arithmetic traps: */
#define ARITH_INTOVF		0x01	/* integer overflow */
#define ARITH_INTDIV0		0x02	/* integer divide by 0 */
#define ARITH_FPOVF		0x03	/* float overflow */
#define ARITH_FPDIV0		0x04	/* float divide by 0 */
#define ARITH_FPUNDF		0x05	/* float underflow */
#define ARITH_DECOVF		0x06	/* decimal overflow */
#define ARITH_RANGE		0x07	/* subscript range */
#define ARITH_FFPOVF		0x08	/* fault: float overflow */
#define ARITH_FFPDIV0		0x09	/* fault: float divide by 0 */
#define ARITH_FFPUNDF		0x0a	/* fault: float underflow */


#endif /* __EVENT_H__ */

