/*  exec.h                                                      (CRC:cecad51b)
 *      relative path: ./include/sys
 *      creation date: 1992
 *      last modified: Wed May 22 21:41:05 1996
 *      header update: Wed Dec 18 07:17:44 1996
 *  
 *       Include files
 *        System Include files
 *  
 *  Copyright (c) 1992 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is a part of the Shag/OS operating system distribution, and is
 *  free software for non-profit uses; it may be freely used and distributed
 *  for non-profit purposes under the terms of the Shag/OS License, version 1.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND, 
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 *  AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT WILL THE AUTHOR, OR
 *  ANYONE RESPONSIBLE FOR THE LEGAL MODIFICATION OR REDISTRIBUTION OF THIS
 *  SOFTWARE, BE LIABLE FOR ANY DAMAGES, INCLUDING BUT NOT LIMITED TO ANY LOST
 *  PROFITS, LOST SAVINGS OR ANY INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER
 *  RESULTING FROM IMPAIRED OR LOST DATA, SOFTWARE OR COMPUTER FAILURE OR ANY
 *  OTHER CAUSE, EVEN IF THE AUTHOR IS ADVISED OF THE POSSIBILITY OF SUCH
 *  DAMAGES, OR FOR ANY OTHER CLAIM BY THE USER OR FOR ANY THIRD PARTY CLAIM.
 *  
 *  You should have received a copy of the Shag/OS License along with this
 *  program, in a file called 'license1.txt'.  If this license is missing,
 *  please contact the author and describe how this software was obtained.
 *  
 *  The author can be reached for questions, comments, and suggestions at:
 *  
 *      Frank Barrus, 4435 River Rd. Apt #3, Scottsville, NY 14546, USA
 *      email:   shaggy@csh.rit.edu
 *  
 *      For sending bug reports, please read the file 'bugs', and always
 *      include the product name and version, and the word 'BUG' in the subject.
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

#ifndef __EXEC_H__
#define __EXEC_H__

#include <sys/types.h>


struct exec {
  ushort a_magic;	/* magic number */
  ushort a_machtype;	/* machine type */
  ulong a_text;		/* length of text, in bytes */
  ulong a_data;		/* length of data, in bytes */
  ulong a_bss;		/* length of uninitialized data area for file, in bytes */
  ulong a_syms;		/* length of symbol table data in file, in bytes */
  ulong a_entry;	/* start address */
  ulong a_trsize;	/* length of relocation info for text, in bytes */
  ulong a_drsize;	/* length of relocation info for data, in bytes */
};


/* Code indicating object file or impure executable  */
#define OMAGIC 0407
/* Code indicating pure executable  */
#define NMAGIC 0410
/* Code indicating demand-paged executable  */
#define ZMAGIC 0413
/* Code indicating core file  */
#define CMAGIC 0421
/* Code indicating library file */
#define LMAGIC 0430
/* Code indicating demand-paged executable with header in text */
#define QMAGIC 0314

struct nlist {
	ulong	n_strx;
	uchar	n_type;
	char 	n_other;
	ushort 	n_desc;
	ulong 	n_value;
};

#define N_UNDF	0
#define N_ABS	2
#define N_TEXT	4
#define N_DATA	6
#define N_BSS	8
#define N_COMM	0x12	/* common - ld internal */
#define N_LABEL	0x14	/* label - ld internal */
#define N_FN	0x1f	/* file name symbol */

#define N_EXT	1
#define N_TYPE	0x1e	/* type mask */

#define N_SHARE	0x20	/* global sym -- register in /lib */

#define N_STAB	0xe0	/* if any bits set, an SDB entry */
#define	N_INDR	0xa	/* indirect reference immediately follows */
#define N_SETA	0x14
#define N_SETT	0x16
#define N_SETD	0x18
#define N_SETB	0x1a
#define N_SETV	0x1c

struct relocation_info {
	ulong	r_address;
	unsigned r_symbolnum:24;
	unsigned r_pcrel:1;
	unsigned r_length:2;
	unsigned r_extern:1;
	unsigned r_pad:4;
};

#endif /* __EXEC_H__ */

