/*  fcntl.c                                                     (CRC:310d3020)
 *      relative path: ./posix
 *      creation date: 1992
 *      last modified: Wed May 22 21:36:18 1996
 *      header update: Wed Dec 18 07:17:55 1996
 *  
 *       POSIX standard library
 *  
 *  Copyright (c) 1992 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is a part of the Shag/OS operating system distribution, and is
 *  free software for non-profit uses; it may be freely used and distributed
 *  for non-profit purposes under the terms of the Shag/OS License, version 1.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND, 
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 *  AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT WILL THE AUTHOR, OR
 *  ANYONE RESPONSIBLE FOR THE LEGAL MODIFICATION OR REDISTRIBUTION OF THIS
 *  SOFTWARE, BE LIABLE FOR ANY DAMAGES, INCLUDING BUT NOT LIMITED TO ANY LOST
 *  PROFITS, LOST SAVINGS OR ANY INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER
 *  RESULTING FROM IMPAIRED OR LOST DATA, SOFTWARE OR COMPUTER FAILURE OR ANY
 *  OTHER CAUSE, EVEN IF THE AUTHOR IS ADVISED OF THE POSSIBILITY OF SUCH
 *  DAMAGES, OR FOR ANY OTHER CLAIM BY THE USER OR FOR ANY THIRD PARTY CLAIM.
 *  
 *  You should have received a copy of the Shag/OS License along with this
 *  program, in a file called 'license1.txt'.  If this license is missing,
 *  please contact the author and describe how this software was obtained.
 *  
 *  The author can be reached for questions, comments, and suggestions at:
 *  
 *      Frank Barrus, 4435 River Rd. Apt #3, Scottsville, NY 14546, USA
 *      email:   shaggy@csh.rit.edu
 *  
 *      For sending bug reports, please read the file 'bugs', and always
 *      include the product name and version, and the word 'BUG' in the subject.
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

/* fcntl.c
 *   posix compliant file control functions
 */

#include <sys/types.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/resource.h>
#include <error.h>
#include <fcntl.h>
#include <dir.h>


extern label_t _size, _pos;


obj_t open(char *path, int flags, ...)
{
obj_t fd,sd,rv;
char buf[MAXLABEL];
int size;
	fd = dir_parse(OBJ_CWD, path);
	if(err(fd)) {
		if(flags & O_CREAT) {
			sd = dir_parse_path(OBJ_CWD, path, buf, FALSE);
			fd = dir_parse(OBJ_ROOT,"/class/File");
			fd = _(fd,_new,___,___,___);
			_(sd, _create, Label(buf), fd, ___);
		} else
			return(fd);
	} 
	if(flags&O_TRUNC)
		err_(fd,_set,_size,0,___);
	rv = err_(fd,_write,Ptr(NULL),0,___);
	if(err(rv)) {
		sd = dir_parse(OBJ_ROOT,"/class/Stream");
		sd = _(sd,_new,fd,___,___);
	} else {
		sd = fd;
	}
	if(flags&O_APPEND) {
		size = err_(fd, _get, _size, ___, ___);
		if(!err(size))
			_(sd,_set,_pos, size,___);
	}
	return(sd);
}


obj_t creat(char *path, int mode)
{
	return(open(path, O_WRONLY|O_CREAT|O_TRUNC, mode));
}


status_t close(obj_t fd)
{
	_(fd,_free,___,___,___);
	return(0);
}


off_t lseek(obj_t fd, off_t offset, int whence)
{
	switch(whence) {
	case SEEK_SET:
		err_(fd,_set,_pos, offset, ___);
		break;
	case SEEK_CUR:
		offset += err_(fd,_get,_pos,___,___); 
		err_(fd,_set,_pos, offset, ___);
		break;
	case SEEK_END:
		offset += err_(fd,_get,_size,___,___); 
		err_(fd,_set,_pos, offset, ___);
		break;
	}
	return(offset);
}

int read(int fildes, void *buf, ulong nbytes)
{
	return(err_(fildes, _read, Ptr(buf), nbytes, ___));
}

int write(int fildes, const void *buf, ulong nbytes)
{
	return(err_(fildes, _write, Ptr(buf), nbytes, ___));
}

obj_t dup(obj_t fd)
{
	printf("dup %x\n", fd);
	return(fd);
}
