/*  fs_static.c                                                 (CRC:d310f840)
 *      relative path: ./sys
 *      creation date: 1992
 *      last modified: Wed Dec 18 07:18:08 1996
 *      header update: Wed Dec 18 07:18:08 1996
 *  
 *       Microkernel system files
 *  
 *  Copyright (c) 1992 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is a part of the Shag/OS operating system distribution, and is
 *  free software for non-profit uses; it may be freely used and distributed
 *  for non-profit purposes under the terms of the Shag/OS License, version 1.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND, 
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 *  AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT WILL THE AUTHOR, OR
 *  ANYONE RESPONSIBLE FOR THE LEGAL MODIFICATION OR REDISTRIBUTION OF THIS
 *  SOFTWARE, BE LIABLE FOR ANY DAMAGES, INCLUDING BUT NOT LIMITED TO ANY LOST
 *  PROFITS, LOST SAVINGS OR ANY INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER
 *  RESULTING FROM IMPAIRED OR LOST DATA, SOFTWARE OR COMPUTER FAILURE OR ANY
 *  OTHER CAUSE, EVEN IF THE AUTHOR IS ADVISED OF THE POSSIBILITY OF SUCH
 *  DAMAGES, OR FOR ANY OTHER CLAIM BY THE USER OR FOR ANY THIRD PARTY CLAIM.
 *  
 *  You should have received a copy of the Shag/OS License along with this
 *  program, in a file called 'license1.txt'.  If this license is missing,
 *  please contact the author and describe how this software was obtained.
 *  
 *  The author can be reached for questions, comments, and suggestions at:
 *  
 *      Frank Barrus, 4435 River Rd. Apt #3, Scottsville, NY 14546, USA
 *      email:   shaggy@csh.rit.edu
 *  
 *      For sending bug reports, please read the file 'bugs', and always
 *      include the product name and version, and the word 'BUG' in the subject.
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

/* fs_static.c
 */

#define KERNEL
#include <sys/types.h>
#include <sys/resource.h>
#include <stdio.h>
#include <sys/global.h>
#include <object.h>
#include <error.h>
#include <dir.h>


extern StaticFS sfs_base[];
extern long sfs_nentries;
extern long sfs_size;


extern obj_t _size;


static obj_t static_init(ObjFrame *frame, StaticFS *f, int index)
{
	frame->data = &sfs_base[index];
	frame->size = sizeof(StaticFS);
	return(OFLAG_HASVEC|OFLAG_KERNEL);
}

static obj_t static_readat(ObjFrame *frame, StaticFS *f,
			 uchar *buf, int len, int index)
{
	if(!len)
		return(0);
	if(index+len > f->size)
		len = f->size - index;
	if(len <= 0)
		return(0);
	memcpy(buf,&f->data[index],len);
	return(len);
}


static MethodTable static_method[] = {
	{&_init,(obj_method)static_init},
	{&_readat,(obj_method)static_readat}};

static AttrTable static_attr[] = {
	{&_size,offsetof(StaticFS,size),ATYPE_ULONG|ATYPE_NOWRITE}};

static ObjClass Class_StaticFile = {
	static_method,
	static_attr,
	arraysize(static_method),
	arraysize(static_attr)};


void nuke_fs(void)
{
	kprintf("nuking fs: (%x:%x)  ", sfs_base, sfs_size);
	ksfree(sfs_base, sfs_size);
	kprintf("done\n");
}


void StaticFile_init(void)
{
int i;
obj_t td,wd,fd;
char buf[50];
	kprintf("sfs_init(): sfs_size=%d sfs_nentries=%d ", 	
		sfs_size, sfs_nentries);
	for(i=0; i<sfs_nentries; i++) {
		sfs_base[i].name += (ulong)sfs_base;
		sfs_base[i].data += (ulong)sfs_base;
	}
	class_staticfile = 
		_(class_object, _class, Ptr(&Class_StaticFile), ___, ___);

	td = dir_parse_path(kinfo.Root,"/usr/tmp/",buf,TRUE);
	for(i=0; i<sfs_nentries; i++) {
		wd = dir_parse_path(td,sfs_base[i].name,buf,TRUE);
		fd = _(class_staticfile, _new, i, ___, ___);
		_(wd,_create,Label(buf),fd,___);
	}

	kprintf("done\r\n");
}

