/*  $Id: mkernel.c,v 1.1 1996/12/11 01:12:57 shaggy Exp shaggy $
 *      micro-kernel initialization
 *      relative path: ./sys
 *  
 *       Microkernel system files
 *  
 *  Copyright (c) 1992,1996 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is a part of the Shag/OS operating system distribution, and is
 *  free software for non-profit uses; it may be freely used and distributed
 *  for non-profit purposes under the terms of the Shag/OS License, version 1.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND, 
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 *  AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT WILL THE AUTHOR, OR
 *  ANYONE RESPONSIBLE FOR THE LEGAL MODIFICATION OR REDISTRIBUTION OF THIS
 *  SOFTWARE, BE LIABLE FOR ANY DAMAGES, INCLUDING BUT NOT LIMITED TO ANY LOST
 *  PROFITS, LOST SAVINGS OR ANY INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER
 *  RESULTING FROM IMPAIRED OR LOST DATA, SOFTWARE OR COMPUTER FAILURE OR ANY
 *  OTHER CAUSE, EVEN IF THE AUTHOR IS ADVISED OF THE POSSIBILITY OF SUCH
 *  DAMAGES, OR FOR ANY OTHER CLAIM BY THE USER OR FOR ANY THIRD PARTY CLAIM.
 *  
 *  You should have received a copy of the Shag/OS License along with this
 *  program, in a file called 'license1.txt'.  If this license is missing,
 *  please contact the author and describe how this software was obtained.
 *  
 *  The author can be reached for questions, comments, and suggestions at:
 *  
 *      Frank Barrus, 4435 River Rd. Apt #3, Scottsville, NY 14546, USA
 *      email:   shaggy@csh.rit.edu
 *  
 *      For sending bug reports, please read the file 'bugs', and always
 *      include the product name and version, and the word 'BUG' in the subject.
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

/*  $Log: mkernel.c,v $
 *  Revision 1.1  1996/12/11  01:12:57  shaggy
 *  Initial revision
 *
 *      creation date: 1992
 */



#define KERNEL
#define GLOBAL
#include <sys/types.h>
#include <sys/resource.h>
#include <sys/global.h>
#include <object.h>
#include <error.h>
#include <stdio.h>
#include <dir.h>

#include <boot/libsolo.h>


Thread *thread = NULLP(Thread);
Thread *null_thread = NULLP(Thread);



status_t call_init(int param);
extern ulong etext[], edata[], end[], sfs_base[];

ulong kstack_data[KSTACK_SIZE];

/* tell the runtime code (srt0) where our stack is, so it doesn't
 * try to allocate another one
 */
ulong* initial_stack_top =
	kstack_data+sizeof(kstack_data)/sizeof(*kstack_data);

extern label_t _msleep, _suspend, _exec;


int
main(int argc, char **argv)
{
boolean b;
/* it can safely be assumed that all interrupts are disabled
 * when we get here...
 */
	vm_init();
	panic_event.event = 0;		/* clear panic event */
	thread_context = (Context*)kstack_data;
	kinfo.Root = 0;			/* no root object yet */
	kernel.ref = 1;			/* always a reference,
					 * unless null/0 is killed */
	qclear((Queue*)&eventq);

/* initialize system */
	interrupt_init();	/* set up interrupt tables */
	system_init();		/* initialize other machine specific parts */

/* get us a console to print to and announce our presence */
	console_init();
	kprintf("\nShag/OS Classic v%d.%02d\n\n", VER_MAJOR, VER_MINOR);
	kprintf("machine=%s phys_mem=%d\n", MACHINE, 0);   /* XXX fix size! */
	kprintf("machparam= %s\n", machparam);
	kprintf("%s@%s: %s\n\n", mkuser,mkhost,mkdate);
	kprintf("etext=%x edata=%x end=%x sfs_base=%x\n",
		etext, edata, end, sfs_base);
	bootparam_print();

	
	Name_init();		/* name services */
	Object_init();		/* object-oriented calls */
	Dir_init();		/* directory services */
	Buffer_init();		/* stream buffers */
	Thread_init();		/* threads */
	Task_init();		/* tasks */
	StaticFile_init();	/* static files */
	event_init();		/* events */
	dir_register();		/* register kernel functions */

	kprintf("MicroKernel initialized -- enabling interrupts...\n");
/* turn on interrupts and clock */
	CSWITCH_OFF(b)
	interrupt_enable();
	clock_init();		
	CSWITCH_ON(b)

	kprintf("Loading init:\n");
	_(OBJ_THREAD, _new, Label("init"), Ptr(call_init), ___);

	while(1) 
		_(OBJ_THREAD, _suspend, ___, ___, ___);
/*
		waitfor(EVENT_CHILDTERM, PARAM_ANY);
		kprintf("Init killed! Restarting...\n");
		thread_spawn((vaddr)call_init,Label("init"),0);
	}
*/
}


status_t call_init(int param)
{
obj_t td,fd,rv;
	fd = dir_parse(kinfo.Root,"/drv/init");
	if(err(fd)) {
		print_error("/drv/init", fd);
		return(fd); 
	}
	td = _(class_task,_new,___,___,___);
	rv = _(td,_exec,fd,___,___);
	_(td,_free,___,___,___);
	if(err(rv)) {
		kprintf("init failed: %s\n", strerror(rv));
		halt();
	}
	return(OK);
}

