/*  $Id: task.c,v 1.1 1996/12/20 05:01:56 shaggy Exp shaggy $
 *      task management
 *      
 *      relative path: sys
 *  
 *      Microkernel system files
 *  
 *  Copyright (c) 1992,1996 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is a part of the Shag/OS operating system distribution, and is
 *  free software for non-profit uses; it may be freely used and distributed
 *  for non-profit purposes under the terms of the Shag/OS License, version 1.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND, 
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 *  AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT WILL THE AUTHOR, OR
 *  ANYONE RESPONSIBLE FOR THE LEGAL MODIFICATION OR REDISTRIBUTION OF THIS
 *  SOFTWARE, BE LIABLE FOR ANY DAMAGES, INCLUDING BUT NOT LIMITED TO ANY LOST
 *  PROFITS, LOST SAVINGS OR ANY INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER
 *  RESULTING FROM IMPAIRED OR LOST DATA, SOFTWARE OR COMPUTER FAILURE OR ANY
 *  OTHER CAUSE, EVEN IF THE AUTHOR IS ADVISED OF THE POSSIBILITY OF SUCH
 *  DAMAGES, OR FOR ANY OTHER CLAIM BY THE USER OR FOR ANY THIRD PARTY CLAIM.
 *  
 *  You should have received a copy of the Shag/OS License along with this
 *  program, in a file called 'license1.txt'.  If this license is missing,
 *  please contact the author and describe how this software was obtained.
 *  
 *  The author can be reached for questions, comments, and suggestions at:
 *  
 *      Frank Barrus, 4435 River Rd. Apt #3, Scottsville, NY 14546, USA
 *      email:   shaggy@csh.rit.edu
 *  
 *      For sending bug reports, please read the file 'bugs', and always
 *      include the product name and version, and the word 'BUG' in the subject.
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

/*  $Log: task.c,v $
 *  Revision 1.1  1996/12/20  05:01:56  shaggy
 *  Initial revision
 *
 *      creation date: 1992
 */


#define KERNEL
#include <sys/types.h>
#include <sys/resource.h>
#include <sys/global.h>
#include <error.h>
#include <sys/exec.h>
#include <dir.h>


struct reloc {
	vaddr	base;		/* base for current reloc (tbase or dbase) */
	ulong	tbase;		/* start of text segment */
	ulong	dbase;		/* start of data segment */
	ulong	bbase;		/* start of bss segment */
	ulong	cursize;	/* size of current reloc (tsize or dsize) */
	ulong	bsize;		/* size of bss segment */
	ulong	blast;		/* last used bss address */
	ulong	dstart;		/* start of data segment in file */
	ulong	bstart;		/* start of bss segment in file */
	int	nsym;		/* number of symbols */
	int	nreloc;		/* number of relocation entries */
	int	entry;		/* execution entry point */
	struct nlist *sym;	/* symbol table */
	struct relocation_info *reloc;	/* relocation table */
	vaddr	temp_base;	/* temp shared area for text or data */
	vaddr	temp_bss;	/* temp shared area for bss */
	ulong	strsize;	/* size of string data area */
	char	*str;		/* string data area */
	boolean valid;		/* was reloc successful? */
	boolean undef;		/* were there any unresolved symbols? */
	obj_t	libd;		/* library function table descriptor */
	int	nlib;		/* number of libraries to look in */
	obj_t	lib[4];		/* shared libraries to check */
	ushort	magic;		/* magic number */
};

extern label_t _data, _exec, _load, _loadlib, _task, _mkdir, _self, _kernel;

static status_t task_load_data(Task *task, obj_t fd, int pos, vaddr adr,
							 int size);
static status_t task_reloc(Task *task, struct reloc *reloc);
static int get_bss_size(struct reloc *reloc);


static ZoneP task_zone;




static vaddr task_load(Task *task, vaddr *entry, obj_t fd)
{
struct exec header;
uchar buf[1024];
vaddr base,adr;
int n, pos;
struct relocation_info *t_reloc, *d_reloc;
struct reloc reloc;
char *str;
ulong accmode;
	n = _(fd,_readat,Ptr(&header),sizeof(struct exec),0);
	if(err(n)) return(n);
	if (n < sizeof(struct exec))
		return((status_t)ERR_SHORTFILE);
	pos = n;
	accmode = VM_URW;
	if(header.a_magic == LMAGIC) 
		task = &kernel;	/* XXX - BIG security hole */
	else if(task == &kernel)
		accmode = VM_KRW;
	if(header.a_magic == ZMAGIC || header.a_magic == NMAGIC
			|| header.a_magic == OMAGIC || header.a_magic == LMAGIC) {
		base = vm_alloc(&task->data, 
			align(header.a_text,PAGE_SIZE) + header.a_data, accmode);
		if(header.a_magic != OMAGIC && header.a_magic != LMAGIC) {
			n = _(fd,_readat,Ptr(buf),1024-sizeof(struct exec),pos);
			if(err(n)) return(n);
			if (n < 1024-sizeof(struct exec))
				return((status_t)ERR_SHORTFILE);
			pos = 1024;
		}
		pos = task_load_data(task,fd,pos,base,header.a_text);
		if(err(pos)) return(pos);
		pos = task_load_data(task,fd,pos,
			align(base+header.a_text,PAGE_SIZE),header.a_data);
		if(err(pos)) return(pos);
	} else
		base = 0;
	if(header.a_magic == ZMAGIC || header.a_magic == NMAGIC) {
		kprintf("error: loading non-relocatable file\n");
		if(header.a_bss)
			vm_alloc(&task->data, header.a_bss, VM_URW);
		vm_setaccess(&task->data,0,header.a_text,VM_URKW);
		if(entry)
			*entry = base;
	} else if(header.a_magic == OMAGIC || header.a_magic == LMAGIC) {
		if(header.a_magic == LMAGIC) 
			reloc.libd = dir_parse_path(OBJ_ROOT,"/lib/posix/",(char*)buf,TRUE);
		else
			reloc.libd = OBJ_NULL;
		reloc.magic = header.a_magic;
		reloc.sym = (struct nlist *)kmalloc(header.a_syms);
		/* XXX - had to add 8 to the allocation size because GCC
		 * sometimes generates code that accesses the last byte
		 * of the structure with a longword, so it overlaps by 3
		 * bytes into the next page */
		t_reloc = (struct relocation_info *)kmalloc(header.a_trsize+8);
		d_reloc = (struct relocation_info *)kmalloc(header.a_drsize+8);
		n = _(fd,_readat,Ptr(t_reloc),header.a_trsize,pos);
		if(err(n)) return(n);
		pos +=n ;
		n = _(fd,_readat,Ptr(d_reloc),header.a_drsize,pos);
		if(err(n)) return(n);
		pos +=n ;
		n = _(fd,_readat,Ptr(reloc.sym),header.a_syms,pos);
		if(err(n)) return(n);
		if(err(n)) return(n);
		pos +=n ;
		n = _(fd,_readat,Ptr(&reloc.strsize),4,pos);
		if(err(n)) return(n);
		pos +=n ;
		reloc.str = (char*)kmalloc(reloc.strsize);
		n = _(fd,_readat,Ptr(&reloc.str[4]),reloc.strsize-4,pos);
		if(err(n)) return(n);
		pos +=n ;

		reloc.valid = TRUE;
		reloc.undef = FALSE;
		reloc.tbase = base;
		reloc.dbase = align(reloc.tbase+header.a_text, PAGE_SIZE);
		reloc.bbase = reloc.dbase+header.a_data;
		reloc.bsize = header.a_bss;
		reloc.dstart  = header.a_text;
		reloc.bstart = reloc.dstart+header.a_data;
		reloc.nsym = header.a_syms/sizeof(struct nlist);

		reloc.nlib = 0;
		reloc.lib[reloc.nlib] = dir_parse(kinfo.Root,"/lib/posix");
		if(!err(reloc.lib[reloc.nlib]))
			reloc.nlib++;
		reloc.lib[reloc.nlib] = dir_parse(kinfo.Root,"/lib/mkern");
		if(!err(reloc.lib[reloc.nlib]))
			reloc.nlib++;

		reloc.blast = reloc.bsize;
		reloc.bsize += get_bss_size(&reloc);
		/* get more memory if BSS can't be contained in data area: */
		if((reloc.bbase+reloc.bsize) > align(reloc.bbase,PAGE_SIZE)) 
			reloc.bbase = vm_alloc(&task->data,
					 reloc.bsize, accmode);
/*
		kprintf("text: %x,%x\n", reloc.tbase, header.a_text);
		kprintf("data: %x,%x\n", reloc.dbase, header.a_data);
		kprintf("bss:  %x,%x(%x)\n", reloc.bbase, 
						n, reloc.bsize);
*/

		reloc.base = reloc.tbase;
		reloc.cursize = header.a_text;
		reloc.reloc = t_reloc;
		reloc.nreloc = header.a_trsize/sizeof(struct relocation_info);
		task_reloc(task,&reloc);

		reloc.base = reloc.dbase;
		reloc.cursize = header.a_data; 
		reloc.reloc = d_reloc;
		reloc.nreloc = header.a_drsize/sizeof(struct relocation_info);
		task_reloc(task,&reloc);

		if(entry) {
			reloc.entry = -1;
			for(n=0; n<reloc.nsym; n++) {
				str = &reloc.str[reloc.sym[n].n_strx];
				if(!strcmp(str, "_main"))
					reloc.entry = reloc.sym[n].n_value;
				if(!strcmp(str, "_Main"))
					reloc.entry = reloc.sym[n].n_value;
				if(reloc.sym[n].n_type & N_SHARE)  {
					adr = reloc.sym[n].n_value;
					switch(reloc.sym[n].n_type & N_TYPE) {
					case N_TEXT: adr += reloc.tbase; break;
					case N_DATA: adr += reloc.dbase; break;
					case N_BSS: adr += reloc.bbase; break;
					}
					_(reloc.libd, _create, Label(&str[1]),
						adr, ___);
				}
			}
		}
		
		if(reloc.undef)
			kprintf("\n");
		if(entry) {
			if(reloc.entry < 0) {
				kprintf("error: No Main()\n");
				return(ERR_FAILURE);
			}
			*entry = reloc.tbase + reloc.entry;
		}
		if(!reloc.valid || reloc.undef) 
			return(ERR_FAILURE);

		kfree(reloc.sym);
		kfree(t_reloc);
		kfree(d_reloc);
		kfree(reloc.str);

		if(task != &kernel || header.a_magic==LMAGIC)
			vm_setaccess(&task->data,base,header.a_text,VM_UR);
	} else
		return((status_t)ERR_INVFILE);
	return(base);
}


static int get_bss_size(struct reloc *reloc)
{
int i;
struct nlist *sp;
char *str;
int n;	
/* XXX - may be including symbols that never get referenced... */
	n = 0;
	for(i=0; i<reloc->nsym; i++) {
		sp = &reloc->sym[i];
		switch(sp->n_type & N_TYPE) {
		case N_UNDF:
			if(!sp->n_value) {
				str = &reloc->str[sp->n_strx];
				if(str[0]=='_' && str[1]=='_') 
					n += 4;
				break;
			}
			n += sp->n_value;
			break;
		}
	}
	return(n);
}


static status_t sym_lookup(Task *task, struct reloc *reloc, int symnum) 
{
struct nlist *sp;
int i, n;
char *str;
obj_t fd;
extern void __main(void);
	sp = &reloc->sym[symnum];
	switch(sp->n_type & N_TYPE) {
	case N_UNDF:
		if(!sp->n_value) {
			str = &reloc->str[sp->n_strx];
			if(!strcmp(str,"___main")) {
				n = (int)__main;
				sp->n_value = n;
				sp->n_type = N_ABS | N_EXT;
				break;
			} else if(str[0]=='_' && str[1]=='_') {
				reloc->blast = align(reloc->blast,4);
				sp->n_type = N_LABEL; 
				sp->n_value = reloc->blast;
				reloc->blast += 4;
				n = reloc->bbase + sp->n_value;
				fd = Label(&str[2]);
				*(ulong*)(reloc->temp_bss+sp->n_value) = fd;
				break;
			} else {
				n = 0;
				while(n < reloc->nlib) {
					if(str[0] != '_') {
						char tmp[80];
						tmp[0] = '.';
						strcpy(&tmp[1],str);
						fd  = err_(reloc->lib[n],_get,
						Label(tmp),___,___);
					} else	
						fd  = err_(reloc->lib[n],_get,
						Label(&str[1]),___,___);
					if(!err(fd)) {
						n = fd;
						sp->n_value = n;
						sp->n_type = N_ABS | N_EXT;
						return(n);
					}
					n++;
				}
			} 
			if(!reloc->undef) { 
				kprintf("Undefined: %s", str);
				reloc->undef = TRUE;
			} else
				kprintf(", %s", str);
			n = 0;
			reloc->valid = FALSE;
			sp->n_value = n;
			sp->n_type = N_ABS;
			break;
		} 
		sp->n_type = N_COMM;
		i = sp->n_value;
		reloc->blast = align(reloc->blast,4);
		sp->n_value = reloc->blast;
		reloc->blast += i;
		n = reloc->bbase + sp->n_value;
		break;
	case N_ABS:
		n = sp->n_value;
		break;
	case N_TEXT:
		n = reloc->tbase + sp->n_value;
		break;
	case N_DATA:
		n = reloc->dbase + sp->n_value;
		break;
	case N_BSS:
	case N_COMM:
	case N_LABEL:
		n = reloc->bbase + sp->n_value;
		break;
	default:
		kprintf(" %s(invalid symtype %x)", &reloc->str[sp->n_strx],
			sp->n_type);
		reloc->valid = FALSE;
		n = 0;
	}
	return(n);
}


static status_t task_reloc(Task *task, struct reloc *reloc)
{
int i, s;
ulong v,v_orig;
struct relocation_info *r;
	reloc->temp_base = vm_temp_share(&task->data,reloc->base,
						reloc->cursize,VM_KRW);
	reloc->temp_bss = vm_temp_share(&task->data,reloc->bbase,
						reloc->bsize,VM_KRW);
	for(i=0; i<reloc->nreloc; i++) {
		r = &reloc->reloc[i];
		switch(r->r_length) {
		case 0:
			v = *(uchar*)(reloc->temp_base+r->r_address);
			s = 1;
			break;
		case 1:
			v = *(ushort*)(reloc->temp_base+r->r_address);
			s = 2;
			break;
		case 2:
			v = *(ulong*)(reloc->temp_base+r->r_address);
			s = 4;
			break; 
		default:
			v = 0;
			s = 0;
			kprintf("Invalid data size\n");
			reloc->valid = FALSE;
		}
		if(r->r_pcrel)
			v += (reloc->base+r->r_address+s);
		v_orig = v;
		if(r->r_extern) {
			v += sym_lookup(task,reloc,r->r_symbolnum);
		} else {
			switch(r->r_symbolnum & ~1) {
			case N_TEXT:
				v += reloc->tbase; break;
			case N_DATA:
				v += (reloc->dbase-reloc->dstart); break;
			case N_BSS:
				v += (reloc->bbase-reloc->bstart); break;
			case N_ABS:
				break;
			default:
				kprintf("Invalid reloc code: %x base=%x/%x/%x\n",
					r->r_symbolnum, reloc->base,
r->r_address, v );
				reloc->valid = FALSE;
			}
		}
		if(r->r_pcrel)
			v -= (r->r_address+reloc->base+reloc->tbase+s);
		switch(r->r_length) {
		case 0:
			*(uchar*)(reloc->temp_base+r->r_address) = v;
			break;
		case 1:
			*(ushort*)(reloc->temp_base+r->r_address) = v;
			break;
		case 2:
			*(ulong*)(reloc->temp_base+r->r_address) = v;
			break;
		}
	}
	vm_temp_free(reloc->temp_base);
	vm_temp_free(reloc->temp_bss);
	return(OK);
}


static status_t task_load_data(Task *task, obj_t fd, int pos, vaddr adr, 
								int size)
{
int n;
vaddr temp;
	temp = vm_temp_share(&task->data,adr,size,VM_KRW);
	n = _(fd,_readat,Ptr(temp),size,pos);
	vm_temp_free(temp);
	if(n < size)
		return((status_t)ERR_SHORTFILE);	
	pos += n;
	return(pos);
}


status_t sym_getbyvalue(obj_t fd, ulong value, char *buf, int buflen)
{
struct exec header;
struct nlist sym;
int symbase, strbase;
int nsyms;
int n;
ulong val,strx;
	_(fd, _readat, Ptr(&header), sizeof(struct exec), 0);
	symbase = sizeof(header)+header.a_text+header.a_data+header.a_trsize+header.a_drsize;
	strbase = symbase+header.a_syms;
	nsyms = header.a_syms/sizeof(struct nlist);
	val = 0;
	strx = 4;
	n = 0;
	while(n < nsyms) {
		_(fd, _readat, Ptr(&sym), sizeof(struct nlist),
			symbase+n*sizeof(struct nlist));
		if(sym.n_value < value && sym.n_value >= val) {
			val = sym.n_value;
			strx = sym.n_strx;
		}
		n++;
	}
	err_(fd, _readat, Ptr(buf), buflen, strbase+strx);
	return(val);
}


int sym_getbyname(obj_t fd, char *str)
{
struct exec header;
struct nlist sym;
int symbase, strbase;
int nsyms;
int n;
char buf[128];
	_(fd, _readat, Ptr(&header), sizeof(header), 0);
	symbase = sizeof(header)+header.a_text+header.a_data+header.a_trsize+header.a_drsize;
	strbase = symbase+header.a_syms;
	nsyms = header.a_syms/sizeof(struct nlist);
	n = 0;
	while(n < nsyms) {
		_(fd, _readat, Ptr(&sym), sizeof(struct nlist),
			symbase+n*sizeof(struct nlist));
		_(fd, _readat, Ptr(buf), sizeof(buf), strbase+sym.n_strx);
		if(!strcmp(buf,str)) {
			return(sym.n_value);
		}
		n++;
	}
	return(ERR_FAILURE);
}


void sym_register(obj_t fd, obj_t wd)
{
struct exec header;
struct nlist sym;
int symbase, strbase;
int nsyms;
int n;
char buf[128];
	_(fd, _readat, Ptr(&header), sizeof(header), 0);
	symbase = sizeof(header)+header.a_text+header.a_data+header.a_trsize+header.a_drsize;
	strbase = symbase+header.a_syms;
	nsyms = header.a_syms/sizeof(struct nlist);
	n = 0;
	while(n < nsyms) {
		_(fd, _readat, Ptr(&sym), sizeof(struct nlist),
			symbase+n*sizeof(struct nlist));
		if(sym.n_type & N_SHARE) {
			_(fd, _readat, Ptr(buf), sizeof(buf), strbase+sym.n_strx);
			if(buf[0] == '_')
				_(wd,_create,Label(&buf[1]),sym.n_value, ___);
			else {
			char tmp[128];
				tmp[0] = '.';
				strcpy(&tmp[1],buf);
				_(wd,_create,Label(tmp),sym.n_value,___);
			}
		}
		n++;
	}
}


status_t kernsym_getbyvalue(int val, char *buf, int buflen)
{
obj_t fd;
	fd = dir_parse(kinfo.Root, "/mkern.sym");
	return(sym_getbyvalue(fd, val, buf, buflen));
}

int kernsym_getbyname(char *str)
{
obj_t fd;
	fd = dir_parse(kinfo.Root, "/mkern.sym");
	return(sym_getbyname(fd, str));
}




static obj_t task_init(ObjFrame *frame, Task *task, Task *newtask)
{
obj_t wd;
	wd = _(kinfo.Root, _get, _task, ___, ___);
	if((obj_t)newtask == OBJ_NULL) {
		frame->data = task = (Task*)zalloc(task_zone);
		memset(task,0,sizeof(Task));
		task->ref = 1;
		qclear(&task->mem);
		task->memfree = NULL;
		LOCK_CLEAR(task->memlock);
		vm_task_open(&task->data);
		/* reserve a page to trap NULL references: */
		vm_reserve(&task->data,PAGE_SIZE);
		_(wd, _create, objdata(frame->oid), frame->oid, ___);
	} else {
		frame->data = task = newtask;
		_(wd, _create, _kernel, frame->oid, ___);
	}
	frame->size = sizeof(Task);
	task->self = frame->oid;
	return(OFLAG_HASVEC|OFLAG_KERNEL);
}


static obj_t task_free(ObjFrame *frame, Task *task)
{
obj_t wd;
	if(task && !(--(task->ref))) {
		wd = _(kinfo.Root, _get, _task, ___, ___);
		_(wd, _remove, objdata(task->self), ___, ___);
		vm_task_close(&task->data);
		zfree(task_zone, task);
	}
	return(OK);
}


static obj_t task_mem_alloc(ObjFrame *frame, Task *task, int size)
{
	return((obj_t)task_mem_malloc(task,size));
}

static obj_t task_memfree(ObjFrame *frame, Task *task, void *adr)
{
	task_mem_free(task, adr);
	return(OK);
}

static obj_t task_vm_map(ObjFrame *frame, Task *task, paddr phys, int len)
{
	return((obj_t)vm_map(&task->data,phys,len,VM_URW));
}

static obj_t task_vm_alloc(ObjFrame *frame, Task *task, int len)
{
	return((obj_t)vm_alloc(&task->data,len,VM_URW));
}

static obj_t task_vm_reserve(ObjFrame *frame, Task *task, int len)
{
	return((obj_t)vm_reserve(&task->data,len));
}


static obj_t task_exec(ObjFrame *frame, Task *task, obj_t fd)
{
obj_t rv;
vaddr pc;
Task *oldtask;
	rv = task_load(task,&pc,fd);
	if(err(rv))
		return(rv);
	oldtask = thread->task;
	thread_settask(thread, task);
/*
	kprintf("call_user:  pc=%x stack=%x\n", pc,
		*(ulong*)((vaddr)thread_context + CONTEXT_USP));
*/
	
	if(task == &kernel) 
		rv = ((obj_method)pc)(0,0,0,0,0);
	else  
		rv = call_user(pc,0,0,0,0,0);
	thread_settask(thread, oldtask);
	return(rv);
}


MethodTable task_method[] = {
	{&_init,(obj_method)task_init},
	{&_free,(obj_method)task_free},
	{&_mem_alloc,(obj_method)task_mem_alloc},
	{&_mem_free,(obj_method)task_memfree},
	{&_vm_map,(obj_method)task_vm_map},
	{&_vm_alloc,(obj_method)task_vm_alloc},
	{&_vm_reserve,(obj_method)task_vm_reserve},
	{&_exec,(obj_method)task_exec}};

static ObjClass Class_Task = {
	task_method,
	NULLP(AttrTable),
	arraysize(task_method),
	0};


void Task_init(void)
{
obj_t wd;
	task_zone = zinit(sizeof(Task),"task");
	wd = _(kinfo.Root,_new,_mkdir,_task,___);
	_(wd,_create,_self,OBJ_TASK,___);
	class_task = _(class_object, _class, Ptr(&Class_Task), ___, ___);
	_(class_task, _new, Ptr(&kernel), ___, ___);
}

