/*  thread.c                                                    (CRC:86424bbf)
 *      relative path: ./sys
 *      creation date: 1992
 *      last modified: Wed Dec 18 07:18:19 1996
 *      header update: Wed Dec 18 07:18:20 1996
 *  
 *       Microkernel system files
 *  
 *  Copyright (c) 1992 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is a part of the Shag/OS operating system distribution, and is
 *  free software for non-profit uses; it may be freely used and distributed
 *  for non-profit purposes under the terms of the Shag/OS License, version 1.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND, 
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 *  AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT WILL THE AUTHOR, OR
 *  ANYONE RESPONSIBLE FOR THE LEGAL MODIFICATION OR REDISTRIBUTION OF THIS
 *  SOFTWARE, BE LIABLE FOR ANY DAMAGES, INCLUDING BUT NOT LIMITED TO ANY LOST
 *  PROFITS, LOST SAVINGS OR ANY INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER
 *  RESULTING FROM IMPAIRED OR LOST DATA, SOFTWARE OR COMPUTER FAILURE OR ANY
 *  OTHER CAUSE, EVEN IF THE AUTHOR IS ADVISED OF THE POSSIBILITY OF SUCH
 *  DAMAGES, OR FOR ANY OTHER CLAIM BY THE USER OR FOR ANY THIRD PARTY CLAIM.
 *  
 *  You should have received a copy of the Shag/OS License along with this
 *  program, in a file called 'license1.txt'.  If this license is missing,
 *  please contact the author and describe how this software was obtained.
 *  
 *  The author can be reached for questions, comments, and suggestions at:
 *  
 *      Frank Barrus, 4435 River Rd. Apt #3, Scottsville, NY 14546, USA
 *      email:   shaggy@csh.rit.edu
 *  
 *      For sending bug reports, please read the file 'bugs', and always
 *      include the product name and version, and the word 'BUG' in the subject.
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

/* thread.c
 *  process spawning and control
 *  machine independent functions
 */

#define KERNEL
#include <sys/types.h>
#include <sys/global.h>
#include <sys/resource.h>
#include <limits.h>
#include <event.h>
#include <error.h>
#include <stdio.h>


extern label_t _parent, _name, _cwd, _root, _data;
extern label_t _quantum, _max_quantum, _rtime, _stime, _itime, _ltime;
extern label_t _task, _thread, _data, _self; 
extern label_t _stdin, _stdout, _stderr;
extern label_t _mkdir, _w_param, _w_event;
extern label_t _dvmem, _dpmem, _dsmem;
extern label_t _tvmem, _tpmem, _tsmem;
extern label_t _svmem, _spmem, _ssmem;
extern label_t _msleep, _suspend, _wait, _signal;

static ZoneP thread_zone;

void thread_settask(Thread *node, Task *task);

#include <mdthread.c>


void thread_next(void)
{
	thread->cswitch_enable = cswitch_enable;
	cswitch_enable = FALSE;
	while((Queue*)(thread = (Thread*)qnext(&runq)) == &runq) {
		qremnode((Queue*)null_thread);
		qaddhead(&runq,(Queue*)null_thread);
	} 
	thread->quantum = thread->max_quantum;
	thread_context = &thread->context;	
	context_switch();
	cswitch_enable = thread->cswitch_enable;
}



void thread_settask(Thread *node, Task *task)
{
boolean b;
	if(task == node->task) 
		return;
	CSWITCH_OFF(b)
	if(node->task)
		_(node->task->self, _free, ___, ___, ___);
	node->task = task;
	task->ref++;
	if(task == &kernel)
		vm_task_ptmap(node, NULLP(Vmap));
	else
		vm_task_ptmap(node, &task->data);
	CSWITCH_ON(b)
}


static __inline__ void thread_spawn(Thread *new, obj_t name, vaddr func, obj_t param)
{
	new->name = name;
	new->task = NULL;
	vm_stack_temp_open(&new->stack, &new->temp); 
        new->kstack = (vaddr)vm_stack_alloc(&new->stack, KSTACK_SIZE, VM_KRW);
        new->ustack = (vaddr)vm_stack_alloc(&new->stack, 8*K, VM_URW);
	mdthread_spawn(new, func, param);
	vm_stack_temp_ptmap(new,&new->stack, &new->temp);	
	thread_settask(new,thread->task);
	qaddhead(&runq, (Queue*)new);
	kinfo.tcount++;
}


static __inline__ void thread_destroy(Thread *node)
{
Event *event;
obj_t wd;
	wd = _(kinfo.Root, _get, _thread,___,___);
	_(wd, _remove, objdata(node->self),___,___);
	event = (Event*)zalloc(event_zone);
	event->event = EVENT_CHILDTERM;
	event->param = node->w_param;
	event->sender = node->self;
	event->receiver = node->parent;
	event_send(event);		/* tell the parent */
	vm_stack_temp_close(&node->stack, &node->temp);   
	while(!qempty((Queue*)&node->eventq)) {
		event = (Event*)qremhead((Queue*)&node->eventq);
		zfree(event_zone,event); 
	}
	if(node->task)
		_(node->task->self, _free, ___, ___, ___);
	mdthread_destroy(node);
	zfree(thread_zone,node);
}



static obj_t thread_init(ObjFrame *frame, Thread *new, obj_t name, vaddr func, obj_t  param)
{
obj_t wd;
	frame->data = new = (Thread*)zalloc(thread_zone);
	frame->size = sizeof(Thread);
	if(thread && (Queue*)thread != &runq) {
		memcpy(new,thread,sizeof(Thread));
		new->parent = thread->self;
		if(new->task)
			new->task->ref++;
	} else {
		memset(new,0,sizeof(Thread));
		new->max_quantum = UCHAR_MAX;
	}
	new->cswitch_enable = TRUE;
	new->quantum = new->max_quantum;
	new->rtime = 0;
	new->stime = kinfo.systime;
	qclear((Queue*)&new->eventq);
	new->self = frame->oid;
	wd = _(kinfo.Root, _get, _thread,___,___);
	_(wd, _create, objdata(new->self), new->self,___);
	if(!thread)
		null_thread = new;
	else
		thread_spawn(new, name, func, param);
	return(OFLAG_HASVEC|OFLAG_KERNEL);
}


static obj_t thread_free(ObjFrame *frame, Thread *node, obj_t val)
{
boolean b;
	CSWITCH_OFF(b)
	if(!node->w_event)
		kinfo.tcount--;
	qremnode((Queue*)node);
	qaddtail(&deadq,(Queue*)node);
	qremnode((Queue*)cleanup_thread);
	qaddhead(&runq, (Queue*)cleanup_thread);
	if(cleanup_thread->w_event)
		kinfo.tcount++;
	cswitch_enable = TRUE;
	node->w_param = val;
	if(node == thread)
		thread_next();
	CSWITCH_ON(b)
	return(OK);
}


static obj_t thread_get(ObjFrame *frame, Thread *node, obj_t attr)
{
	if(attr == _data)
		return((obj_t)node);
	if(attr == _ltime)
		return(kinfo.systime-node->stime);
	if(attr == _itime)
		return(kinfo.systime-node->stime-node->rtime);
	if(attr == _dvmem && node->task)
		return(node->task->data.ptlen*PAGE_SIZE);
	if(attr == _dpmem && node->task)
		return(vm_getpsize(&node->task->data,FALSE));
	if(attr == _dsmem && node->task)
		return(vm_getpsize(&node->task->data,TRUE));
	if(attr == _tvmem)
		return(node->temp.ptlen*PAGE_SIZE);
	if(attr == _tpmem)
		return(vm_getpsize(&node->temp,FALSE));
	if(attr == _tsmem)
		return(vm_getpsize(&node->temp,TRUE));
	if(attr == _svmem)
		return(node->stack.ptlen*PAGE_SIZE);
	if(attr == _spmem)
		return(vm_getpsize(&node->stack,FALSE));
	if(attr == _ssmem)
		return(vm_getpsize(&node->stack,TRUE));
	return(ERR_NOATTR);
}


static obj_t thread_suspend(ObjFrame *frame, Thread *node)
{
boolean b;
	CSWITCH_OFF(b)
/* XXX - should check if it's in the run queue first */
	kinfo.tcount--;
	qremnode((Queue*)node);
	node->w_event = EVENT_WAKEUP;
	qaddhead(&waitq,(Queue*)node);
	if(node == thread)
		thread_next();
	CSWITCH_ON(b)
	return(OK);
}


static obj_t thread_signal(ObjFrame *frame, Thread *node,
			   ulong evtype, ulong param, obj_t sender)
{
Event *event;	
	event = (Event*)zalloc(event_zone);
	event->event = evtype;
	event->param = param;
	if(sender == OBJ_NULL)
		sender = thread->self;
	event->sender = sender;
	event->receiver = node->self;
	event_send(event);	
	return(OK);
}


static obj_t thread_msleep(ObjFrame *frame, Thread *node, int msec)
/* if time=0, the current thread just relinquishes CPU
 * otherwise a timer event is set to wake up the
 * process */
{
Event *event;
int n;
boolean b;
	CSWITCH_OFF(b)
	thread->max_quantum += kinfo.quantum_sleep;
	if(thread->max_quantum > kinfo.maxquantum)
		thread->max_quantum = kinfo.maxquantum;
	if(msec == 0) {
		thread_next();
	} else {
		n = (msec*kinfo.hz)/1000;
		event = (Event*)zalloc(event_zone);
		event->event = EVENT_WAKEUP;
		event->param = n;
		event->time = kinfo.systime + n;
		event->sender = thread->self;
		event->receiver = thread->self;
		schedule_event(event);
		waitfor(EVENT_WAKEUP, PARAM_ANY);	
	}
	CSWITCH_ON(b)
	return(0);
}

static obj_t thread_wait(ObjFrame *frame, Thread *node,
			 ulong evtype, ulong param, Event *event)
{
Event *recv;
	if((obj_t)event == OBJ_NULL)
		return(waitfor(evtype,param));
	recv = event_recv(evtype,param);
	memcpy(event,recv,sizeof(Event));
	zfree(event_zone, recv);
	return(event->param);
}


static MethodTable thread_method[] = {
	{&_init,(obj_method)thread_init},
	{&_free,(obj_method)thread_free},
	{&_get,(obj_method)thread_get},
	{&_msleep,(obj_method)thread_msleep},
	{&_signal,(obj_method)thread_signal},
	{&_wait,(obj_method)thread_wait},
	{&_suspend,(obj_method)thread_suspend}};

static AttrTable thread_attr[] = {
	{&_parent, offsetof(Thread, parent), ATYPE_ULONG},
	{&_name, offsetof(Thread, name), ATYPE_OBJ},
	{&_cwd, offsetof(Thread, cwd), ATYPE_OBJ},
	{&_root,offsetof(Thread, root), ATYPE_OBJ},
	{&_quantum, offsetof(Thread, quantum), ATYPE_UCHAR},
	{&_max_quantum, offsetof(Thread, max_quantum), ATYPE_UCHAR},
	{&_rtime, offsetof(Thread, rtime), ATYPE_ULONG},
	{&_stime, offsetof(Thread, stime), ATYPE_ULONG},
	{&_task, offsetof(Thread, task), ATYPE_ULONG},
	{&_stdin, offsetof(Thread, std_obj[0]), ATYPE_OBJ},
	{&_stdout, offsetof(Thread, std_obj[1]), ATYPE_OBJ},
	{&_stderr, offsetof(Thread, std_obj[2]), ATYPE_OBJ},
	{&_w_event, offsetof(Thread, w_event), ATYPE_ULONG},
	{&_w_param, offsetof(Thread, w_param), ATYPE_ULONG}};

static ObjClass Class_Thread = {
	thread_method,
	thread_attr,
	arraysize(thread_method),
	arraysize(thread_attr)};


void load_average(void)
{
	printf("%3d.%02d %3d.%02d %3d.%02d (tcount=%d)\n", 
		kinfo.tload1/100,kinfo.tload1%100,
		kinfo.tload10/100,kinfo.tload10%100,
		kinfo.tload60/100,kinfo.tload60%100, kinfo.tcount);
}


void sys_stat(void)
{
	while(1) {
		load_average();
		_(OBJ_THREAD, _msleep, 1000, ___, ___);
	}
}


static void thread_cleanup(void)
{
Thread *node;
	while(1) {
		while((Queue*)(deadq.next) != &deadq) {
			node = (Thread*)qremhead((Queue*)&deadq);
			thread_destroy(node);
		}
		thread_suspend(NULL, thread);
	}
}




void Thread_init(void)
{
obj_t td, wd;
Thread *new;
	kprintf("thread_init(): ");
	thread_zone = zinit(sizeof(Thread),"thread");
	wd = _(kinfo.Root, _new,_mkdir,_thread,___);
	_(wd, _create, _self, OBJ_THREAD, ___);
	qclear(&runq);
	qclear(&waitq);
	qclear(&deadq);
	class_thread = _(class_object, _class, Ptr(&Class_Thread), ___, ___);
	td = _(class_thread, _new, ___, ___, ___);
	new = null_thread;
	kprintf("thread=%x sizeof(Thread)=%d ", new, sizeof(Thread));
	new->name = Label("null");
	new->root = kinfo.Root;
	new->cwd = kinfo.Root;
	new->task = &kernel;
	thread_context = &new->context;
	qaddhead(&runq,(Queue*)new);
	thread = new;
	kinfo.tcount = 1;
	kinfo.tcount1 = 0; 
	kinfo.tcount10 = 0; 
	kinfo.tcount60 = 0; 
	kinfo.tload1 = 0;
	kinfo.tload10 = 0;
	kinfo.tload60 = 0;
	kinfo.time1 = 0;
	kinfo.time10 = 0;
	kinfo.time60 = 0;
	mdthread_init();
	context_set(thread_context);
	cswitch_enable = TRUE;
	thread_next();
	td = _(OBJ_THREAD, _new, Label("maid"), Ptr(thread_cleanup), ___);
	cleanup_thread = (Thread*)_(td, _get, _data, ___, ___);
	kprintf("done\n");
}

