/*  timer.c                                                     (CRC:d1de2320)
 *      relative path: ./sys
 *      creation date: 1992
 *      last modified: Wed Dec 18 07:18:21 1996
 *      header update: Wed Dec 18 07:18:22 1996
 *  
 *       Microkernel system files
 *  
 *  Copyright (c) 1992 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is a part of the Shag/OS operating system distribution, and is
 *  free software for non-profit uses; it may be freely used and distributed
 *  for non-profit purposes under the terms of the Shag/OS License, version 1.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND, 
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 *  AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT WILL THE AUTHOR, OR
 *  ANYONE RESPONSIBLE FOR THE LEGAL MODIFICATION OR REDISTRIBUTION OF THIS
 *  SOFTWARE, BE LIABLE FOR ANY DAMAGES, INCLUDING BUT NOT LIMITED TO ANY LOST
 *  PROFITS, LOST SAVINGS OR ANY INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER
 *  RESULTING FROM IMPAIRED OR LOST DATA, SOFTWARE OR COMPUTER FAILURE OR ANY
 *  OTHER CAUSE, EVEN IF THE AUTHOR IS ADVISED OF THE POSSIBILITY OF SUCH
 *  DAMAGES, OR FOR ANY OTHER CLAIM BY THE USER OR FOR ANY THIRD PARTY CLAIM.
 *  
 *  You should have received a copy of the Shag/OS License along with this
 *  program, in a file called 'license1.txt'.  If this license is missing,
 *  please contact the author and describe how this software was obtained.
 *  
 *  The author can be reached for questions, comments, and suggestions at:
 *  
 *      Frank Barrus, 4435 River Rd. Apt #3, Scottsville, NY 14546, USA
 *      email:   shaggy@csh.rit.edu
 *  
 *      For sending bug reports, please read the file 'bugs', and always
 *      include the product name and version, and the word 'BUG' in the subject.
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

/* timer.c
 *   routines to control system timing
 *   and also the scheduler
 */

#define KERNEL
#include <sys/types.h>
#include <sys/resource.h>
#include <sys/global.h>
#include <event.h>
#include <mdinfo.h>

void scheduler(void)
{
Event *event;
	if(!thread)
		return;
	while((Queue*)(eventq.next) != &eventq && 
             kinfo.systime >= ((Event*)(eventq.next))->time)	{
		event = (Event*)qremhead((Queue*)&eventq);
		event_send(event);
	}
	if (thread->quantum == 0) {
		if(thread->max_quantum <
				kinfo.minquantum+kinfo.quantum_cpubound)
			thread->max_quantum = kinfo.minquantum;
		else
			thread->max_quantum -= kinfo.quantum_cpubound;
		thread_next();
	}
}


static void clock_svc(void)
{
	kinfo.systime++;
	thread->rtime++;

	if (thread->quantum-- == 0) 
		thread->quantum = 0;

	/* load-average computations */
	kinfo.tcount1 += kinfo.tcount;
	if(++kinfo.time1 >= kinfo.hz) {
		kinfo.tcount10 += kinfo.tcount1;
		kinfo.tload1 = (100*kinfo.tcount1)/kinfo.hz;
		kinfo.tcount1 = 0;
		kinfo.time1 = 0;
		if(++kinfo.time10 >= 10) {
			kinfo.tcount60 += kinfo.tcount10;
			kinfo.tload10 = (100*kinfo.tcount10)/(10*kinfo.hz);
			kinfo.tcount10 = 0;
			kinfo.time10 = 0;
			if(++kinfo.time60 >= 6) {
				kinfo.tload60 = (100*kinfo.tcount60)/(60*kinfo.hz);
				kinfo.tcount60 = 0;
				kinfo.time60 = 0;
			}
		}
	} 
	if(cswitch_enable) 
		scheduler();
}


void schedule_event(Event *event)
{
Queue *evptr;
		LOCK(eventlock);
		evptr = (Queue*)eventq.next;
		while(evptr != &eventq && 	
			((Event*)evptr)->time < event->time)
			evptr = (Queue*)(evptr->next);
		qaddhead((Queue*)evptr->prev,(Queue*)event);
		UNLOCK(eventlock);
}


static ulong clock_calibration;

static __inline__ void clock_calibrate(void)
{
ulong t,ot;
ulong *tp;
	kprintf("calibrating clock: ");
	tp = &kinfo.systime;
	clock_calibration = 1000;
	ot = *tp;
	while(ot == (t = *tp)) 
		ot = t;		/* sync up with clock */
	msleep(1000);
	clock_calibration = (clock_calibration*kinfo.hz)/(*tp-t);
	kprintf("%d\n", clock_calibration);
}

void msleep(int msec)
{
register int i;
	i = clock_calibration*msec;
	while(i--)
		continue;
}



ushort clock_get_hz(void)
{
        return(kinfo.hz);
}

void clock_set_hz(ushort new_hz)
{
int cflag;
        CRITICAL_BEGIN(cflag)
        clock_set_interval(new_hz);
        kinfo.hz = new_hz;
        kinfo.minquantum = MINQUANTUM;
        kinfo.maxquantum = MAXQUANTUM;
        kinfo.quantum_wait = QUANTUM_WAIT;
        kinfo.quantum_sleep = QUANTUM_SLEEP;
        kinfo.quantum_cpubound = QUANTUM_CPUBOUND;
        CRITICAL_END(cflag)
}


void clock_init(void)
{
	kinfo.systime = 0;
	set_interrupt((vaddr)clock_svc, 0, INT_CLOCK);
	clock_set_hz(HZ_DEFAULT);
	clock_calibrate();
}

