/*  trap.c                                                      (CRC:eb4ff029)
 *      relative path: ./sys
 *      creation date: 1992
 *      last modified: Wed Dec 18 07:18:23 1996
 *      header update: Wed Dec 18 07:18:23 1996
 *  
 *       Microkernel system files
 *  
 *  Copyright (c) 1992 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is a part of the Shag/OS operating system distribution, and is
 *  free software for non-profit uses; it may be freely used and distributed
 *  for non-profit purposes under the terms of the Shag/OS License, version 1.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND, 
 *  INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 *  AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT WILL THE AUTHOR, OR
 *  ANYONE RESPONSIBLE FOR THE LEGAL MODIFICATION OR REDISTRIBUTION OF THIS
 *  SOFTWARE, BE LIABLE FOR ANY DAMAGES, INCLUDING BUT NOT LIMITED TO ANY LOST
 *  PROFITS, LOST SAVINGS OR ANY INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER
 *  RESULTING FROM IMPAIRED OR LOST DATA, SOFTWARE OR COMPUTER FAILURE OR ANY
 *  OTHER CAUSE, EVEN IF THE AUTHOR IS ADVISED OF THE POSSIBILITY OF SUCH
 *  DAMAGES, OR FOR ANY OTHER CLAIM BY THE USER OR FOR ANY THIRD PARTY CLAIM.
 *  
 *  You should have received a copy of the Shag/OS License along with this
 *  program, in a file called 'license1.txt'.  If this license is missing,
 *  please contact the author and describe how this software was obtained.
 *  
 *  The author can be reached for questions, comments, and suggestions at:
 *  
 *      Frank Barrus, 4435 River Rd. Apt #3, Scottsville, NY 14546, USA
 *      email:   shaggy@csh.rit.edu
 *  
 *      For sending bug reports, please read the file 'bugs', and always
 *      include the product name and version, and the word 'BUG' in the subject.
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

/* trap.c
 *  generic trap handlers
 */

#define KERNEL
#include <sys/types.h>
#include <sys/resource.h>
#include <sys/global.h>
#include <event.h>
#include <error.h>
#include <stdarg.h>
#include <stdio.h>


char *trap_msg(ulong type)
{
	switch(type&0xff00) {
		case TRAP_INVALID:	return("invalid interrupt");
		case TRAP_FAULT:	return("general fault");
		case TRAP_BKPT:		return("breakpoint");
		case TRAP_RESVOP:	return("rsvd op");
		case TRAP_RESVADR:	return("rsvd addr");
		case TRAP_PRIVINST:	return("priv inst");
		case TRAP_KSTACK:	return("kstack invalid");
		case TRAP_TRACE:	return("trace pending");	
		case TRAP_DOUBLE:	return("dbl fault");
		case TRAP_INVTASK:	return("invalid task");
		case TRAP_ACCV:		return("access violation");
		case TRAP_SEGV:	return("segmentation fault");
		case TRAP_PGFAULT:	return("page fault");
		case TRAP_ARITH: 	switch(type&0xff) {
			case ARITH_INTOVF:	return("int overflow");
			case ARITH_INTDIV0:	return("int divide by 0");
			case ARITH_FPOVF:	return("fp overflow");
			case ARITH_FPDIV0:	return("fp divide by 0");
			case ARITH_FPUNDF:	return("fp underflow");
			case ARITH_DECOVF:	return("dec overflow");	
			case ARITH_RANGE:	return("range error");
			case ARITH_FFPOVF:	return("fault: fp overflow");
			case ARITH_FFPDIV0:	return("fault: fp divide by 0");
			case ARITH_FFPUNDF:	return("fault: fp underflow");
			default:		return("unknown arith trap");
		}
		default:		return("unknown trap");
	}
}



/* if a pc of 0 is specified, the calling pc is looked up,
 * and a stack traceback is also included in the output
 */
void panic(vaddr pc, char *fmt, ...)
{
int cflag;
char buf[80];
boolean trace;
boolean kpanic;
static boolean in_panic_save = FALSE;
boolean in_panic;
va_list args;
	CRITICAL_BEGIN(cflag);
	in_panic = in_panic_save;
	in_panic_save = TRUE;
	kpanic = (thread == null_thread || int_level > 1);
	if(kpanic) {
		kprintf("Kernel Panic: ");
		panic_event.event++;
	} else
		kprintf("Fatal Error: ");
	va_start(args,fmt);
	vkprintf(fmt,args);
	va_end(args);
	trace = FALSE;
	if(pc < 8 && !in_panic) {
		pc = pc_traceback(2+pc);
		trace = !kpanic;
	}
	kprintf("; pc=%08x", pc);
	if(pc > KERNBASE && kinfo.Root && !in_panic) {
		kernsym_getbyvalue(pc, buf, sizeof(buf));
		kprintf("(%s)",buf);
	}
	if(thread && !in_panic) {
		ObjString(buf,sizeof(buf),thread->name);
		kprintf(" (%d:%s)", objdata(thread->self), buf);
		if(!kpanic) 
			kprintf(" -- Killed\r\n");
	}
	kprintf("\n");
	if(trace && !in_panic) 
		stack_traceback(NULL,TRUE);
	in_panic_save = FALSE;
	if(!kpanic)
		_(OBJ_THREAD, _free, ___, ___, ___);
	else
		halt();
}

